// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2006
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// 1.5 XD/ws/code/objectgrid/src/com/ibm/websphere/objectgrid/CopyMode.java, WAS.objectgrid, WASX.XD, b0614.107 2/14/06 17:20:39
//
// Module  :  CopyMode.java
//
// Source File Description:
//
//    Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M3_B.3  WASX.XD   Mar 30, 2005 kjlaw    new code drop
// XD60_M3_B.10 WASX.XD   04/07/2005   sutter    Re-factor datagrid/datagrid.impl
// b22228       7.1.1      1/10/2010   kulack    Changes in support for byte array data serializer
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid;

import com.ibm.websphere.objectgrid.plugins.ObjectTransformer;
import com.ibm.websphere.objectgrid.plugins.io.ValueDataSerializer;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.SerializedValue;

/**
 * This class is used to define the "copy" mode when the
 * <code>setCopyMode</code> method of the <code>BackingMap</code> interface is used.
 * The application is expected to pass one of the final static
 * variables that are defined in this class to the
 * <code>setCopyMode</code> method.
 *
 *
 * @see BackingMap#setCopyMode(CopyMode, Class)
 * @see ObjectTransformer#copyValue(Object)
 *
 * @since WAS XD 6.0, XC10
 */
public final class CopyMode
{
    /**
     * The COPY_ON_READ_AND_COMMIT mode is the default mode.  This mode ensures
     * that an application never has a reference to the value object that is in the
     * BackingMap, and instead the application is always working with a copy of the value that
     * is in the BackingMap.  The copy ensures the application can never
     * inadvertently corrupt the data that is cached in the BackingMap.
     * When an application transaction calls an <code>ObjectMap.get</code> method for a given key,
     * and it is the first access of the ObjectMap entry for that key, a copy
     * of the value is returned.  When the transaction is committed, any changes
     * the application committed are copied to the BackingMap to ensure that
     * the application does not have reference to the committed value in the BackingMap.
     */
    public final static CopyMode COPY_ON_READ_AND_COMMIT = new CopyMode(1);

    /**
     * The COPY_ON_READ mode improves performance over the
     * COPY_ON_READ_AND_COMMIT mode by eliminating the copy that occurs
     * when a transaction is committed.  To preserve integrity of BackingMap
     * data, the application promises to destroy every reference it has to
     * an entry once the transaction is committed.  This mode results in a
     * <code>ObjectMap.get</code> method returning a copy of the value rather than a reference
     * to the value to ensure that changes made by the application to the
     * value does not affect the BackingMap value until the transaction is committed.
     * However, when the transaction does commit, a copy of changes is not made.
     * Instead, the reference to the copy that was returned by ObjectMap.get
     * is stored in the BackingMap.  This is the reason the application
     * must agree to destroy all map entry references once the transaction
     * is committed.  If application fails to keep its promise, the application
     * could cause the data cached in BackingMap to become corrupted.  If an
     * application is using this mode and it is having problems, then switch
     * to the COPY_ON_READ_AND_COMMIT mode to see if the problem still exists.  If the problem
     * goes away, then more than likely the application is failing to destroy
     * all of its references after the transaction has committed.
     */
    public final static CopyMode COPY_ON_READ = new CopyMode(2);

    /**
     * The COPY_ON_WRITE mode improves performance over the COPY_ON_READ_AND_COMMIT
     * mode by eliminating the copy that occurs when <code>ObjectMap.get</code> is called for the
     * first time by a transaction for a given key. Instead, the ObjectMap.get method
     * returns a proxy to the value rather than a direct reference to the value object
     * itself. The proxy ensures that a copy of the value is not made unless the
     * application calls a set method on the value interface that is passed on the
     * <code>BackingMap.setCopyMode(CopyMode, Class)</code> method. Thus, the proxy provides
     * a "copy on write" implementation.  When a transaction commits, the
     * <code>BackingMap</code> examines the proxy to determine if any copy was made
     * as a result of a set method being called.  If a copy was made, then the reference
     * to that copy is stored in the BackingMap. The big advantage of this mode
     * is a value is never copied on read or at commit when the transaction never
     * calls a set method to mutate the value.
     *
     * @see BackingMap#setCopyMode(CopyMode, Class)
     */
    public final static CopyMode COPY_ON_WRITE = new CopyMode(3);

    /**
     * The NO_COPY mode allows an application to promise that it will
     * never modify a value object obtained using an ObjectMap.get method
     * in exchange for performance improvements.  If this mode is used,
     * no copy of the value is ever made.  If the application breaks its
     * promise and does modify values, then data in the BackingMap will be corrupted.
     * This mode is primarily useful for read only maps where data
     * is never modified by the application.  If the application is using this
     * mode and it is having problems, then switch to COPY_ON_READ_AND_COMMIT
     * mode to see if the problem still exists.  If the problem
     * goes away, then more than likely the application is not keeping its
     * promise and is modifying the value returned by ObjectMap.get method (either
     * during transaction or after transaction has committed).
     */
    public final static CopyMode NO_COPY = new CopyMode(4);

    /**
     * The COPY_TO_BYTES mode is similar to the COPY_ON_READ_AND_COMMIT mode in that it ensures
     * that an application never has a reference to the value object that is in the
     * BackingMap.  The value that the application works with is a newly inflated
     * version of the serialized version that is in the BackingMap. The copy ensures the application
     * can never inadvertently corrupt the data that is cached in the BackingMap since a byte form
     * of the value is what is stored in the BackingMap instead of the Object form.<p>
     *
     * A copy of the value is returned when an application transaction calls an
     * <code>ObjectMap.get</code> method for a given key,
     * and it is the first time that the ObjectMap entry is accessed for that key.
     * When the transaction is committed, any changes
     * the application committed are copied to bytes in the BackingMap to ensure that
     * the application does not have reference to the committed value in the BackingMap.
     * @since 7.0
     */
    public final static CopyMode COPY_TO_BYTES = new CopyMode(5);

    /**
     * When set, all ObjectMap APIs that return a {@link SerializedValue} rather than the
     * original Java Object, allowing access to the serialized form of the data, preventing inflation
     * of object into Java Object form.
     * @since 7.1.1
     * @see ValueDataSerializer
     */
    public final static CopyMode COPY_TO_BYTES_RAW = new CopyMode(6);

    /**
     * Id that uniquely identifies the CopyMode instance.
     */
    final private int ivModeId;

    /**
     * Hidden CTOR for enumerated type idiom to use.
     *
     * @param modeId is the unique id for the contructed CopyMode instance.
     */
    private CopyMode( int modeId )
    {
        ivModeId = modeId;
    }

    /**
     * Hidden default CTOR for enumerated type idiom to use.
     */
    private CopyMode()
    {
        this(1);
    }

    /**
     * Returns a string representation of the <code>CopyMode</code>.
     *
     * @return  a string representation of the <code>CopyMode</code>.
     */
    public String toString() {
        switch (ivModeId) {
        case 1: return "COPY_ON_READ_AND_COMMIT CopyMode"; //$NON-NLS-1$
        case 2: return "COPY_ON_READ CopyMode"; //$NON-NLS-1$
        case 3: return "COPY_ON_WRITE CopyMode"; //$NON-NLS-1$
        case 4: return "NO_COPY CopyMode"; //$NON-NLS-1$
        case 5: return "COPY_TO_BYTES CopyMode"; //$NON-NLS-1$
        case 6: return "COPY_TO_BYTES_RAW CopyMode"; //$NON-NLS-1$
        default: return "UNKNOWN CopyMode"; //$NON-NLS-1$
        }
    }

    /**
     * Is the copy mode one of the copy modes that indicate
     * copy to bytes?
     *
     * @return boolean indicating if copy mode is one of COPY_TO_BYTES
     * or COPY_TO_BYTES_RAW.
     */
    public boolean isBytes() {
        if (this == COPY_TO_BYTES_RAW || this == COPY_TO_BYTES) {
            return true;
        }
        return false;
    }
}

