// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Source File Description:
//
//  Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// b11146       WASX.XD   04/15/2008   jian.tang time-based database update
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;

import com.ibm.websphere.objectgrid.jpa.JPAPropertyFactory;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.util.SerializationHelper;

/**
 * This TimeBasedDBUpdateConfig represents a configuration for a time-based database
 * updater. A TimeBasedDBUpdateConfig object contains information on how to get
 * the newly inserted and updated records from the database and how to update the data
 * in the correspondent ObjectGrid maps.
 * <p>
 * Users use {@link BackingMap#setTimeBasedDBUpdateConfig(TimeBasedDBUpdateConfig)} method
 * to set a TimeBasedDBUpdateConfig in the backing map configuration.
 * <p>
 * A time-based database updater periodically interacts with Java Persistence
 * API (JPA) to get the latest changes (inserts and updates) from the database using a specific
 * timestamp field. Therefore, a JPA provider is required to use this feature.
 * The database here can be any backend supported by the chosen JPA provider.
 * <P>
 * The time-based database updater periodically query the database using JPA interfaces to
 * get the JPA entities which represent the newly inserted and updated records in the database.
 * In order to do that, every record in the database should have a "timestamp" to identify the time
 * or sequence when the record was last updated or inserted. The "timestamp" here does not have to be
 * in a timestamp format. It can be in an integer or long format, as long as it generates a unique
 * increasing value.
 * <p>
 * If a TimeBasedDBUpdateConfig is defined in the BackingMap or a "timeBasedDBUpdate" element is configured
 * for a backing map in the XML configuration, a time-based database updater is automatically launched.
 * For the distributed ObjectGrid, it will only be launched in the partition 0.
 *
 * @see BackingMap#setTimeBasedDBUpdateConfig(TimeBasedDBUpdateConfig)
 * @since WAS XD 6.1.0.3
 */
public class TimeBasedDBUpdateConfig implements Externalizable {

    private static final long serialVersionUID = 1L;

    /**
     * An type of DBUpdateMode specifies how the database changes
     * should be pushed to the ObjectGrid maps.
     */
    public static class DBUpdateMode {
        private byte type = -1;

        private DBUpdateMode(byte i) {
            type = i;
        }

        /**
         * INVALIDATE_ONLY type indicates to invalidate the entries in the
         * ObjectGrid map if the corresponding records in the database have changed.
         */
        public final static DBUpdateMode INVALIDATE_ONLY = new DBUpdateMode((byte) 1);

        /**
         * UPDATE_ONLY type indicates to update the existing entries in the ObjectGrid
         * map with the latest values from database. However, all the newly inserted records
         * to the database are ignored.
         */
        public final static DBUpdateMode UPDATE_ONLY = new DBUpdateMode((byte) 2);

        /**
         * INSERT_UPDATE type indicates to update the existing entries in the ObjectGrid
         * map with the latest values from database. Also, all the newly inserted records
         * to the database are inserted into the ObjectGrid map.
         */
        public final static DBUpdateMode INSERT_UPDATE = new DBUpdateMode((byte) 3);

        public String toString() {
            switch (type) {
            case 1:
                return Constants.DBUPDATE_INVALIDATE_ONLY_STRING;
            case 2:
                return Constants.DBUPDATE_UPDATE_ONLY_STRING;
            case 3:
                return Constants.DBUPDATE_INSERT_UPDATE_STRING;
            default:
                return "UNKNOWN";
            }
        }
    }

    private DBUpdateMode mode = null;

    private String puName = null;

    private String entityClassName = null;

    private String tsField = null;

    private JPAPropertyFactory propertyFactory = null;
    
    /**
     * Set the entity class name used to interact with the Java Persistence API (JPA)
     * provider. The entity class name is used to get JPA entities which represent database records.
     *
     * @param entityClassName The entityClassName to set.
     */
    public void setEntityClassName(String entityClassName) {
        this.entityClassName = entityClassName;
    }

    /**
     * Set the time-based database update mode.
     * <p>
     * By default, the time-based database update mode is set to {@link DBUpdateMode#INVALIDATE_ONLY}.
     * @param mode the time-based database update mode
     * @see DBUpdateMode
     */
    public void setDBUpdateMode(DBUpdateMode mode) {
        this.mode = mode;
    }

    /**
     * Set the JPA persistence unit name.
     * <p>
     * By default, the persistence unit name is the name of the first persistence unit defined in your
     * persistence.xml file.
     * @param persistenceUnitName the JPA persitence unit name to set.
     */
    public void setPersistenceUnitName(String persistenceUnitName) {
        this.puName = persistenceUnitName;
    }

    /**
     * The timestamp field is a field of the entity class. 
     * A timestamp field is used to identify the time or sequence when a
     * database backend record was last updated or inserted.
     * <p>
     * If a timestamp field is not set, then ObjectGrid will try to get the
     * annotated {@link com.ibm.websphere.objectgrid.jpa.dbupdate.annotation.Timestamp}
     * from the entity class. If the annotation cannot be found, an
     * IllegalArgumentException will be thrown.
     * 
     * @param timestampField the timestamp field to set.
     */
    public void setTimestampField(String timestampField) {
        this.tsField = timestampField;
    }

    /**
     * @return the entity class name used for interacting with JPA.
     */
    public String getEntityClassName() {
        return entityClassName;
    }

    /**
     * @return the time-based database update mode.
     */
    public DBUpdateMode getDBUpdateMode() {
        return mode;
    }

    /**
     * @return the persistence unit name set on this object.
     */
    public String getPersistenceUnitName() {
        return puName;
    }

    /**
     * @return the timestamp field set on this object.
     */
    public String getTimestampField() {
        return this.tsField;
    }

    /**
     * Set the JPA property factory to override the default persistence properties. 
     * @param map the overriding JPA property factory
     */
    public void setJpaPropertyFactory(JPAPropertyFactory map) {
        propertyFactory = map;
    }
    
    /**
     * @return the JPA property factory used to override the default persistence properties.
     */
    public JPAPropertyFactory getJpaPropertyFactory() {
        return propertyFactory;
    }
    
    public void readExternal(ObjectInput in) throws IOException, ClassNotFoundException {
        byte b = in.readByte();
        switch (b) {
        case 1:
            mode = DBUpdateMode.INVALIDATE_ONLY;
            break;
        case 2:
            mode = DBUpdateMode.UPDATE_ONLY;
            break;
        case 3:
            mode = DBUpdateMode.INSERT_UPDATE;
            break;
        default:
            mode = null;
        }

        puName = SerializationHelper.readNullableUTF(in);
        entityClassName = SerializationHelper.readNullableUTF(in);
        tsField = SerializationHelper.readNullableUTF(in);

    }

    public void writeExternal(ObjectOutput out) throws IOException {
        if (mode == DBUpdateMode.INVALIDATE_ONLY) {
            out.writeByte(1);
        } else if (mode == DBUpdateMode.UPDATE_ONLY) {
            out.writeByte(2);
        } else if (mode == DBUpdateMode.INSERT_UPDATE) {
            out.writeByte(3);
        } else {
            out.writeByte(0);
        }
        SerializationHelper.writeNullableUTF(out, puName);
        SerializationHelper.writeNullableUTF(out, entityClassName);
        SerializationHelper.writeNullableUTF(out, tsField);
    }
}
