// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2011
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  ObjectGridConfigFactory.java
//
// Source File Description:
//
//    Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M3_B.x  WASX.XD   02/23/2005   rajeshk   new code drop
// d344949      WXD601.XD 02/07/2006   jhanders  move server security configuration out of public interface
// 344949       WXD601.XD 02/10/2006   vandersn  cleanup and updates for classes moved to private package
// b896                   05/01/2006   vandersn  open config APIs
// b28
// b21871       XS7.1.1   02/20/2011   jian.tang add createOSGiServicePlugin method
//
package com.ibm.websphere.objectgrid.config;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.ObjectGridRuntimeException;
import com.ibm.ws.ffdc.FFDCFilter;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.config.IObjectGridConfiguration;
import com.ibm.ws.objectgrid.config.InternalConfigFactory;
import com.ibm.ws.xs.util.dopriv.DoPrivUtil;

/**
 * <p>This is the configuration factory for ObjectGrid configuration entities. Users are expected to
 * use static methods of this factory to create ObjectGrid configuration objects.</p>
 *
 * <p>Here is a list of static methods used to create configuration objects:</p>
 * <ul>
 * <li><code>createObjectGridConfiguration(String)</code>: create an ObjectGridConfiguration object</li>
 * <li><code>createBackingMapConfiguration(String)</code>: create a <code>BackingMapConfiguration</code> object by passing a backing map name</li>
 * <li><code>createConfigProperty(ConfigPropertyType, String, String)</code>: create a <code>ConfigProperty</code> object</li>
 * <li><code>createPlugin(PluginType, String)</code>: create a <code>Plugin</code> object</li>
 * </ul>
 *
 * <p>Below is an example of creating an ObjectGrid configuration.
 * A Plugin is added to the <code>ObjectGridConfiguration</code> object:
 * <code>com.ibm.websphere.objectgrid.plugins.ObjectGridEventListener</code> plugin.
 * </p>
 *
 * <p>A <code>BackingMapConfiguration</code> called "myBackingMap" is then created and added to the <code>ObjectGridConfiguration</code>.
 * This <code>BackingMapConfiguration</code> also has an Evictor Plugin configured. </p>
 *
 * <p>Once the ObjectGridConfiguration object has been created , it can be used to call either of these methods
 * <ul>
 * <li><code>com.ibm.websphere.objectgrid.ObjectGridManager.putOverrideObjectGridConfigurations(String, List)</code></li>
 * <li><code>com.ibm.websphere.objectgrid.ObjectGridManager.setOverrideObjectGridConfigurations(Map)</code></li>
 * </ul>
 * to set configuration objects, prior to connecting.
 * <p><pre>
 // Create an ObjectGridConfiguration object
 ObjectGridConfiguration ogConfig = ObjectGridConfigFactory.createObjectGridConfiguration(ogName);

 // create ObjectGridEventListener plugin
 Plugin eventListener = ObjectGridConfigFactory.createPlugin(PluginType.OBJECTGRID_EVENT_LISTENER,"com.ibm.test.MyOgEventListener");

 // Add plugin to ObjectGridConfiguration object
 ogConfig.addPlugin(eventListener);

 // Create a BackingMapConfiguration object
 BackingMapConfiguration bmConfig = ObjectGridConfigFactory.createBackingMapConfiguration("mybackingMap");

 // Add BackingMapConfiguration object to ObjectGridConfiguration object
 ogConfig.addBackingMapConfiguration(bmConfig);

 // Set the number of buckets to 1000
 bmConfig.setNumberOfBuckets(1000);

 // Create a Evictor plugin for this backing map.
 Plugin evictor = ObjectGridConfigFactory.createPlugin(
 BackingMapConfiguration.PLUGIN_EVICTOR,
 com.acme.myEvictorImpl.class.getName());

 // add Evictor Plugin to the BackingMapConfiguration
 bmConfig.addPlugin(evictor);

 // Create a ConfigProperty for the Evictor plugin
 ConfigProperty sizeProperty = ObjectGridConfigFactory.createConfigProperty(
   ConfigPropertyType.INT_PRIM, "size", "153");

 // Add the ConfigProperty to the Evictor plugin
 evictor.setConfigProperty(sizeProperty);
 * </pre></p>
 * @see com.ibm.websphere.objectgrid.config.ObjectGridConfiguration
 * @see com.ibm.websphere.objectgrid.config.BackingMapConfiguration
 * @see com.ibm.websphere.objectgrid.config.Plugin
 * @see com.ibm.websphere.objectgrid.config.ConfigProperty
 * @since WAS XD 6.0.1.2, XC10
 */
public final class ObjectGridConfigFactory {

    private static String PLUGIN_IMPL = "com.ibm.ws.objectgrid.config.PluginImpl";

    private static String OBJECTGRID_CONFIG_IMPL = "com.ibm.ws.objectgrid.config.ObjectGridConfigurationImpl";

    private static String CONFIG_PROPERTY_IMPL = "com.ibm.ws.objectgrid.config.ConfigPropertyImpl";

    private static final String CLASS_NAME = "com.ibm.websphere.objectgrid.config.ObjectGridConfigFactory";

    private static final TraceComponent tc = Tr.register(
        ObjectGridConfigFactory.class,
        Constants.TR_GROUP_NAME,
        Constants.TR_RESOURCE_BUNDLE_NAME);

    /**
     * Create a Plugin for a BackingMapConfiguration or an ObjectGridConfiguration
     *
     * @param pluginType
     *            the PluginType.
     * @param className
     *            of the Plugin implementation class to instantiate
     * @return a Plugin instance
     * @see PluginType#OBJECTGRID_EVENT_LISTENER
     * @see PluginType#TRANSACTION_CALLBACK
     * @see PluginType#COLLISION_ARBITER
     * @see PluginType#EVICTOR
     * @see PluginType#MAP_EVENT_LISTENER
     * @see PluginType#OBJECTGRID_LIFECYCLE_LISTENER
     * @see PluginType#MAP_LIFECYCLE_LISTENER
     */
    public static Plugin createPlugin(PluginType pluginType, String className) {
        Plugin pt = null;
        try {
            pt = (Plugin) DoPrivUtil.forName(PLUGIN_IMPL)
                    .newInstance();
            pt.setPluginType(pluginType);
            pt.setClassName(className);
        }
        catch (InstantiationException e) {
            FFDCFilter.processException(e, CLASS_NAME + ".createPlugin", "80");
            if (tc.isDebugEnabled()) {
                Tr.debug(tc, "InstantiationException " + e);
            }
            throw new ObjectGridRuntimeException("InstantiationException", e);
        }
        catch (IllegalAccessException e) {
            FFDCFilter.processException(e, CLASS_NAME + ".createPlugin", "86");
            if (tc.isDebugEnabled()) {
                Tr.debug(tc, "IllegalAccessException " + e);
            }
            throw new ObjectGridRuntimeException("IllegalAccessException", e);
        }
        catch (ClassNotFoundException e) {
            FFDCFilter.processException(e, CLASS_NAME + ".createPlugin", "92");
            if (tc.isDebugEnabled()) {
                Tr.debug(tc, "ClassNotFoundException " + e);
            }
            throw new ObjectGridRuntimeException("ClassNotFoundException", e);
        }
        return pt;
    }

    /**
     * Create an ObjectGridConfiguration object.
     * @param objectGridConfigName the name that will be assigned to this ObjectGridConfiguration object
     * @return the ObjectGridConfiguration object
     */
    public static ObjectGridConfiguration createObjectGridConfiguration(String objectGridConfigName) {
        IObjectGridConfiguration ogcConfig = null;
        try {
            ogcConfig = (IObjectGridConfiguration) DoPrivUtil
                .forName(OBJECTGRID_CONFIG_IMPL)
                .newInstance();
            ogcConfig.setName(objectGridConfigName);
        }
        catch (InstantiationException e) {
            FFDCFilter.processException(e, CLASS_NAME + ".createObjectGridConfiguration", "111");
            if (tc.isDebugEnabled()) {
                Tr.debug(tc, "InstantiationException " + e);
            }
            throw new ObjectGridRuntimeException("InstantiationException", e);
        }
        catch (IllegalAccessException e) {
            FFDCFilter.processException(e, CLASS_NAME + ".createObjectGridConfiguration", "117");
            if (tc.isDebugEnabled()) {
                Tr.debug(tc, "IllegalAccessException " + e);
            }
            throw new ObjectGridRuntimeException("IllegalAccessException", e);
        }
        catch (ClassNotFoundException e) {
            FFDCFilter.processException(e, CLASS_NAME + ".createObjectGridConfiguration", "123");
            if (tc.isDebugEnabled()) {
                Tr.debug(tc, "ClassNotFoundException " + e);
            }
            throw new ObjectGridRuntimeException("ClassNotFoundException", e);
        }

        return ogcConfig;
    }
    /**
     * Create a BackingMapConfiguration object
     * @param backingMapConfigName the name to assign to this BackingMapConfiguration
     * @return the BackingMapConfiguration object
     */
    public static BackingMapConfiguration createBackingMapConfiguration(String backingMapConfigName) {
        return InternalConfigFactory.createBackingMapConfiguration(backingMapConfigName,false);
    }

    /**
     * Create a ConfigPropety object for use on a Plugin.
     *
     * <p>The Plugin should have a set method that corresponds to the name of this ConfigProperty.
     * The method must accept a parameter of the ConfigPropertyType that is specified on this ConfigProperty.
     * For example, if the name of this ConfigProperty is set to "size", and the type is
     * <code>ConfigPropertyType.INT_PRIM</code>,
     * then the Plugin must have the method setSize(int).  The value of the ConfigProperty will be passed
     * to the setter of the Plugin when an ObjectGrid is created based on this configuration.
     * @param configPropType ConfigPropertyType of the ConfigProperty.  Part of the set method's signature, the type of parameter the set method requires.  Valid
     * @param name of the ConfigProperty.  It must correspond to the name of a set method on the Plugin.
     * @param value of the ConfigProperty. This value will be passed to the set method on the Plugin.
     *
     * @return the ConfigProperty object
     * @see ConfigPropertyType#STRING_JAVA_LANG
     * @see ConfigPropertyType#BOOLEAN_JAVA_LANG
     * @see ConfigPropertyType#BOOLEAN_PRIM
     * @see ConfigPropertyType#BYTE_JAVA_LANG
     * @see ConfigPropertyType#BYTE_PRIM
     * @see ConfigPropertyType#CHARACTER_JAVA_LANG
     * @see ConfigPropertyType#CHAR_PRIM
     * @see ConfigPropertyType#DOUBLE_JAVA_LANG
     * @see ConfigPropertyType#DOUBLE_PRIM
     * @see ConfigPropertyType#FLOAT_JAVA_LANG
     * @see ConfigPropertyType#FLOAT_PRIM
     * @see ConfigPropertyType#INTEGER_JAVA_LANG
     * @see ConfigPropertyType#INT_PRIM
     * @see ConfigPropertyType#LONG_JAVA_LANG
     * @see ConfigPropertyType#LONG_PRIM
     * @see ConfigPropertyType#SHORT_JAVA_LANG
     */
    public static ConfigProperty createConfigProperty(ConfigPropertyType configPropType, String name, String value) {

        ConfigProperty cpt = null;
        try {
            cpt = (ConfigProperty) DoPrivUtil.forName(CONFIG_PROPERTY_IMPL).newInstance();
            cpt.setName(name);
            cpt.setConfigPropertyType(configPropType);
            cpt.setValue(value);
        }
        catch (InstantiationException e) {
            FFDCFilter.processException(e, CLASS_NAME + ".createConfigProperty", "264");
            if (tc.isDebugEnabled()) {
                Tr.debug(tc, "InstantiationException " + e);
            }
            throw new ObjectGridRuntimeException("InstantiationException", e);
        }
        catch (IllegalAccessException e) {
            FFDCFilter.processException(e, CLASS_NAME + ".createConfigProperty", "270");
            if (tc.isDebugEnabled()) {
                Tr.debug(tc, "IllegalAccessException " + e);
            }
            throw new ObjectGridRuntimeException("IllegalAccessException", e);
        }
        catch (ClassNotFoundException e) {
            FFDCFilter.processException(e, CLASS_NAME + ".createConfigProperty", "276");
            if (tc.isDebugEnabled()) {
                Tr.debug(tc, "ClassNotFoundException " + e);
            }
            throw new ObjectGridRuntimeException("ClassNotFoundException", e);
        }

        return cpt;
    }

    /**
     * Create an OSGi Service Plugin for a BackingMapConfiguration or an ObjectGridConfiguration
     *
     * @param pluginType
     *            the PluginType.
     * @param osgiServiceName
     *            the OSGi service name
     * @return a Plugin instance
     * @see PluginType#OBJECTGRID_EVENT_LISTENER
     * @see PluginType#TRANSACTION_CALLBACK
     * @see PluginType#COLLISION_ARBITER
     * @see PluginType#EVICTOR
     * @see PluginType#MAP_EVENT_LISTENER
     * @see PluginType#OBJECTGRID_LIFECYCLE_LISTENER
     * @see PluginType#MAP_LIFECYCLE_LISTENER
     *
     * @since 7.1.1
     */
    public static Plugin createOSGiServicePlugin(PluginType pluginType, String osgiServiceName) {
        Plugin pt = null;
        try {
            pt = (Plugin) DoPrivUtil.forName(PLUGIN_IMPL)
                    .newInstance();
            pt.setPluginType(pluginType);
            pt.setOSGiService(osgiServiceName);
        }
        catch (Exception e) {
            FFDCFilter.processException(e, CLASS_NAME + ".createOSGiServicePlugin", "300");
            if (tc.isDebugEnabled()) {
                Tr.debug(tc, "Exception " + e);
            }
            throw new ObjectGridRuntimeException("Exception when creating an OSGi service Plugin", e);
        }
        return pt;
    }

}