// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2012
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  ContinuousQueryNotificationEvent.java
//
// Source File Description:
//
// An interface which defines notification events sent to continuous query listeners.

package com.ibm.websphere.objectgrid.continuousquery;

import com.ibm.websphere.objectgrid.Session;
import com.ibm.websphere.objectgrid.SessionHandle;

/**
 * An interface which defines notification metadata that is sent to continuous
 * query listeners when the result set of a defined continous query changes.
 *
 * @param <KeyType>  Type of the key object in this event, from the key type used in the map being queried
 * @param <ValueType> Type of the value object in this event, from the value type used in the map being queried
 * @since 8.6
 */
public interface ContinuousQueryNotificationEvent<KeyType,ValueType>  {

	/**
	 * 
	 * Indicates the reason a notification event has occurred for this cache entry.
	 *
	 */
	public enum Reason { 
		/**
		 * The cache entry was inserted into the grid and matches the query, or an entry which did not match the
		 * query was updated such that it now matches the query.
		 */
		ADDED, 
		/**
		 * The cache entry, which previously matched the query, was removed from the grid or updated such that it no longer matches.
		 */
		REMOVED, 
		/**
		 * The cache entry, which previously matched the query, was updated and still matches the query.
		 * Note: The result of getValue() for REMOVE events will always be null, even if the query was defined to include values.
		 * @see ContinuousQueryManager#defineContinuousQuery
		 */
		UPDATED, 
		/**
		 * A cache entry which matched the query was removed from the grid due to a map clear operation.
		 */
		CLEAR }
	
	/**
	 * Retrieve the key of the cache entry that triggered this notification
	 * 
	 * @return the cache entry key
	 */
	public KeyType getKey();
	
	/**
	 * Retrieve the value of the cache entry that triggered this notification
	 * 
	 * @return the cache entry value
	 */
	public ValueType getValue();
	
	/**
	 * Retrieve a SessionHandle object which can be used to retrieve the cache entry that triggered this notification.
	 * 
	 * @return the SessionHandle for the partition that owns the cache entry.
	 * @see Session#setSessionHandle(SessionHandle)
	 */
	public SessionHandle getSessionHandle();
	
	/**
	 * Retrieve the reason for this notification event 
	 * 
	 * @return the Reason enum 
	 */
	public Reason getReason();	

	/**
	 * If true, this notification only includes the key of the matching cache entry.
	 * If false, this notification also includes the value 
	 * @see ContinuousQueryNotificationEvent#getValue()
	 * @return whether or not this notification includes the cache value.
	 */
	public boolean isKeysOnly();
}
