// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2012
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  CompareFilter.java
//
// Source File Description:
//
// This is a base class for filters comparing an attribute with a scalar.
package com.ibm.websphere.objectgrid.continuousquery.filter;

import java.util.Arrays;

import com.ibm.websphere.objectgrid.continuousquery.ContinuousQueryFilter;
import com.ibm.websphere.objectgrid.io.XsDataInputStream;
import com.ibm.websphere.objectgrid.plugins.ObjectTransformer;
import com.ibm.websphere.objectgrid.plugins.io.DataSerializer.DataAttributeInflatable;
import com.ibm.websphere.objectgrid.plugins.io.datadescriptor.DataDescriptor;
import com.ibm.websphere.objectgrid.plugins.io.datadescriptor.MapDataDescriptor;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.DataObjectContext;

/**
 * This is a base class for filters comparing an attribute with a scalar.
 *
 * @param <KeyType>  Type of the key object for the map being queried
 * @param <ValueType> Type of the value object for the map being queried
 * @param <AttributeType> Type of the attribute referenced by the attribute path
 * @param <MatchType> Type of the object being compared to
 * @since 8.6
 */
public abstract class CompareFilter<KeyType,ValueType,AttributeType,MatchType> extends AbstractCQFilter<KeyType,ValueType,AttributeType,MatchType> {

    private static final long serialVersionUID = 7451883230616103005L;
    protected String attributePath;
    protected MatchType objectToCompare;
    
    public CompareFilter() { }

    /**
     * Construct a filter for comparing an arbitrary object with the filtering criteria.
     * 
     * The attribute path format is dependent on how the associated BackingMap is configured:
     * <ul>
     * <li><b>POJO</b> - A map that stores Java Objects without an EntityManager API entity or MapSerializerPlugin configured.  
     * The filter uses reflection to interact with the objects.
     * <li><b>MapSerializerPlugin</b> - A map that has a MapSerializerPlugin configured.
     * <li><b>Entity</b> - A map that has an EntityManager API entity associated - not supported by continuous query. 
     * </ul><p>
     * <b>POJO maps</b><br>
     * For POJO maps which use traditional java serialization, or a custom {@link ObjectTransformer}, attributes are accessed using reflection. The field name and property name are defined by the 
     * JavaBeans specification, or the method name defined in the class. 
	 * By default paths are assumed to reference attributes on the value object associated with the cache entry.
     * The path delimiter is ".", which is specified in {@link ContinuousQueryFilter#POJO_PATHSEPARATOR}.
     * <p>
     * A key object attribute is specified by including the "key" path prefix, which is specified in {@link ContinuousQueryFilter#POJO_ADDRESSABLEKEYNAME}.
     * <p>
     * To specify a method invocation, append "()" to the method name.
     * <p>
     * If null or an empty string is specified, the entire value object is used as the attribute value.
     * <p>
     * If {@link ContinuousQueryFilter#POJO_ADDRESSABLEKEYNAME} is specified by itself, the entire key object is used as the attribute value. 
     * <p>
     * 
     * POJO attribute path grammar:
     * <pre>
	 * &lt;attributePath&gt; ::= NULL | "" | "key" | &lt;repeatableSegment&gt; | "key" &lt;pathSeparator&gt; &lt;repeatableSegment&gt;
	 * &lt;repeatableSegment&gt; ::= &lt;segment&gt; | &lt;segment&gt; &lt;pathSeparator&gt; &lt;repeatableSegment&gt; 
 	 * &lt;segment&gt; ::= &lt;attributeName&gt; | &lt;methodName&gt; "()"
 	 * &lt;attributeName&gt; ::= a valid field name on the POJO
 	 * &lt;methodName&gt; ::= a no argument public method on the POJO
     * </pre>
     * 
     * POJO attribute path examples:<br>
     * <ul>
     * <li><code>name</code> - The "name" attribute in the value object.
     * <li><code>key.ssn</code> - The "ssn" attribute in the key object.
     * <li><code>employee.store.totalSales()</code> - The result of the "totalSales" method from the "store" embedded type from the "employee" embedded type in the value object. 
     * </ul>
     * <p>
     * <br>
     * <b>MapSerializerPlugin maps</b><br>
     * For maps that use a MapSerializerPlugin, the attribute path is passed directly 
     * to the {@linkplain DataAttributeInflatable#getAttributeContexts(String...)} method
     * and the attribute values are fetched using the 
     * {@linkplain DataAttributeInflatable#inflateDataObjectAttributes(DataObjectContext, XsDataInputStream, Object)}
     * method. The attribute path is a path to an attribute within the key or value.  
     * <p>
     * A key attribute is specified by including the "key" path prefix, or the prefix specified by the {@link MapDataDescriptor#getAddressableKeyName()} method.
     * <p> 
     * The path delimiter is specified in the {@link DataDescriptor#getPathSeparator()} for the DataDescriptor for each KeyDataSerializer or ValueSerializer.
     * <p>
     * If null or an empty string is specified, the entire value object is used as the attribute value.
     * <p>
     * If {@link MapDataDescriptor#getAddressableKeyName()} is specified by itself, the entire key object is used as the attribute value. 
     * <p>
     * 
     * Serializer attribute path grammar:
     * <pre>
     * &lt;attributePath&gt; ::= NULL | "" | &lt;addressableKeyName&gt; | &lt;repeatableSegment&gt; | &lt;addressableKeyName&gt; &lt;pathSeparator&gt; &lt;repeatableSegment&gt;
	 * &lt;repeatableSegment&gt; ::= &lt;attributeName&gt; | &lt;attributeName&gt; &lt;pathSeparator&gt; &lt;repeatableSegment&gt; 
     * &lt;attributeName&gt; ::= a valid field name on the POJO
     * &lt;addressableKeyName&gt; ::= "key" or the key identifier defined in {@link MapDataDescriptor#getAddressableKeyName()} 
     * &lt;pathSeparator&gt; ::= the path separator defined in {@link DataDescriptor#getPathSeparator()}
     * </pre>
     * 
     * 
     * MapSerializerPlugin attribute path examples:<br>
     * (assuming the default addressable key name of "key" and path separator of ".")
     * <ul>
     * <li><code>name</code> - Indexes the "name" attribute in the value object.
     * <li><code>key.ssn</code> - Indexes the "ssn" attribute in the key object.
     * <li><code>employee.store</code> - The "store" attribute from the "employee" embedded type in the value. 
     * </ul>
     * Note: In this scenario all processing of the attribute path is dependent on the MapSerializerPlugin implementation.  MapSerializerPlugin implementations can
     * implement method invocation as part of the attribute path processing, but are not required to do so. 
	 * <p>
	 * <br>
	 * <b>eXtreme Data Format maps</b><br>
     * For maps using the eXtreme Data Format (XDF), the attribute path is passed directly 
     * to the XDF component which will extract the attribute.  The attribute path is a path to an attribute within the key or value.
     * <p>  
     * A key attribute is specified by including the "key" path prefix, or the prefix specified by the {@link MapDataDescriptor#getAddressableKeyName()} method.
     * <p> 
     * The path delimiter is specified in the {@link DataDescriptor#DEFAULT_PATHSEPARATOR}.
     * <p>
     * If null or an empty string is specified, the entire value object is used as the attribute value.
     * <p>
     * If {@link MapDataDescriptor#getAddressableKeyName()} is specified by itself, the entire key object is used as the attribute value. 
     * <p>
     * 
     * XDF attribute path grammar:
     * <pre>
     * &lt;attributePath&gt; ::= NULL | "" | &lt;addressableKeyName&gt; | &lt;repeatableSegment&gt; | &lt;addressableKeyName&gt; &lt;pathSeparator&gt; &lt;repeatableSegment&gt;
	 * &lt;repeatableSegment&gt; ::= &lt;attributeName&gt; | &lt;attributeName&gt; &lt;pathSeparator&gt; &lt;repeatableSegment&gt; 
     * &lt;attributeName&gt; ::= a valid field name on the POJO
     * &lt;addressableKeyName&gt; ::= "key" or the key identifier defined in {@link MapDataDescriptor#getAddressableKeyName()} 
     * &lt;pathSeparator&gt; ::= the path separator defined in {@link DataDescriptor#DEFAULT_PATHSEPARATOR}
     * </pre>
     * 
     * XDF attribute path examples:<br>
     * (assuming the default addressable key name of "key" and path separator of ".")
     * <ul>
     * <li><code>name</code> - Indexes the "name" attribute in the value object.
     * <li><code>key.ssn</code> - Indexes the "ssn" attribute in the key object.
     * <li><code>employee.store</code> - The "store" attribute from the "employee" embedded type in the value. 
     * </ul>
     * Note: method invocation is not supported when using Continuous Query with XDF maps.
	 *
     * Continuous query does not support entity maps.
     * 
     * @param attributePath 
     *              The path for retrieving a value for the object to be filtered.
     *              If the value object should be compared without any manipulation, use null.
     *              If the key object should be compared without any manipulation, use the addressible key name.
     * @param object The value to be compared with
     */
    public CompareFilter(String attributePath, MatchType object) {
        this.attributePath = attributePath;
        this.objectToCompare = object;
    }

    /**
     * Composes a string representation of the compare filter using the operator.
     * 
     * @param operator the symbol that represents the operation
     * @return "<attributePath> <operator> <object>"
     *          where 
     *          <attributePath> refers to the path for accessing a value from the object to be filtered
     *          <operator> refers to the input parameter,
     *          and <object> refers to the string representation of the object to compare the filtering value with
     */
    protected String createString(String operator) {
        StringBuffer rc = new StringBuffer();
        rc.append(attributePath == null ? "value" : attributePath).append(" ").append(operator).append(" ");
        if (objectToCompare != null && objectToCompare instanceof byte[]) {
            rc.append(Arrays.toString((byte[]) objectToCompare));
        } else {
            rc.append(objectToCompare.toString());
        }
        return rc.toString();
    }

}
