// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2006,2008
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  Query.java
//
// Source File Description:
//
//    See javadoc.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// b998         6.1       06/06/2006   cdjohnson New part.
// b1162        6.1       06/23/2006   cdjohnson Feature updates for M4
// b1576        6.1       08/11/2006   cdjohnson Feature updates for M5
// b2402        XD 6.1    11/27/2006   cdjohnson Feature updates for M7
// b10896       XD 6101   12/17/2007   cdjohnson Add setForUpdate
// b12664       XD 6103   06/24/2008   cdjohnson Add getPlan
// b12849       XD 6104   08/19/2008   cdjohnson Add setHint for index selection
// ---------------------------------------------------------------------------------------

package com.ibm.websphere.objectgrid.em;

import java.util.Iterator;

import com.ibm.websphere.objectgrid.ObjectMap;
import com.ibm.websphere.objectgrid.PartitionManager;
import com.ibm.websphere.projector.md.EntityMetadata;

/**
 * Interface used to control entity query execution.<p>
 *
 * Use the {@link EntityManager#createQuery(String)} method to create
 * a <code>Query</code>.  Each query instance can be used multiple times
 * using the EntityManager in which it was retrieved.<p>
 *
 * Each query result produces an entity, where the entity key is the row id (of type long)
 * and the entity value contains the field results of the SELECT clause.  Each query result
 * can be included in subsequent queries.<p>
 *
 * @since WAS XD 6.1
 * @see com.ibm.websphere.objectgrid.em.EntityManager
 */
public interface Query {

    /**
     * Hint to the query engine to use the specified index.<p>
     *
     * String Format:
     * <pre>
     * &lt;IndexHint&gt;     ::= &lt;Entity Name&gt; "." &lt;AttributeName&gt;
     * &lt;Entity Name&gt;   ::= ObjectGrid entity name
     * &lt;AttributeName&gt; ::= An indexed attribute.
     * </pre>
     *
     * Example:
     * <pre>
     * Query q = em.createQuery("SELECT p FROM Person p WHERE p.name=?1 AND p.city=?2 AND p.state=?3");
     * q.setHint(Query.HINT_USEINDEX, "Person.city");
     * q.setParameter(1, "Smith");
     * q.setParameter(2, "Rochester");
     * q.setParameter(3, "MN");
     * Iterator it = q.getResultIterator();
     * </pre>
     * @see #setHint(String, Object)
     * @since WAS XD 6.1.0.4
     */
    public String HINT_USEINDEX = "HINT_USEINDEX";

    /**
     * Execute a SELECT query and return the query results in an ObjectMap with the results in
     * query-specified order.  The result ObjectMap is only valid only for the current transaction.<p>
     *
     * The map key is the result number (of type long) starting at 0.<p>
     *
     * The map value is of type <code>{@link com.ibm.websphere.projector.Tuple}</code>
     * where each attribute and association is named based on it's ordinal
     * position within the query's select clause.  Use the <code>{@link ObjectMap#getEntityMetadata()}</code>
     * method to retrieve the <code>EntityMetadata</code> for the Tuple object stored within
     * the map.<p>
     *
     * This method is the fastest method for retrieving query result data where there can be multiple
     * results.  The name of the resulting entity can be retrieved using the
     * {@link ObjectMap#getEntityMetadata()} and {@link EntityMetadata#getName()} methods.<p>
     *
     * Example:<br>
     * The following query returns 2 rows:<br>
     * <pre>
     * String ql = SELECT e.name, e.id, d from Employee e join e.dept d WHERE d.number=5
     * Query q = em.createQuery(ql);
     * ObjectMap resultMap = q.getResultMap();
     * long rowID=0;
     * Tuple tResult = (Tuple) resultMap.get(new Long(rowID));
     * while(tResult != null) {
     *     // The first attribute is name and has an attribute name of "1"
     *     // But has an ordinal position of 0.
     *     String name = (String)tResult.getAttribute(0);
     *     Integer id = (String)tResult.getAttribute(1);
     *
     *     // Dept is an association with a name of "3", but
     *     // an ordinal position of 0 since it's the first association.
     *     // The association is always a OneToOne relationship,
     *     // so there is only one key.
     *     Tuple deptKey = tResult.getAssociation(0,0);
     *     rowID++;
     *     tResult = (Tuple) resultMap.get(new Long(rowID));
     *     ...
     * }
     * </pre>
     *
     * @return an ObjectMap that contains the results of the query.
     * @throws PersistenceException thrown if query validation fails.
     *             This exception may be deferred until the query is run.
     */
    public ObjectMap getResultMap();

    /**
     * Execute a SELECT query and return the query results using an Iterator where
     * each result is either an Object (for a single valued query) or Object[] (for
     * a multiple valued query).  The values in the Object[] result are stored in
     * query order.  The result Iterator is valid only for the current transaction.<p>
     *
     * This method is the preferred method for retrieving query results within the
     * EntityManager's context.  The optional {@link #setResultEntityName(String)} method
     * can be used to name the resulting entity so that it can be used in further queries.<p>
     *
     * Example:<br>
     * The following query returns 2 rows:<br>
     * <pre>
     *
     * String ql = SELECT e.name, e.id, e.dept from Employee e WHERE e.dept.number=5
     * Query q = em.createQuery(ql);
     * Iterator results = q.getResultIterator();
     * while(results.hasNext()) {
     *     Object[] curEmp = (Object[]) results.next();
     *     String name = (String) curEmp[0];
     *     Integer id = (Integer) curEmp[1];
     *     Dept d = (Dept) curEmp[2];
     *     ...
     * }
     * </pre>
     *
     * @return an Iterator with the query results in the form of <code>Object</code> or <code>Object[]</code>
     * @throws PersistenceException thrown if query validation fails.
     *             This exception may be deferred until the query is run.
     * @throws TransactionRequiredException
     *             if invoked without an active transaction.
     */
    public Iterator getResultIterator();


    /**
     * Execute a SELECT query and return the query results using an entity Iterator where
     * the entity type is determined by the <code>resultType</code> parameter.  The result
     * Iterator is valid only for the current transaction.<p>
     *
     * Use this method when it is desirable to use the EntityManager APIs to access the
     * resulting entities.
     *
     * Example:<br>
     * The following query returns all of the employees and the department they belong to
     * for one division, ordering by salary.  We want to print out the 5 employees with the
     * highest salaries and then select work with employees from only one department in the
     * same working set:<br>
     * <pre>
     * em.getTransaction().begin();
     *
     * String ql = SELECT e.name, e.id, e.dept from Employee e WHERE e.dept.division='Manufacturing' ORDER BY e.salary DESC
     * Query q = em.createQuery(ql);
     * q.setResultEntityName("AllEmployees");
     * Iterator results = q.getResultIterator(EmployeeResult.class);
     * int curEmployee=0;
     * while(results.hasNext() && curEmployee++ < 5) {
     *     EmployeeResult curEmp = (EmployeeResult) results.next();
     *     System.out.println(curEmp);
     *     // Remove the employee from the resultset.
     *     em.remove(curEmp);
     * }
     *
     * // Flush the changes to the result map.
     * em.flush();
     *
     * // Run a query against the local working set without the employees we removed
     * String ql = SELECT e.name, e.id, e.dept from AllEmployees e WHERE e.dept.name="Hardware""
     * Query q = em.createQuery(ql);
     * Iterator results = q.getResultIterator(EmployeeResult.class);
     * while(results.hasNext()) {
     *     EmployeeResult curEmp = (EmployeeResult) results.next();
     *     System.out.println(curEmp);
     * }
     *
     *
     * <literal>@Entity</literal>
     * public class EmployeeResult {
     *     String name;
     *     Integer id;
     *     <literal>@ManyToOne</literal>
     *     Dept dept;
     *
     *     public String toString() {
     *        return "Name=" + name + ", ID=" + id + ", Department=" + dept.name();
     *     }
     * }
     *
     * </pre>
     *
     * @param resultType the type of object to convert the results into.
     * @return an Iterator that contains the results of the query.
     * @throws PersistenceException thrown if query validation fails.
     *             This exception may be deferred until the query is run.
     * @throws TransactionRequiredException
     *             if invoked without an active transaction.
     */
    public Iterator getResultIterator(Class resultType);

    /**
     * Execute a SELECT query that returns a single result.<p>
     *
     * If the SELECT clause has more than one field defined, then the result
     * will be a <code>Object[]</code>
     * where each element in the object array is based on it's ordinal
     * position within the query's select clause.<p>
     *
     * <pre>
     * String ql = SELECT e from Employee e WHERE e.id=100"
     * Employee e = em.createQuery(ql).getSingleResult();
     *
     * String ql = SELECT e.name, e.dept from Employee e WHERE e.id=100"
     * Object[] empData = em.createQuery(ql).getSingleResult();
     * String empName= (String) empData[0];
     * Department empDept = (Department) empData[1];
     * </pre>
     *
     * @return the result Object or array of Objects
     * @throws NoResultException
     *             if there is no result
     * @throws NonUniqueResultException
     *             if more than one result
     * @throws PersistenceException thrown if query validation fails
     */
    public Object getSingleResult();

    /**
     * Set the maximum number of results to retrieve.
     *
     * @param maxResult
     *            the maximum number of results to retrieve.
     * @return the same query instance
     * @throws IllegalArgumentException
     *             if argument is negative
     */
    public Query setMaxResults(int maxResult);

    /**
     * Set the position of the first result to retrieve.
     *
     * @param startPosition position of the first result, numbered from 0
     * @return the same query instance
     * @throws IllegalArgumentException
     *             if argument is negative
     */
    public Query setFirstResult(int startPosition);

    /**
     * Specify the name of the query result entity.<p>
     *
     * Each time the <tt>getResultIterator</tt> or <tt>getResultMap</tt> methods
     * are invoked, an entity with an ObjectMap are dynamically created to hold the results
     * of the query.  If not specified, or
     * <tt>null</tt>, the entity and ObjectMap name will be automatically generated.<p>
     *
     * Since all query results are available for the duration of a transaction, a query
     * name may not be reused in a single transaction. <p>
     *
     * @param entityName the name of the entity
     * @return the same query instance
     */
    public Query setResultEntityName(String entityName);

    /**
     * Set a query hint. If the hint name is not recognized, it is silently ignored.
     *
     * @param hintName the name of the hint.  See the constant values defined in this interface.
     * @param value the value of the hint.
     * @return the same query instance
     * @throws IllegalArgumentException
     *             thrown if the second argument is not valid for the hint.
     *             This exception may be deferred until the query is run.
     * @since WAS XD 6.1.0.4
     */
    public Query setHint(String hintName, Object value);

    /**
     * Bind an argument to a named parameter.
     *
     * @param name
     *            the parameter name
     * @param value the value of the parameter.
     * @return the same query instance
     * @throws IllegalArgumentException
     *             if parameter name does not correspond to parameter in query string or argument is of incorrect type.
     *             This exception may be deferred until the query is run.
     */
    public Query setParameter(String name, Object value);

    /**
     * Bind an argument to a positional parameter.
     *
     * @param position the position of the parameter.  The first parameter is 1.
     * @param value the value of the parameter.
     * @return the same query instance
     * @throws IllegalArgumentException
     *             if position does not correspond to positional parameter of query or argument is of incorrect typel
     *             This exception may be deferred until the query is run.
     */
    public Query setParameter(int position, Object value);

    /**
     * Set the flush mode type to be used when the query is executed, overriding
     * the flush mode type set on the EntityManager.
     * @param flushMode the FlushModeType to set for this query instance.
     * @return the same query instance
     */
    public Query setFlushMode(FlushModeType flushMode);

    /**
     * Set the partition to route the query to.<p>
     *
     * Required if the maps in the query are partitioned and the EntityManager
     * does not have affinity to a single schema root entity's partition.<p>
     *
     * Use the {@link PartitionManager} to determine the number of partitions
     * for a given entity's backing map.
     *
     * @param partitionId the partition to route the query to.
     * @return the same query instance
     */
    public Query setPartition(int partitionId);

    /**
     * Set the update intent for the query.  If set to true, subsequent
     * query executions will lock records appropriately for updating.<p>
     *
     * If not set, the default is false.
     *
     * @param forUpdate if true, lock records for update.
     * @return the same query instance
     * @since WAS XD 6.1.0.1
     */
    public Query setForUpdate(boolean forUpdate);

    /**
     * Return a string representation of the query plan.<p>
     *
     * This method can be used to explain the plan of a query without running the query.  The plan describes
     * when an entity is to be scanned, an index used and the order in which the entities are accessed.
     *
     * @return a string representation of the query plan.
     * @throws PersistenceException thrown if query validation fails.  The transaction will not be marked to rollback.
     * @since WAS XD 6.1.0.3
     */
    public String getPlan();
}
