package com.ibm.websphere.objectgrid.io;
/*
 * IBM Confidential OCO Source Material
 * 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2011
 * The source code for this program is not published or otherwise divested
 * of its trade secrets, irrespective of what has been deposited with the
 * U.S. Copyright Office.
 *
 * Module  :  XsDataStreamFactory.java
 *
 * Change Activity:
 *
 * Reason       Version   Date         Userid    Change Description
 * ------------ --------- ------------ --------- -----------------------------------------
 * b24713       7.1.1     9/15/2011    ppoddar    Created
 */

import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

/**
 * A random access, byte-oriented, portable stream for reading primitive types and string.
 * <br>
 * This stream combines the facilities of
 * <UL>
 * <LI>primitive data reading of {@link java.io.DataOutput} plus unsigned values
 * <LI>random access capability of a {@link java.nio.ByteBuffer}.
 * </UL>
 * <p>
 *
 * This class is not intended as a base class for user code to extend.
 * Instances of XsDataInputStream are provided solely by the WebSphere eXtreme Scale
 * runtime for use by plug-ins.
 *
 * @since 7.1.1
 */
public abstract class XsDataInputStream extends InputStream {
	/**
	 * Positions this stream to read from the given position.
	 * @param i a position in this stream
	 * @return this stream positioned at the given position.
	 * @throws IOException if the position is less than 0 or more than the size of this
	 * stream
	 */
    public abstract XsDataInputStream position(int i) throws IOException;

    /**
     * Gets the current position from where the next byte will be read.
     */
    public abstract int position() throws IOException;

    /**
     * Gets the total number of bytes in this stream.
     */
    public abstract int size() throws IOException;

    /**
     * Affirms if this stream has been closed.
     */
    public abstract boolean isClosed();

	/**
	 * Reads 1-byte as a boolean value at the current position in current byte order.
	 */
    public abstract boolean readBoolean() throws IOException;

	/**
	 * Reads 1-byte as a signed byte value at the current position in current byte order.
	 */
    public abstract byte readByte() throws IOException;

	/**
	 * Reads 1-byte as an unsigned byte value at the current position in current byte order.
	 */
    public abstract int readUnsignedByte() throws IOException;

    /**
     * Reads <tt>b.length</tt> bytes, if available.
     * If requisite bytes are not available, an IOException is thrown.
     * If the given byte array is null, throws a NullPointerException.
     *
     * @param b array of bytes to be populated. Must not be null.
     * @param off an offset index. Must be positive.
     * @param len number of bytes to read. Must be positive.
     */
    public abstract void readFully(byte[] b, int off, int len) throws IOException;

    /**
     * Reads <tt>b.length</tt> bytes, if available.
     * If requisite bytes are not available, an IOException is thrown.
     * If the given byte array is null, throws a NullPointerException.
     */
    public abstract void readFully(byte[] b) throws IOException;

	/**
	 * Reads 2-byte as an signed integer value in the current position in current byte order.
	 * The 2-byte integer value is interpreted as a 2-byte <tt>UTF-16</tt> character.
	 */
    public abstract char readChar() throws IOException;

    /**
     * Reads <tt>c.length</tt> characters as an array of 2-byte signed integer values in
     * the current position in current byte order.  The 2-byte integer values are interpreted
     *  as 2-byte <tt>UTF-16</tt> characters.
     */
    public abstract void readChars(char[] c) throws IOException;

	/**
	 * Reads a 4-byte integer at the current position in current byte order.
	 */
    public abstract int  readInt() throws IOException;

	/**
	 * Reads a 4-byte floating point number at the current position in current byte order.
	 */
    public abstract float readFloat() throws IOException;

	/**
	 * Reads a 8-byte floating point number at the current position in current byte order.
	 */
    public abstract double readDouble() throws IOException;

	/**
	 * Reads a 8-byte integer number at the current position in current byte order.
	 */
    public abstract long readLong() throws IOException;

	/**
	 * Reads a 2-byte signed integer number at the current position in current byte order.
	 */
    public abstract short readShort() throws IOException;

	/**
	 * Reads a 2-byte unsigned integer number at the current position in current byte order.
	 */
    public abstract int readUnsignedShort() throws IOException;

    /**
     * Reads a variable length integer number at the current position in the current byte order.
     */
    public abstract int readVarintAsInt() throws IOException;

    /**
     * Reads a variable length long number at the current position in the current byte order.
     */
    public abstract long readVarintAsLong() throws IOException;


    /**
     * Gets canonical name of the current string encoding scheme.
     * By default, the encoding scheme is <tt>UTF-8</tt>.
     */
    public abstract String getEncodingScheme();

    /**
     * Sets the current string encoding scheme.
     * @param enc canonical name of an encoding scheme.
     */
    public abstract void setEncodingScheme(String enc);

	/**
	 * Sets the byte order of the data.
	 *
	 * @param order <tt>BIG_ENDIAN</tt> or <tt>LITTLE_ENDIAN</tt> byte order
	 * @return if the given order is different than the native order of the platform.
	 */
	public abstract boolean setByteOrder(ByteOrder order);

	/**
	 * Gets the byte order for writing multi-byte data.
	 *
	 * @return <tt>BIG_ENDIAN</tt> or <tt>LITTLE_ENDIAN</tt> byte order
	 */
	public abstract ByteOrder getByteOrder();

	/**
	 * Reads a string assuming that the number of bytes of the string to be read
	 * precedes the actual bytes as a 2-byte short number. The bytes are read
	 * and decoded with the current encoding scheme.
	 * <br>
	 * This method emulates the semantics of equivalent {@link DataInputStream#readUTF() method}
	 * of <tt>readUTF</tt>.
	 *
	 * @return the string being read
	 * @throws IOException
	 */
    public abstract String readString() throws IOException;

	/**
	 * Reads a string assuming that the number of bytes of the string to be read
	 * precedes the actual bytes as a 2-byte short number. The bytes are read
	 * and decoded with the given encoding scheme.
	 *
	 * @return the string being read
	 * @throws IOException
	 */
    public abstract String readString(String encoding) throws IOException;

	/**
	 * Reads a string assuming that the string to be read is terminated by a single bye of value 0.
	 * The bytes are read and decoded with the current encoding scheme.
	 *
	 * @return the string being read
	 * @throws IOException
	 */
	public abstract String readStringAsNullTerminated() throws IOException;

	/**
	 * Reads a string assuming that the string to be read is terminated by a single bye of value 0.
	 * The bytes are read and decoded with the given encoding scheme.
	 *
	 * @return the string being read
	 * @throws IOException
	 */
	public abstract String readStringAsNullTerminated(String encoding) throws IOException;


	/**
	 * Reads a string as a array of bytes of given length.
	 * The bytes are read and decoded with the current encoding scheme.
	 *
	 * @return the string being read
	 * @throws IOException
	 */
	public abstract String readStringRaw(int L) throws IOException ;

	/**
	 * Reads a string as a array of bytes of given length.
	 * The bytes are read and decoded with the given encoding scheme.
	 *
	 * @return the string being read
	 * @throws IOException
	 */
	public abstract String readStringRaw(String encoding, int L) throws IOException;

	/**
	 * Reads a string as a array of bytes with given parameter specifications.
	 *
	 * The bytes are read and decoded with the given encoding scheme.
	 *
	 * @param lengthPrepended if true, then a 2-byte short is read from the current position
	 * of the stream itself. The value read is assumed to be the number of bytes in the string.
	 * The next two parameters are ignored.
	 * @param L if the <tt>lengthPrepended</tt> and <tt>nullTerminated</tt> parameter are both false,
	 * then this value is considered to be the number of bytes in the string.
	 * @param nullTerminated if true and the <tt>lengthPrepended</tt> is false, then scans the stream from the
	 * current position to find a byte of value 0 as the terminating character of the string to be read.
	 * @return the string being read
	 * @throws IOException
	 */
	public abstract String readString(String encoding, boolean lengthPrepended, int L, boolean nullTerminated)
	throws IOException;


	/**
	 * Reads a string as an array of bytes with the UTF-8 encoding
	 *
	 * The length is stored in a varint at the current position in the stream to indicate the number
	 * of UTF-8 bytes that will follow in the buffer.
	 *
	 * @return the string being read
	 * @throws IOException
	 */
	public abstract String readUTF8String( ) throws IOException;


	/**
	 * Rewinds the stream by reopening, if necessary and repositioning at the state of construction.
	 * @return this stream positioned at the beginning of the stream.
	 */
	public abstract XsDataInputStream rewind();

	/**
	 * Gets the underlying bytes of this stream.
	 */
	public abstract byte[] toByteArray();

	/**
	 * Transfers the underlying bytes of this stream to the given byte buffer at current position.
	 * @param buffer a non-null buffer
	 * @return the buffer after the bytes are written.
	 */
	public abstract ByteBuffer toByteBuffer(ByteBuffer buffer);
}
