package com.ibm.websphere.objectgrid.io;
/*
 * IBM Confidential OCO Source Material
 * 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2011
 * The source code for this program is not published or otherwise divested
 * of its trade secrets, irrespective of what has been deposited with the
 * U.S. Copyright Office.
 *
 * Module  :  XsDataStreamFactory.java
 *
 * Change Activity:
 *
 * Reason       Version   Date         Userid    Change Description
 * ------------ --------- ------------ --------- -----------------------------------------
 * b24713       7.1.1     9/15/2011    ppoddar    Created
 */

import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

/**
 * An abstract extension of {@link OutputStream standard i/o stream} for writing primitive types and string
 * in random accessible, byte-oriented, cross-language manner specifically for internal usage of 
 * WebSphere eXtreme Scale plug-ins.
 * <br>
 * This stream combines the facilities of
 * <UL>
 * <LI>sequential, on-demand resource allocation of {@link java.io.ByteArrayOutputStream}
 * <LI>primitive data writing of {@link java.io.DataOutput}
 * <LI>random access capability of a {@link java.nio.ByteBuffer}.
 * </UL>
 * <p>
 * <br>
 * <b>Resource Allocation</b>: The stream must grow on demand as the bytes
 * are being written. The actual bytes being written <em>should</em> be allocated in a
 * a flexible manner that allows both in-memory and off-heap allocation.
 * <p>
 * The written bytes are accessible via {@link #toByteArray()} methods as a single,
 * contiguous array of bytes. Or can be transferred to a given {@linkplain ByteBuffer}
 * via {@linkplain #toByteBuffer(ByteBuffer)} method.
 * <p>
 * <b>Random Access</b>: The data is written at the current {@link #position()}.
 * The caller can, however, {@link #position(int) position} the stream at any
 * location. If the position is beyond the current {@link #size() size} of the
 * stream, new bytes must be allocated.
 * <p>
 * <p><b>Data Type Support</b>: The standard fixed-length, signed, primitive Java data types
 * are supported as in {@link java.io.DataOutput} interface. However, for cross-language data
 * portability, this Java based interface also supports unsigned value of Java primitive types via
 * wider signed type value. For example, an unsigned 2-byte short could be
 * written by {@link #writeUnsignedShort(int)} supplying a 4-byte integer value as input.
 * <p>
 * <b>String Encoding</b>: Strings are written, by default, as encoded (by UTF-8 scheme)
 * array of bytes preceded by a 2-byte integral number designating number of encoded bytes.
 * <br>
 * The encoding scheme can be specified per string basis or default scheme can be changed via
 * {@link #setEncodingScheme(String)} method. The default encoding
 * scheme at construction is <tt>UTF-8</tt>.
 * <br>
 * Optionally, raw encoded bytes could be written without the length header.
 * <br>
 * Optionally, a single-byte null character could be appended after the encoded bytes.
 * <br>
 * <p>
 * <b>Byte Arrays</b>: byte arrays are written directly to the output stream with
 * no conversion. Attempting to write a null byte array will result in a NullPointerException.
 * <p>
 * <b>Null Handling</b>: Null Strings or array can not be written and will throw
 * {@linkplain NullPointerException} when supplied as input.
 * <p>
 * <b>Byte Ordering</b>: Multi-byte data can be written in either
 * <tt>BIG ENDIAN</tt> or <tt>LITTLE ENDIAN</tt> form. By default, the data is
 * written in the {@link ByteOrder#BIG_ENDIAN <tt>BIG_ENDIAN</tt> order}.
 * The user can {@linkplain #setByteOrder(ByteOrder) set} the default
 * byte order.
 * <p>
 * This class is not intended as a base class for user code to extend.
 * Instances of XsDataInputStream are provided solely by the WebSphere eXtreme Scale
 * runtime for use by plug-ins.
 *
 * @since 7.1.1
 */
public abstract class XsDataOutputStream extends OutputStream {
	/**
	 * Gets the current size of this stream in number of bytes.
	 * Size of the stream denotes the maximum position where a byte is ever written.
	 */
    public abstract int size();

    /**
     * Gets the current position of this stream.
     * The current position denotes where the next byte will be written.
     */
    public abstract int position();

    /**
     * Sets the position of this stream at the given byte index.
     * The position denotes where the next byte will be written.
     */
    public abstract XsDataOutputStream position(int i) throws IOException;

    /**
     * Restores this stream at the initial state at its construction.
     */
    public abstract XsDataOutputStream rewind() throws IOException;

    /**
     * Gets a <em>copy</em> of the underlying bytes.
     *
     * @return an array of bytes. Length of the array is equal to {@linkplain #size() current size}.
     */
    public abstract byte[] toByteArray();

    /**
     * Writes the content of this stream on to the given buffer starting at its current position.
     * @param buf a buffer where the content of this stream will be written.
     * @return the given buffer with modified content.
     * @throws IOException
     */
    public abstract ByteBuffer toByteBuffer(ByteBuffer buf) throws IOException;


    /**
     * Writes a single byte for the given boolean value.
     */
    public abstract void writeBoolean(boolean bool) throws IOException;

    /**
     * Writes a single byte for the signed value of the given byte.
     */
    public abstract void writeByte(byte b) throws IOException;

    /**
     * Writes a single byte as the unsigned byte value (i.e. least significant byte)
     * of the given integer.
     */
    public abstract void writeUnsignedByte(int b) throws IOException;

    /**
     * Writes a 2-byte value for the given character in the current byte order.
     */
    public abstract void writeChar(char c) throws IOException;

    /**
     * Writes an array of 2-byte values for the given character in the current byte order.
     */
    public abstract void writeChars(char c[]) throws IOException;
    
    /**
     * Writes a 4-byte value for the given integral number in the current byte order.
     */
    public abstract void  writeInt(int i) throws IOException;

    /**
     * Writes a 4-byte value for the given floating point number in the current byte order.
     */
    public abstract void writeFloat(float f) throws IOException;

    /**
     * Writes a 8-byte value for the given floating point number in the current byte order.
     */
    public abstract void writeDouble(double d) throws IOException;

    /**
     * Writes a 8-byte value for the given integral number in the current byte order.
     */
    public abstract void writeLong(long l) throws IOException;

    /**
     * Writes a 2-byte value for the given integral number in the current byte order.
     */
    public abstract void writeShort(short s) throws IOException;

    /**
     * Writes 2-byte as the unsigned short value (i.e. 2 least significant bytes)
     * of the given integer in the current byte order.
     */
    public abstract void writeUnsignedShort(int s) throws IOException;

    /**
     * Writes a variable length integer value for the given integral number.
     */
    public abstract void writeVarint( int i) throws IOException;

    /**
     * Writes a variable length integer value for the given integral number.
     */
    public abstract void writeVarint( long i) throws IOException;
    
    
    /**
     * Writes the given string as array of bytes encoded in current encoding scheme.
     * The number of encoded bytes is written as a 2-byte integral value before the actual bytes.
     * @param s a string. Must not be null.
     * @return total number of bytes written to the stream
     * @throws IOException
     * @throws NullPointerException is the given string is null.
     * @throws IllegalArgumentException when the byte length cannot be encoded in a 2-byte integral value
     */
    public abstract int writeString(String s) throws IOException;

    /**
     * Writes the given string as array of bytes encoded by given encoding scheme.
     * The number of encoded bytes is written as a 2-byte integral value before the actual bytes.
     * @param s a string. Must not be null.
     * @param encoding canonical name of a supported encoding scheme
     * @return total number of bytes written to the stream
     * @throws IOException
     * @throws NullPointerException is the given string is null.
     * @throws UnsupportedEncodingException
     * @throws IllegalArgumentException when the byte length cannot be encoded in a 2-byte integral value
     */
    public abstract int writeString(String s, String encoding) throws IOException, UnsupportedEncodingException;

    /**
     * Writes the given string as array of bytes encoded by current encoding scheme.
     * The number of bytes is <em>not</em> written before the actual bytes. But
     * a single byte null character is written at the end.
     * @param s a string. Must not be null.
     * @return total number of bytes written to the stream
     * @throws IOException
     */
    public abstract int writeStringAsNullTerminated(String s) throws IOException;

    /**
     * Writes the given string as array of bytes encoded by the given encoding scheme.
     * The number of bytes is <em>not</em> written before the actual bytes. But
     * a single byte null character is written at the end.
     * @param s a string. Must not be null.
     * @param encoding canonical name of a supported encoding scheme
     * @return total number of bytes written to the stream
     * @throws IOException
     */
    public abstract int writeStringAsNullTerminated(String s, String encoding) throws IOException;

    /**
     * Writes the given string as array of bytes encoded by the current encoding scheme.
     * The number of bytes is <em>not</em> written before the actual bytes.
     * @param s a string. Must not be null.
     * @return total number of bytes written to the stream
     * @throws IOException
     */
    public abstract int writeStringRaw(String s) throws IOException;

    /**
     * Writes the given string as array of bytes encoded by the given encoding scheme.
     * The number of encoded bytes is <em>not</em> written before the actual bytes.
     * @param s a string. Must not be null.
     * @param encoding canonical name of a supported encoding scheme
     * @return total number of bytes written to the stream
     * @throws IOException
     */
    public abstract int writeStringRaw(String s, String encoding) throws IOException;

    /**
     * Writes the given string as array of bytes encoded by the given encoding scheme.
     * @param s a string. Must not be null.
     * @param encoding canonical name of a supported encoding scheme
     * @param raw if true, 2-byte length header is not written before the bytes.
     * @param nullTerminate if true a single-byte null character is written at end of encoded bytes.
     * @return total number of bytes written to the stream
     * @throws IOException
     * @throws IllegalArgumentException when raw is false, and the byte length cannot be encoded in a 2-byte integral value
     */
    public abstract int writeString(String s, String encoding, boolean raw, boolean nullTerminate) throws IOException;

    /**
     * Writes the given string as an array of bytes encoded as UTF-8.  This method always 
     * prepends the string with a varint value to indicate the number of UTF-8 bytes that 
     * are encoded.
     * @param s a string. Must not be null
     * @return total number of bytes written to the stream
     * @throws IOException
     */
    public abstract int writeUTF8String(String s) throws IOException;
    
    
    /**
     * Gets canonical name of the current string encoding scheme.
     * By default, the encoding scheme is <tt>UTF-8</tt>.
     */
    public abstract String getEncodingScheme();

    /**
     * Sets the current string encoding scheme.
     * @param enc canonical name of an encoding scheme.
     */
    public abstract void setEncodingScheme(String enc);

	/**
	 * Sets the byte order of the data.
	 *
	 * @param order <tt>BIG_ENDIAN</tt> or <tt>LITTLE_ENDIAN</tt> byte order
	 * @return if the given order is different than the native order of the platform.
	 */
	public abstract boolean setByteOrder(ByteOrder order);

	/**
	 * Gets the byte order for writing multi-byte data.
	 *
	 * @return <tt>BIG_ENDIAN</tt> or <tt>LITTLE_ENDIAN</tt> byte order
	 */
	public abstract ByteOrder getByteOrder();
	
	/**
	 * Affirms if this stream has been closed.
	 */
	public abstract boolean isClosed();
}
