// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008, 2009
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Source File Description:
//
//  Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// d11207       WASX.XD   03/04/2008   bnewport  OpenJPA Load and Preload
// d12005       6.1.0.3   05/15/2008   jian.tang Enable preload map
// b12562       6.1.0.3   06/19/2008   jian.tang Support ManyToOne as Id
// b14919       6.1.0.5   02/16/2009   jian.tang Throw OptimisticCollisionException at commit time
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.jpa;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.BackingMap;
import com.ibm.websphere.objectgrid.ObjectGrid;
import com.ibm.websphere.objectgrid.OutputFormat;
import com.ibm.websphere.objectgrid.TxID;
import com.ibm.websphere.objectgrid.em.EntityManager;
import com.ibm.websphere.objectgrid.plugins.Loader;
import com.ibm.websphere.objectgrid.plugins.LoaderException;
import com.ibm.websphere.objectgrid.plugins.LogElement;
import com.ibm.websphere.objectgrid.plugins.LogSequence;
import com.ibm.websphere.objectgrid.plugins.OptimisticCollisionException;
import com.ibm.websphere.objectgrid.plugins.io.annotations.PluginOutputFormat;
import com.ibm.websphere.projector.Tuple;
import com.ibm.websphere.projector.md.EntityMetadata;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.ObjectGridManagerImpl;
import com.ibm.ws.objectgrid.jpa.JPAHelper;
import com.ibm.ws.objectgrid.jpa.JPAProviderHelperImpl;
import com.ibm.ws.projector.TupleBuilder;
import com.ibm.ws.projector.md.OverrideAccessType;

/**
 * This is the entity loader that uses Java Persistence API (JPA) to load and
 * store data in the "database". Therefore, a JPA provider is required to use this feature.
 * The "database" here can be any backend supported by the chosen JPA provider.
 * <p>
 * A TransactionCallback plugin, {@link JPATxCallback},
 * should be configured on the ObjectGrid to provide the transaction
 * coordination between the ObjectGrid transaction and the JPA entity manager
 * transaction for this loader to work.
 * <p>
 * A JPA persistence unit name and entity class name are required to use JPA. The
 * persistence unit name is provided to the JPATxCallback instance and the entity
 * class name is provided to this loader. This establishes a one to one relation
 * between  ObjectGrid instances and JPA persistence units and a one to one
 * relation between ObjectGrid maps and JPA entities.
 * <p>
 * When defining the ObjectGrid entity metadata for the entity map, an ObjectGrid entity class
 * has already been defined. If the JPA entity class is the same as this ObjectGrid entity class,
 * there is no need to provide the entity class any more. If the JPA entity class
 * is different, it is required to be set using the attribute "entityClassName".
 * <p>
 * This loader is used for ObjectGrid entity maps only.
 * Use {@link JPALoader}
 * for ObjectGrid maps that store POJOs.
 * <p>
 * If entities have relationships to other entities, each entity map must have an independent JPAEntityLoader.
 * The JPAEntityLoader cannot be used to load the root entity and its children.
 * <p>
 * If the preloadPartition property is set to a valid partition, the preloadMap method will preload all the data
 * into the ObjectGrid maps from that partition. The preloadMap uses {@link ClientLoader}
 * to preload a map. By default, the preloadPartition is set to -1 to disable the map preload.
 * <p>
 * This JPAEntityLoader implements the RetryableLoader because it can handle a retry record
 * from ObjectGrid.
 * When the retry indication slot contains a <code>Boolean.TRUE<code>, the JPAEntityLoader handles the
 * ObjectGrid log elements in the following ways:
 * <OL>
 * <LI>For an insert or update type log element, it calls {@link EntityManager#merge(Object)} method to merge the entity. </LI>
 * <LI>For a delete type log element, if calls {@link EntityManager#find(Class, Object)} method to find the entity. If the
 * entity is found, it calls {@link EntityManager#remove(Object)} method to remove it.</LI>
 * </OL>
 * @see com.ibm.websphere.objectgrid.jpa.ClientLoader
 * @since WAS XD 6.1.0.3
 */
@PluginOutputFormat(keyFormat=OutputFormat.RAW, valueFormat=OutputFormat.RAW)
public class JPAEntityLoader extends JPALoader {
    private final static String CLASS_NAME = JPAEntityLoader.class.getName();

    private final static TraceComponent tc = Tr.register(CLASS_NAME, Constants.TR_JPA_GROUP_NAME,
            Constants.TR_RESOURCE_BUNDLE_NAME);

    private EntityMetadata emd;

//    IdClassMetadata[] jpaIdClassMD = null;

    /**
     * Write the data from ObjectGrid entity maps to JPA.
     * <p>
     * The LogSequence contains a sequence of log elements which contain entity key and value tuples.
     * These tuples are converted in to JPA POJO entities, which are then used to interact with JPA to
     * insert, update, or remove the JPA entities.
     *
     * @see Loader#batchUpdate(TxID, LogSequence)

     */
    @Override
    public void batchUpdate(TxID txid, LogSequence sequence) throws LoaderException, OptimisticCollisionException {
        if (retryIndicationSlot != -1) {
            Boolean retry = (Boolean) txid.getSlot(retryIndicationSlot);
            if (retry != null && retry.booleanValue()) {
                entityBatchUpdate(txid, sequence, true);
            } else {
                entityBatchUpdate(txid, sequence, false);
            }
        } else {
            entityBatchUpdate(txid, sequence, false);
        }

    }

    private void entityBatchUpdate(TxID tx, LogSequence seq, boolean isRetry) throws LoaderException, OptimisticCollisionException {
        // this is a JPA EntityManager
        javax.persistence.EntityManager em = null;

        EntityManager ogEM = tx.getSession().getEntityManager();

        Object key = null;
        try {
            em = txCallback.getEntityManager(tx);
            
            Iterator<LogElement> changes = seq.getPendingChanges();
            Object value = null;
            while (changes.hasNext()) {
                LogElement elem = changes.next();
                
                if (elem.isCascaded()) {
                    continue;
                }

                key = elem.getKey();

                switch (elem.getType().getCode()) {
                case LogElement.CODE_INSERT:
                    value = ogEM.find(emd.getMetadataClass(), key);
                    if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
                        Tr.debug(tc, this.mapName + ": persist (retry=" + isRetry + ") " + value);
                    if (isRetry) {
                        em.merge(value);
                    } else {
                        em.persist(value);
                    }
                    break;
                case LogElement.CODE_UPDATE:
                case LogElement.CODE_UPSERT:
                    value = ogEM.find(emd.getMetadataClass(), key);
                    if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
                        Tr.debug(tc, this.mapName + ": merge (retry=" + isRetry + ") " + value);
                    Object o = em.merge(value);
                    break;
                case LogElement.CODE_DELETE:
                    value = em.find(emd.getMetadataClass(), JPAHelper.convertKeyTupleToJPAKey(txCallback
                            .getPersistenceUnitName(), ogEM, em, (Tuple) key, keyMd));

                    if (value != null) {
                        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
                            Tr.debug(tc, this.mapName + ": remove (retry=" + isRetry + ") " + value);

                        // This try catch block is used to catch the OptimisticLockException, we have to catch 
                        // it here 
                        em.remove(value);
                    }
                    
                    break;
                }
            }
            // We cannot do flush here, otherwise it will complain about unmanaged object if cascade
            // is not turned on for the relations
            //            em.flush();
            
        } catch (javax.persistence.OptimisticLockException ole) {
            OptimisticCollisionException oce = jpaProviderHelper.createOptimisticCollisionException(
                    "An OptimisticLockException is thrown in JPAEntityLoader of ObjectGrid " + ogName
                            + " map " + mapName + ": " + ole.getMessage(), this.ogName, this.mapName, key,
                    ole);
            throw oce;
        } catch (LoaderException le) {
            throw le;
        } catch (Error e) {
            throw e;
        } catch (Exception e) {
            Throwable t = e;
            if (jpaProviderHelper.isOptimisticLockException(e)) {
                OptimisticCollisionException oce = jpaProviderHelper.createOptimisticCollisionException(
                        "An OptimisticLockException is thrown in JPAEntityLoader of ObjectGrid " + ogName
                                + " map " + mapName + ": " + e.getMessage(), this.ogName, this.mapName, key,
                        e);
                throw oce;
            }  else if (exceptionMapper != null) {
                 t = exceptionMapper.map(e);
            }
            
            if (t instanceof LoaderException) {
                throw (LoaderException) t;
            }
            
            LoaderException le = new LoaderException("Exception caught in ObjectGrid " + ogName + " map " + mapName
                    + ": " + t.getMessage());
            le.initCause(t);
            throw le;
        }
    }

    /**
     * Returns the list of values corresponding to the set of keys passed in by
     * retrieving the data using JPA.
     * <p>
     * The data retrieved from JPA are POJO entities, which are then converted into
     * ObjectGrid entity tuples. Each key provided will have a corresponding entity
     * tuple in the returned list.
     *
     * @see Loader#get(TxID, List, boolean)
     */
    @Override
    public List get(TxID tx, List keys, boolean forUpdate) throws LoaderException {
        // this is a JPA EntityManager
        javax.persistence.EntityManager em = null; 
        List<Object> result = new LinkedList<Object>();

        try {
            em = txCallback.getEntityManager(tx);
            
            Iterator<Object> iter = keys.iterator();
            while (iter.hasNext()) {
                Tuple key = null;
                try {
                    key = (Tuple) iter.next();
                } catch (ClassCastException cce) {
                    LoaderException le = new LoaderException(
                            "ClassCastException caught in ObjectGrid "
                                    + ogName
                                    + " map "
                                    + mapName
                                    + ". Exception message is "
                                    + cce.getMessage()
                                    + ". One possible cause is that the application uses ObjectMap.get API to get the entity by passing a non-tuple key. Use a tuple key or use EntityManager.find API to get the entity.");
                    le.initCause(cce);
                    throw le;
                }
                if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
                    Tr.debug(tc, this.mapName + ": get " + key);
                Object v = em.find(entityClass, JPAHelper.convertKeyTupleToJPAKey(txCallback.getPersistenceUnitName(),
                        tx.getSession().getEntityManager(), em, key, keyMd));
                if (v == null) {
                    v = Loader.KEY_NOT_FOUND;
                    result.add(v);
                } else {

                    // We will project the whole graph.
                    // That way, we will have the relation keys associated with the POJO.
                    result.add(TupleBuilder.getTupleFromEntity(v, emd, false, partitionId, OverrideAccessType.PROPERTY));
                }
            }

            // Should not Clear the entity manager to detach entities.
            // Otherwise we might hit NPE from OpenJPA
            // em.clear();
        } catch (LoaderException e) {
            throw e;
        } catch (Error e) {
            throw e;
        } catch (Throwable e) {
            if (exceptionMapper != null) {
                e = exceptionMapper.map(e);
            }
            
            if (e instanceof LoaderException) {
                throw (LoaderException) e;
            } 
            
            // Wrap the exception in LoaderException
            LoaderException le = new LoaderException("Exception caught in ObjectGrid " + ogName + " map " + mapName
                    + ". Exception message is " + e.getMessage());
            le.initCause(e);
            throw le;
        }
        return result;
    }


    @Override
    protected void initialize(ObjectGrid grid, BackingMap bmap) throws LoaderException {
        if (!initialized) {

            getJPATxCallback(grid, bmap);

            if (entityClassName == null) {
                entityClass = bmap.getEntityMetadata().getMetadataClass();
                entityClassName = entityClass.getName();
            }

            emd = bmap.getEntityMetadata();

            if (emd == null) {
                throw new LoaderException(
                        "A com.ibm.websphere.objectgrid.jpa.JPAEntityLoader is "
                                + "configured for ObjectGrid "
                                + ogName
                                + " map "
                                + mapName
                                + ") which does not have EntityMetadata associated. Use com.ibm.websphere.objectgrid.jpa.JPALoader instead.");
            }

            // Get the persistence unit name.
            String puName = txCallback.getPersistenceUnitName();

            keyMd = JPAHelper.getJPAMetadata(puName, entityClass, emd);

            jpaProviderHelper = new JPAProviderHelperImpl(keyMd);
            
            exceptionMapper = txCallback.getExceptionMapper();

            initialized = true;
        }
    }

}
