// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008, 2009
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Source File Description:
//
//  Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// d11207       6.1.0.3   03/04/2008   bnewport  OpenJPA Load and Preload
// d12005       6.1.0.3   05/15/2008   jian.tang Enable preload map
// d12175       6.1.0.3   05/19/2008   jian.tang add persistencePropMap to ClientLoader.load method
// b12562       6.1.0.3   06/19/2008   jian.tang Support ManyToOne as Id
// b15138       6.1.0.5   02/24/2009   jian.tang Throw IllegalArgumentException when the entityClassName is not defined for JPALoader
// b14919       6.1.0.5   02/16/2009   jian.tang Throw OptimisticCollisionException at commit time
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.jpa;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.persistence.EntityManager;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.BackingMap;
import com.ibm.websphere.objectgrid.ObjectGrid;
import com.ibm.websphere.objectgrid.OutputFormat;
import com.ibm.websphere.objectgrid.Session;
import com.ibm.websphere.objectgrid.TxID;
import com.ibm.websphere.objectgrid.plugins.ExceptionMapper;
import com.ibm.websphere.objectgrid.plugins.Loader;
import com.ibm.websphere.objectgrid.plugins.LoaderException;
import com.ibm.websphere.objectgrid.plugins.LogElement;
import com.ibm.websphere.objectgrid.plugins.LogSequence;
import com.ibm.websphere.objectgrid.plugins.ObjectTransformer;
import com.ibm.websphere.objectgrid.plugins.OptimisticCollisionException;
import com.ibm.websphere.objectgrid.plugins.RetryableLoader;
import com.ibm.websphere.objectgrid.plugins.TransactionCallback;
import com.ibm.websphere.objectgrid.plugins.io.KeySerializerPlugin;
import com.ibm.websphere.objectgrid.plugins.io.MapSerializerPlugin;
import com.ibm.websphere.objectgrid.plugins.io.SerializerAccessor;
import com.ibm.websphere.objectgrid.plugins.io.ValueSerializerPlugin;
import com.ibm.websphere.objectgrid.plugins.io.annotations.PluginOutputFormat;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.SerializedKey;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.SerializedValue;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.ObjectGridManagerImpl;
import com.ibm.ws.objectgrid.jpa.ClientCallbackImpl;
import com.ibm.ws.objectgrid.jpa.JPAHelper;
import com.ibm.ws.objectgrid.jpa.JPAKeyMetadata;
import com.ibm.ws.objectgrid.jpa.JPAProviderHelperImpl;
import com.ibm.ws.objectgrid.plugins.io.dataobject.keys.KeyDataBytesImpl;
import com.ibm.ws.xs.NLSConstants;
import com.ibm.ws.xs.util.dopriv.DoPrivUtil;

/**
 * This is the loader that uses Java Persistence API (JPA) to load and
 * store data in the "database". Therefore, a JPA provider is required to use this feature.
 * The "database" here can be any backend supported by the chosen JPA provider.
 * <p>
 * A TransactionCallback plugin, {@link JPATxCallback},
 * should be configured on the ObjectGrid to provide the transaction
 * coordination between the ObjectGrid transaction and the JPA entity manager
 * transaction for this loader to work. Otherwise, a LoaderException will be thrown.
 * <p>
 * A JPA persistence unit name and entity class name are required to use JPA. The
 * persistence unit name is provided to the JPATxCallback instance and the entity
 * class name is provided to this loader. This establishes a one to one relation
 * between  ObjectGrid instances and JPA persistence units and a one to one
 * relation between ObjectGrid maps and JPA entities.
 * <p>
 * This loader is used for ObjectGrid maps that store POJO only.
 * Use {@link JPAEntityLoader}
 * for ObjectGrid maps that store ObjectGrid entities.
 * <p>
 * If the preloadPartition property is set to a valid partition, the preloadMap method will preload all the data
 * into the ObjectGrid maps from that partition. The preloadMap uses {@link ClientLoader}
 * to preload a map. By default, the preloadPartition is set to -1 to disable the map preload.
 * <p>
 * This JPALoader implements the RetryableLoader because it can handle a retry record from ObjectGrid.
 * When the retry indication slot contains a <code>Boolean.TRUE</code>, the JPALoader handles the
 * ObjectGrid log elements in the following ways:
 * <OL>
 * <LI>For an insert or update type log element, it calls {@link EntityManager#merge(Object)} method to merge the entity. </LI>
 * <LI>For a delete type log element, if calls {@link EntityManager#find(Class, Object)} method to find the entity. If the
 * entity is found, it calls {@link EntityManager#remove(Object)} method to remove it.</LI>
 * </OL>
 *
 * @see com.ibm.websphere.objectgrid.jpa.ClientLoader
 * @since WAS XD 6.1.0.3
 */
@PluginOutputFormat(keyFormat=OutputFormat.RAW, valueFormat=OutputFormat.RAW)
public class JPALoader implements RetryableLoader {
    private final static String CLASS_NAME = JPALoader.class.getName();

    private final static TraceComponent tc = Tr.register(CLASS_NAME, Constants.TR_JPA_GROUP_NAME,
            Constants.TR_RESOURCE_BUNDLE_NAME);

    protected JPATxCallback txCallback;

    protected String entityClassName;

    protected Class entityClass;

    // make it default so won't shown in javadoc
    JPAKeyMetadata keyMd = null;

    // make it default so won't shown in javadoc
    boolean initialized = false;

    private static Map clientLoadCallbackMap = new HashMap();

    protected int preloadPartition = -1;

    protected int retryIndicationSlot = -1;

    protected String ogName = null;

    protected String mapName = null;

    // make it default so won't shown in javadoc
    int partitionId = -1;

    // make it default so won't shown in javadoc
    JPAProviderHelperImpl jpaProviderHelper = null;

    ExceptionMapper exceptionMapper = null;

    // We need a transformer here to make a copy of the entity.
    // The entity has to be copied before em.persist. See batchUpdate method for more details.
    ObjectTransformer transformer = null;

    MapSerializerPlugin ms;

    /**
     * Default constructor
     */
    public JPALoader() {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
            Tr.debug(tc, "JPA Loader created: " + this);
    }

    public void setRetryIndicationSlot(int slotNumber) {
        retryIndicationSlot = slotNumber;
    }

    void getJPATxCallback(ObjectGrid grid, BackingMap bmap) throws LoaderException {
        TransactionCallback cb = grid.getTransactionCallback();

        ogName = grid.getName();
        mapName = bmap.getName();

        if (cb instanceof JPATxCallback) {
            txCallback = (JPATxCallback) cb;
        }

        if (txCallback == null) {
            Tr.error(tc, NLSConstants.JPA_TX_CALLBACK_NOT_FOUND_CWOBJ3131E);
            throw new LoaderException("A com.ibm.websphere.objectgrid.jpa.JPATxCallback transaction callback plugin "
                    + "is epxected to be configured in the ObjectGrid " + ogName
                    + " to use a com.ibm.websphere.objectgrid.jpa.JPALoader loader.");
        }

        partitionId = bmap.getPartitionId();

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEventEnabled())
            Tr.event(tc, "JPATxCallback =" + txCallback + ", pid=" + partitionId + ", " + this);
    }


    void initialize(ObjectGrid grid, BackingMap bmap) throws LoaderException {
        if (!initialized) {
            getJPATxCallback(grid, bmap);

            if (bmap.getEntityMetadata() != null) {
                throw new LoaderException("A com.ibm.websphere.objectgrid.jpa.JPALoader is configured for ObjectGrid "
                        + ogName + " map " + mapName + ") which has EntityMetadata associated. "
                        + "Use com.ibm.websphere.objectgrid.jpa.JPAEntityLoader instead.");
            }

            SerializerAccessor sac = bmap.getSerializerAccessor();
            MapSerializerPlugin ms = sac.getMapSerializerPlugin();
            if (ms != null) {
                this.ms = ms;
            }

            // Get the persistence unit name.
            String puName = txCallback.getPersistenceUnitName();

            if (entityClass == null) {
                throw new IllegalArgumentException(
                        "The property entityClassName is not configured for the JPALoader of ObjectGrid " + ogName
                                + " map " + mapName + ".");
            }

            keyMd = JPAHelper.getJPAMetadata(puName, entityClass);

            jpaProviderHelper = new JPAProviderHelperImpl(keyMd);

            exceptionMapper = txCallback.getExceptionMapper();

            transformer = bmap.getObjectTransformer();
            if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
                Tr.debug(tc, "JPALoader", "transformer=" + transformer);

            initialized = true;
        }
    }

    /**
     * Write the data from ObjectGrid entity maps to JPA.
     * <p>
     * The LogSequence contains a sequence of log elements which contain keys and values.
     * These keys are values are used to interact with JPA to insert, update, or merge a
     * JPA entity.
     * @param txid the transaction id object
     * @param sequence the set of changes being applied to this object's map
     * @see Loader#batchUpdate(TxID, LogSequence)
     */
    public void batchUpdate(TxID txid, LogSequence sequence) throws LoaderException, OptimisticCollisionException {
        if (retryIndicationSlot != -1) {
            Boolean retry = (Boolean) txid.getSlot(retryIndicationSlot);
            if (retry != null && retry.booleanValue()) {
                batchUpdate(txid, sequence, true);
            } else {
                batchUpdate(txid, sequence, false);
            }
        } else {
            batchUpdate(txid, sequence, false);
        }
    }

    private void batchUpdate(TxID txid, LogSequence sequence, boolean isRetry) throws LoaderException,
            OptimisticCollisionException {
        // this is a JPA EntityManager
        EntityManager em = null;
        Object key = null;
        try {
            KeySerializerPlugin ks = null;
            ValueSerializerPlugin vs = null;

            if (ms != null) {
                ks = ms.getKeySerializerPlugin();
                vs = ms.getValueSerializerPlugin();
            }

            em = txCallback.getEntityManager(txid);

            Iterator<LogElement> changes = sequence.getPendingChanges();
            Object value = null;
            while (changes.hasNext()) {
                LogElement elem = changes.next();
                key = elem.getKey();
                if (ks != null) {
                    // In the case of a key serializer, JPA requires the POJO object.
                    key = ((SerializedKey)key).getObject();
                }

                switch (elem.getType().getCode()) {
                case LogElement.CODE_INSERT:
                    value = elem.getCurrentValue();
                    if (vs != null) {
                        // In the case of a value serializer, JPA requires the POJO object.
                        value = ((SerializedValue)value).getObject();
                    }
                    if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                        Tr.debug(tc, this.mapName + ": persist (retry=" + isRetry + ") " + key + ", " + value);
                    }
                    if (isRetry) {
                        em.merge(value);
                    } else {
                        // 1. Why do we need to make a copy?
                        //   OpenJPA doesn't allow two EntityManager instances used in the same thread.
                        //   Since OpenJPA entities cache EntityManager instance, therefore basially OpenJPA doesn't
                        //   allow an attached entity being used in multiple threads.
                        //   However, with the way ObjectGrid is running, this is possible. Here is the scenario:
                        //   1). Thread 1 comes in insert an entity into the grid.
                        //   2). later on, this entity is pushed to OpenJPA by the WBLoader thread by calling
                        //     JPALoader.batchUpdate method.
                        //   3). The JPALoader.batchUpdate method calles em.persit
                        //   4). During the commit time, ObjectGrid calls em.commit, and then em.clear().
                        //   However, if before em.clear() is called in step 4, and another thread comes in
                        //   to read the entity from ObjectGrid, this will violate the OpenJPA restriction.
                        //   In order to fix this problem, we will make a copy of the origional object
                        //   before calling em.persist(e).
                        // 2. In which case should we make a copy?
                        //   Always copy the value when no value serializer is used.
                        //   Even if it is COPY_TO_BYTES, the elem.getCurrentValue will return a common value b/w plugins and loader
                        //   Same holds for the COPY_TO_READ case.
                        //
                        //   If a value serializer is used, then the POJO we have here has been copied for the duration
                        //   of the transaction. This transaction is already isolated from other threads, and therefore
                        //   the isolating copy has already been made.
                        if (vs == null) {
                            value = transformer.copyValue(value);
                        }
                        em.persist(value);
                    }
                    break;
                case LogElement.CODE_UPDATE:
                case LogElement.CODE_UPSERT:
                    value = elem.getCurrentValue();
                    if (vs != null) {
                        // In the case of a value serializer, JPA requires the POJO object.
                        value = ((SerializedValue)value).getObject();
                    }
                    if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                        Tr.debug(tc, this.mapName + ": merge (retry=" + isRetry + ") " + key + ", " + value);
                    }
                    em.merge(value);
                    break;
                case LogElement.CODE_DELETE:
                    value = em.find(keyMd.getTheClass(), key);
                    if (value != null) {
                        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
                            Tr.debug(tc, this.mapName + ": remove (retry=" + isRetry + ") " + key + ", "
                                    + value);
                        em.remove(value);
                    } else if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
                        Tr.debug(tc, this.mapName + ": remove (retry=" + isRetry + ") " + key
                                + ", VALUE NOT FOUND. Skip delete.");
                    break;
                }
            }
        } catch (javax.persistence.OptimisticLockException ole) {
            OptimisticCollisionException oce = jpaProviderHelper.createOptimisticCollisionException(
                    "An OptimisticLockException is thrown in JPALoader of ObjectGrid " + ogName + " map " + mapName
                            + ": " + ole.getMessage(), this.ogName, this.mapName, key, ole);
            throw oce;
        } catch (Error e) {
            throw e;
        } catch (Exception e) {
            Throwable t = e;

            if (jpaProviderHelper.isOptimisticLockException(e)) {
                OptimisticCollisionException oce = jpaProviderHelper.createOptimisticCollisionException(
                        "An OptimisticLockException is thrown in JPALoader of ObjectGrid " + ogName + " map "
                                + mapName + ": " + e.getMessage(), this.ogName, this.mapName, key, e);
                throw oce;
            } else if (exceptionMapper != null) {
                t = exceptionMapper.map(e);
            }

            if (t instanceof LoaderException) {
                throw (LoaderException) t;
            }

            LoaderException le = new LoaderException("Exception caught in ObjectGrid " + ogName + " map " + mapName
                    + ": " + t.getMessage());
            le.initCause(t);
            throw le;
        }
    }

    /**
     * Returns the list of values corresponding to the set of keys passed in by
     * retrieving the data using JPA. The values are the JPA entities retrieved from
     * JPA as a result of JPA find operation.
     *
     * @param txid the transaction ID object
     * @param keyList List of keys
     * @param forUpdate <code>true</code> when a getForUpdate method is called.
     *                  This parameter can be used to do a "SELECT for UPDATE"
     *                  invocation for proper database locking.
     *
     * JPA as a result of JPA find operation.
     *
     * @see Loader#get(TxID, List, boolean)
     */
    public List get(TxID txid, List keyList, boolean forUpdate) throws LoaderException {
        // this is a JPA EntityManager
        EntityManager em = null;
        List<Object> result = new LinkedList<Object>();
        try {
            em = txCallback.getEntityManager(txid);

            Iterator<Object> iter = keyList.iterator();
            while (iter.hasNext()) {
                Object key = iter.next();
                if (ms != null && ms.getKeySerializerPlugin() != null) {
                    // In the case of a key serializer, JPA requires the POJO object.
                    key = ((KeyDataBytesImpl)key).getObject();
                }
                Object v = em.find(entityClass, key);
                if (v == null)
                    v = Loader.KEY_NOT_FOUND;
                result.add(v);
            }

            // Clear the entity manager to detach entities.
            em.clear();
        } catch (Error e) {
            throw e;
        } catch (Throwable e) {
            if (exceptionMapper != null) {
                e = exceptionMapper.map(e);
            }

            if (e instanceof LoaderException) {
                throw (LoaderException) e;
            }

            // Wrap the exception in LoaderException
            LoaderException le = new LoaderException("Exception caught in ObjectGrid " + ogName + " map " + mapName
                    + ": " + e.getMessage());
            le.initCause(e);
            throw le;
        }
        return result;
    }

    /**
     * This method initializes the loader and preloads the map using the
     * {@link ClientLoader#load(ObjectGrid, String, String, Map, Class, String, Map, boolean, ClientLoadCallback)}
     * method. The persistence unit name and persistence property map are retrieved from the
     * {@link JPATxCallback} TransactionCallback instance configured on the ObjectGrid.
     * <P>
     * For partition maps, the preload will only be called in one partition, which is specified by users
     * using the preloadPartition property. If the preloadPartition value is set to less than 0, or greater than
     * totalNumberOfPartition -1, no preload will happen.
     *
     * @param session a <code>Session</code> reference to use for demarcating transactions.
     * @param backingMap the <code>BackingMap</code> this object is associated with
     *
     * @throws LoaderException if an error occurs during processing of this method
     * Since the default value of preloadPartition is -1, the map will not be preloaded by default.
     * @throws LoaderException if a {@link JPATxCallback} is not configured.
     * <p>
     * @see Loader#preloadMap(Session, BackingMap)
     */
    public void preloadMap(Session session, BackingMap backingMap) throws LoaderException {
        initialize(session.getObjectGrid(), backingMap);

        ObjectGrid og = session.getObjectGrid();
        int numPartitions = backingMap.getPartitionManager().getNumOfPartitions();

        // Only call client loader data in one partition
        if (partitionId == preloadPartition) {
            Tr.info(tc, NLSConstants.PRELOAD_STARTS_CWOBJ3116I, new Object[] { og.getName(), backingMap.getName(),
                    new Integer(partitionId) });

            ClientLoader loader = ClientLoaderFactory.getClientLoader();

            ClientCallbackImpl callback = null;
            if (numPartitions > 1) {
                synchronized (clientLoadCallbackMap) {
                    callback = (ClientCallbackImpl) clientLoadCallbackMap.get(txCallback);
                    if (callback == null) {
                        callback = new ClientCallbackImpl();
                        clientLoadCallbackMap.put(txCallback, callback);
                    }
                }
            }

            try {
                Map propMap = null;
                JPAPropertyFactory factory = txCallback.getJPAPropertyFactory();
                if (factory != null) {
                    int pid = backingMap.getPartitionId();

                    propMap = factory.getPersistenceProperties(og, pid);
                }

                loader.load(og, backingMap.getName(), txCallback.getPersistenceUnitName(), propMap, entityClass, null,
                        null, true, callback);

                Tr.info(tc, NLSConstants.PRELOAD_FINISHES_CWOBJ3117I, new Object[] { og.getName(), backingMap.getName(),
                        new Integer(partitionId) });

            } catch (Throwable e) {
                Tr.error(tc, NLSConstants.PRELOAD_FAILS_CWOBJ3118E, new Object[] { og.getName(), backingMap.getName(),
                        new Integer(partitionId), e });

                LoaderException le = new LoaderException("Exception caught in ObjectGrid " + ogName + " map " + mapName
                        + " during preload: " + e.getMessage());
                le.initCause(e);
                throw le;
            }
        }
    }

    /**
     * @return Returns the preloadPartition.
     */
    public int getPreloadPartition() {
        return preloadPartition;
    }

    /**
     * Specify the partition at which the map preload will be launched. If the preloadPartition is
     * less than 0, or greater than totalNumberOfPartition -1, map preload will not be launched.
     * <p>
     * The default value of preloadPartition is -1. Therefore, by default, map preload will not be launched.
     * @param preloadPartition The preloadPartition to set.
     */
    public void setPreloadPartition(int preloadPartition) {
        this.preloadPartition = preloadPartition;
    }

    /**
     * Retrieve the JPA entity class name that should be loaded.
     * @return the JPA entity class name.
     */
    public String getEntityClassName() {
        return entityClassName;
    }

    /**
     * Set the name of the JPA entity class to load.
     * @param entityClassName the JPA entity class name to set.
     * @throws IllegalArgumentException if the entity class name cannot be found
     */
    public void setEntityClassName(String entityClassName) {
        this.entityClassName = entityClassName;

        try {
            entityClass = DoPrivUtil.contextClassLoaderForName(entityClassName);
        } catch (ClassNotFoundException e) {
            IllegalArgumentException le = new IllegalArgumentException(
                    "ClassNotFoundException caught in loading class " + entityClassName + " for ObjectGrid " + ogName
                            + " map " + mapName);
            le.initCause(e);
            throw le;
        }
    }

    /**
     * Retrieve the JPA transaction callback instance. The JPATxCallback instance can be used to retrieve the
     * EntityManager instance for this transaction.
     * @return the JPATxCallback instance
     * @see JPATxCallback#getEntityManager(TxID)
     */
    public JPATxCallback getTxCallback() {
        return txCallback;
    }

}
