/*
 * IBM Confidential OCO Source Material
 * 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2010-2012
 * The source code for this program is not published or otherwise divested
 * of its trade secrets, irrespective of what has been deposited with the
 * U.S. Copyright Office.
 *
 * $HeadURL$
 * $Id$
 *
 * Change Activity:
 *
 * Reason       Version   Date         Userid    Change Description
 * ------------ --------- ------------ --------- ---------------------------------------
 * b18299       7.1.0.0   03/31/2010  katarya    new part
 * _____________________________________________________________________________________
 */
package com.ibm.websphere.objectgrid.management;

import java.util.logging.Level;

import javax.management.openmbean.CompositeData;
import javax.management.openmbean.TabularData;

/**
 * This MBean interface allows user to manipulate the behaviors of heartbeat and leader manager and other
 * catalog service specific actions. The object name pattern for this MBean is:
 *
 * <pre>
 * com.ibm.websphere.objectgrid:type=CatalogService
 * </pre>
 *
 * If ObjectGrid is running in a WebSphere Application Server process, more key=value pairs may be added to the object
 * name.
 *
 * @since 7.1, XC10
 *
 */
public interface CatalogServiceManagementMBean {
    /**
     * Constant representing a core group membership change notification.  The user data associated
     * with this notification is a CompositeData.<p>
     *
     * The CompositeData includes the following items:
     * <table>
     * <tr>
     *     <td><strong><u>Item Name</u></strong><td><strong><u>Type</u></strong><td><strong><u>Description</u></strong>
     * </tr>
     * <tr>
     *     <td>MemberName</td>
     *     <td>String</td>
     *     <td>The name of the server that is included in the core group.</td>
     * </tr>
     * </table>
     */
    public static String COREGROUP_MEMBERSHIP_CHANGE = "og.server.coregroup.membership.change";

    /**
     * Constant representing an eXtreme Scale server start notification.<p>
     *
     * The UserData argument of the Notification includes a TabularData that includes
     * information for each of the servers.  Each CompositeData (row in the TabularData) contains
     * the following items:
     *
     * <table>
     * <tr>
     *     <td><strong><u>Item Name</u></strong><td><strong><u>Type</u></strong><td><strong><u>Description</u></strong>
     * </tr>
     * <tr>
     *     <td>HAPort</td>
     *     <td>String</td>
     *     <td>The port number of the high availability manager.</td>
     * </tr>
     * <tr>
     *     <td>Host</td>
     *     <td>String</td>
     *     <td>The host/ip address of the server.</td>
     * </tr>
     * <tr>
     *     <td>JMXServiceURL</td>
     *     <td>String</td>
     *     <td>The JMX service url used to access the server.</td>
     * </tr>
     * <tr>
     *     <td>ServerName</td>
     *     <td>String</td>
     *     <td>The name of the server.</td>
     * </tr>
     * <tr>
     *     <td>ZoneName</td>
     *     <td>String</td>
     *     <td>The name of the zone that the server belongs.</td>
     * </tr>
     * </table>
     */
    public static String SERVER_EVENT_STARTED = "og.server.event.start";

    /**
     * Constant representing an eXtreme Scale server stop notification.<p>
     *
     * The UserData argument of the Notification includes a TabularData instance where
     * each CompositeData contains the following items:
     *
     * <table>
     * <tr>
     *     <td><strong><u>Item Name</u></strong><td><strong><u>Type</u></strong><td><strong><u>Description</u></strong>
     * </tr>
     * <tr>
     *     <td>ServerName</td>
     *     <td>String</td>
     *     <td>The name of the server.</td>
     * </tr>
     * </table>
     */
    public static String SERVER_EVENT_STOPPED = "og.server.event.stop";

    /**
     * Constant representing a heartbeat frequency level at a typical rate.
     * <p>
     *
     * A typical heartbeat frequency allows reasonable failover detection and resource utilization. This value is the
     * default.
     */
    public static final int HEARTBEAT_FREQUENCY_LEVEL_TYPICAL = 0;

    /**
     * Constant representing a heartbeat frequency level at an aggressive rate.
     * <p>
     *
     * An increased heartbeat frequency allows failures to be detected more quickly, but can also uses additional CPU
     * and network resources. This level is more sensitive to missing heartbeats when the server is stressed.
     */
    public static final int HEARTBEAT_FREQUENCY_LEVEL_AGGRESSIVE = -1;

    /**
     * Constant representing a heartbeat frequency level at relaxed rate.
     * <p>
     *
     * A decreased heartbeat frequency increases the time to detect failures, but also decreases CPU and network
     * utilization.
     */
    public static final int HEARTBEAT_FREQUENCY_LEVEL_RELAXED = 1;

    /*
     * /** Sets the heartbeat frequency level. The heartbeat frequency level tells the catalog service how aggressively
     * to check for container failures. The default is 0, which is suitable for most environments.
     * <p>
     *
     * Valid values include:
     * <ul>
     * <li> {@link #HEARTBEAT_FREQUENCY_LEVEL_TYPICAL} (default)
     * <li> {@link #HEARTBEAT_FREQUENCY_LEVEL_RELAXED}
     * <li> {@link #HEARTBEAT_FREQUENCY_LEVEL_AGGRESSIVE}
     * </ul>
     *
     * @param frequencyLevel
     *            one of three values: -1, 0 or 1 as defined by the constants that begin with name
     *            <tt>HEARTBEAT_FREQUENCY_LEVEL</tt>.
     * @throws IllegalArgumentException
     *             if the frequencyLevel is out of range.
     *
    public void setHeartBeatFrequencyLevel(int frequencyLevel);
    */

    /**
     * Constant representing the ORB or Object Request Broker transport
     */
    public static final String ORB = "ORB";

    /**
     * Constant representing the eXtremeIO transport
     */
    public static final String XIO = "eXtremeIO";

    /**
     * Retrieves the heartbeat frequency level.
     * <p>
     *
     * Valid values include:
     * <ul>
     * <li> {@link #HEARTBEAT_FREQUENCY_LEVEL_TYPICAL}
     * <li> {@link #HEARTBEAT_FREQUENCY_LEVEL_RELAXED}
     * <li> {@link #HEARTBEAT_FREQUENCY_LEVEL_AGGRESSIVE}
     * </ul>
     *
     * @return the heartbeat frequency level: -1, 0 or 1 as defined by the constants that begin with name
     *         <tt>HEARTBEAT_FREQUENCY_LEVEL</tt>.
     */
    public int getHeartBeatFrequencyLevel();

    /**
     * Retrieves a CompositeData of each eXtreme Scale server that is currently registered with the catalog service.<p>
     *
     * The CompositeData includes the following items:
     *
     * <table>
     * <tr>
     *     <td><strong><u>Item Name</u></strong><td><strong><u>Type</u></strong><td><strong><u>Description</u></strong>
     * </tr>
     * <tr>
     *     <td>serverName</td>
     *     <td>String</td>
     *     <td>The name of the server that is registered with the catalog service.</td>
     * </tr>
     * </table>
     *
     * @return the CompositeData representing the currently registered eXtreme Scale servers.
     */
    public CompositeData getServers();

    /**
     * Retrieves the number eXtreme Scale servers that are currently registered with the catalog service.
     * @return the number of registered eXtreme Scale servers.
     */
    public int getNumberOfServers();

    /**
     * Provides support for logging user messages from processes outside the catalog and/or container servers.
     * Example: XC10 surfaced SNMP trap messages can be flowed from the SNMP agent which throws traps in the
     * console server (sMash) process, not typically an XS catalog/container server.
     *
     * @param level name describing the severity of the event which is compatible with java.util.logging.Level.parse(String name) where name may be either level name (ex. "SEVERE") or an integer value (ex. "1000") - @see {@link Level#parse(String)}
     * @param message for the end user (already sNLS rendered)
     *
     * @since 8.6
     */
    public void logMessage( String level, String message );

    /**
     * Provides indication if the catalog is primary.
     *
     * @return true for primary catalog.
     *
     * @since 8.6
     */
    public boolean isPrimary( );

    /**
     * Provides the numbers of outstanding in-bound and out-bound revisions which need to be replicated cross containers within one domain. 
     * For a given container, outstanding out-bound revisions need to be replicated from primary shards located in this container into replicas located in other containers.
     * In similar way, outstanding in-bound revisions need to be replicated from primary shards located in other containers into corresponding replicas located in this container. 
     * This operation can be used to check the differences in data revisions between containers within one domain.<p>
     * 
     * The result is a TabularData object, where each CompositeData (row in the TabularData) contains the following items:
     *
     * <table>
     * <tr>
     *     <td><strong><u>Item Name</u></strong><td><strong><u>Type</u></strong><td><strong><u>Description</u></strong></td>
     *     <td><strong><u>Since</u></strong></td>
     * </tr>
     * <tr>
     *     <td>Container</td>
     *     <td>String</td>
     *     <td>The container ID.</td>
     *     <td>8.6</td>
     * </tr>
     * <tr>
     *     <td>OutboundRevisions</td>
     *     <td>Long</td>
     *     <td>Number of out-bound revisions for container.</td>
     *     <td>8.6</td>
     * </tr>
     * <tr>
     *     <td>InboundRevisions</td>
     *     <td>Long</td>
     *     <td>Number of in-bound revisions for container.</td>
     *     <td>8.6</td>
     * </tr>
     * </table>
     * 
     * @return A TabularData object with the container replication state information. 
     * @since 8.6
     */
    public TabularData getContainerReplicationState();

    /**
     * Provides the numbers of outstanding in-bound and out-bound revisions which need to be replicated cross domains. 
     * Outstanding out-bound revisions need to be replicated from local primary shards into corresponding remote primary shards.
     * Outstanding in-bound revisions need to be replicated from remote primary shards into corresponding local primary shards. 
     * This operation can be used to check the differences in data revisions between different domains linked by MMR replication.<p>
     * 
     * The result is a TabularData object, where each CompositeData (row in the TabularData) contains the following items:
     *
     * <table>
     * <tr>
     *     <td><strong><u>Item Name</u></strong><td><strong><u>Type</u></strong><td><strong><u>Description</u></strong></td>
     *     <td><strong><u>Since</u></strong></td>
     * </tr>
     * <tr>
     *     <td>Domain</td>
     *     <td>String</td>
     *     <td>The domain name.</td>
     *     <td>8.6</td>
     * </tr>
     * <tr>
     *     <td>Container</td>
     *     <td>String</td>
     *     <td>The container ID.</td>
     *     <td>8.6</td>
     * </tr>
     * <tr>
     *     <td>OutboundRevisions</td>
     *     <td>Long</td>
     *     <td>Number of out-bound revisions for container.</td>
     *     <td>8.6</td>
     * </tr>
     * <tr>
     *     <td>InboundRevisions</td>
     *     <td>Long</td>
     *     <td>Number of in-bound revisions for container.</td>
     *     <td>8.6</td>
     * </tr>
     * </table>
     * 
     * @return A TabularData object with the domain replication state information. 
     * @since 8.6
     *
     */
    public TabularData getDomainReplicationState();

    /**
     * Returns the transport used by the catalog service domain.
     *
     * @see #ORB
     * @see #XIO
     * @return String containing the transport type
     * @since 8.6
     */
    public String getTransport();

 }
