// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2006
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// 1.4 XD/ws/code/objectgrid/src/com/ibm/websphere/objectgrid/plugins/DistributionMode.java, WAS.objectgrid, WASX.XD, b0614.107 2/14/06 17:22:22
//
// Module  :  DistributionMode.java
//
// Source File Description: 
//
//    Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M4_B.17 WASX.XD   06/02/2005   sutter    new code drop 
// 281469       WASX.XD   06/07/2005   sutter    Complete Distributed Commit processing
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.plugins;

/**
 * This class is used to define the "distribution" mode when the
 * <code>serialize()</code> method on the <code>LogSequenceTransformer</code> interface is
 * used.  The caller is expected to pass one of the static constant
 * variables defined in this class to the <code>serialize()</code>
 * method.  The <code>DistributionMode</code> works with the <code>LogSequenceFilter</code>
 * to limit the amount of data distributed via the <code>LogSequence</code>s.
 *  
 * @since WAS XD 6.0
 * 
 * @see LogSequenceFilter
 */
public final class DistributionMode 
{
    /**
     * The INVALIDATE distribution mode is used to communicate invalidates to the
     * receivers of the LogSequence.  This means that any updates (Updates, Deletes, Evicts)
     * in the current LogSequence will be communicated as invalidates on the receiving
     * side.  This allows the receivers of the LogSequence to properly invalidate
     * entries that are now out of date due to the LogSequence that was just processed.
     * The INVALIDATE mode will only send the Keys in the distributed LogSequence since
     * that is all that is required for an unqualified invalidate.
     */
    public final static DistributionMode INVALIDATE = new DistributionMode(1);

    /**
     * The INVALIDATE_CONDITIONAL is very much like the <code>INVALIDATE</code> distribution
     * mode except that it also takes into account the versioned value data associated
     * with each entry.  Thus, before the invalidates are performed on the receiving
     * side, a comparison is done between the versioned value and the value currently 
     * in the receiving side ObjectMap.  If the versioned value is older than the
     * current value, then the invalidate is not performed.
     * 
     * @see #INVALIDATE
     */
    public final static DistributionMode INVALIDATE_CONDITIONAL = new DistributionMode(2);

    /**
     * The PUSH distribution mode is used to communicate changes to the receivers
     * of the LogSequence.  This means that any LogElements in the current
     * LogSequence will be communicated as corresponding LogElements on the receiving
     * side.  This allows the receivers of the LogSequence to properly update their
     * instance of the ObjectMap.  The PUSH mode will process the LogSequence
     * unconditionally, without any concern of optimistic collisions.  This means
     * that an Update done via PUSH mode could overwrite the value in the receiving
     * ObjectMap that was performed via another transaction.
     */
    public final static DistributionMode PUSH = new DistributionMode(3);

    /**
     * The PUSH_CONDITIONAL is very much like the <code>PUSH</code> distribution
     * mode except that it also takes into account the versioned value data associated
     * with each entry.  Thus, before the updates are applied on the receiving
     * side, a comparison is done between the versioned value and the value currently 
     * in the receiving side ObjectMap.  If the versioned value is older than the
     * current value, then the operation is not performed.
     * 
     * @see #PUSH
     */
    public final static DistributionMode PUSH_CONDITIONAL = new DistributionMode(4);

    /**
     * Id that uniquely identifies the DistributionMode instance.
     */
    final private int ivModeId;

    /**
     * Hidden CTOR for enumerated type idiom to use.
     * 
     * @param modeId is the unique id for the constructed DistributionMode instance.
     */
    private DistributionMode( int modeId )
    {
        ivModeId = modeId;
    } 

    /**
     * Hidden default CTOR for enumerated type idiom to use. 
     */
    private DistributionMode()
    {
        this(4);  // default is PUSH_CONDITIONAL
    }    

    /**
     * Checks whether the mode is conditional mode or not. A mode is conditional if it is
     * either <code>PUSH_CONDITIONAL</code> or <code>INVALIDATE_CONDITIONAL</code>.
     * <p>
     * @return true if the mode is conditional; false otherwise
     * 
     * @see #PUSH_CONDITIONAL
     * @see #INVALIDATE_CONDITIONAL
     */
    public boolean isConditional() {
        return ((ivModeId == PUSH_CONDITIONAL.ivModeId) || (ivModeId == INVALIDATE_CONDITIONAL.ivModeId));
    }

    /**
     * Checks whether the mode is a invalidate mode or not. A mode is a invalidate mode if it is
     * either <code>INVALIDATE</code> or <code>INVALIDATE_CONDITIONAL</code>.
     * <p>
     * @return true if the mode is a invalidate mode; false otherwise
     * 
     * @see #INVALIDATE
     * @see #INVALIDATE_CONDITIONAL
     */
    public boolean isInvalidate() {
        return ((ivModeId == INVALIDATE.ivModeId) || (ivModeId == INVALIDATE_CONDITIONAL.ivModeId));
    }

    /**
     * Checks whether the mode is a push mode or not. A mode is a push mode if it is
     * either <code>PUSH</code> or <code>PUSH_CONDITIONAL</code>.
     * <p>
     * @return true if the mode is a push mode; false otherwise
     * 
     * @see #PUSH
     * @see #PUSH_CONDITIONAL
     */
    public boolean isPush() {
        return ((ivModeId == PUSH_CONDITIONAL.ivModeId) || (ivModeId == PUSH.ivModeId));
    }

    /**
     * Return the int representation of mode 
     * @return the int representation of mode 
     */
    public int toInt() {
        return ivModeId;
    }

    /**
     * Returns a string representation of the <code>DistributionMode</code>. 
     *
     * @return  a string representation of the <code>DistributionMode</code>.
     */
    public String toString() {
        switch (ivModeId) {
        case 1: return "INVALIDATE DistributionMode";
        case 2: return "INVALIDATE_CONDITIONAL DistributionMode";
        case 3: return "PUSH DistributionMode";
        case 4: return "PUSH_CONDITIONAL DistributionMode";
        default: return "UNKNOWN DistributionMode";
        }
    }  
}
