// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2006
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// 1.5 XD/ws/code/objectgrid/src/com/ibm/websphere/objectgrid/plugins/Evictor.java, WAS.objectgrid, WASX.XD, b0614.107 2/14/06 17:22:27
//
// Module  :  Evictor.java
//
// Source File Description:
//
//    Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M1_B.1  WASX.XD   01/14/2005   sutter    new code drop
// XD60_M3_B.10 WASX.XD   04/07/2005   sutter    Re-factor datagrid/datagrid.impl
// 294637.1     WASX.XD   08/09/2005   kjlaw     correct javadoc for apply method.
// d347038      WXD601.XD 02/13/2006   kjlaw     Need to maintain TTLData even when map is not activated.
// ------------ --------- ------------ --------- -----------------------------------------

package com.ibm.websphere.objectgrid.plugins;

import com.ibm.websphere.objectgrid.BackingMap;
import com.ibm.websphere.objectgrid.ObjectGrid;
import com.ibm.websphere.objectgrid.ObjectMap;
import com.ibm.websphere.objectgrid.OutputFormat;
import com.ibm.websphere.objectgrid.plugins.io.annotations.PluginOutputFormat;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.SerializedEntry;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.SerializedKey;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.SerializedValue;

/**
 * <p>Data contained in a <code>BackingMap</code> are evicted when the map is
 * full. This plugin is used by the <code>BackingMap</code> to determine when
 * and what to evict from the map based on some algorithm (LRU, LFU, time
 * based, etc).</p>
 * <p>
 * An <code>Evictor</code> implementation that also implements the {@link BackingMapLifecycleListener}
 * interface will be automatically added as an {@link EventListener} on the {@link BackingMap} when the
 * evictor set on the backing map.
 * <p>
 * An <code>Evictor</code> may also implement the {@link BackingMapPlugin} interface in order
 * to receive enhanced BackingMap plug-in lifecycle method calls. The plug-in is then
 * also required to correctly implement each of the bean methods related to
 * introspection of its state (for example <code>isInitialized()</code>,
 * <code>isDestroyed()</code>, etc).
 *
 * @since WAS XD 6.0, XC10
 *
 * @see BackingMap#addMapEventListener(EventListener)
 * @see BackingMap#setEvictor(Evictor)
 * @see EvictorData
 */
public interface Evictor {

    /**
     * <p>Called by a <code>BackingMap</code> instance during the evictor
     * initialization time.</p>
     *
     * <p>The <code>BackingMap</code> calls this method so the Evictor
     * instance can have references to the <code>BackingMap</code> and
     * <code>EvictionEventCallback</code> instances. The evictor can signal
     * events to have specific entries evicted using the
     * <code>EvictionEventCallback</code>.</p>
     *
     * @param map the BackingMap instance
     * @param callback the EvictionEventCallback instance
     *
     * @see BackingMap
     * @see EvictionEventCallback
     */
    void initialize(BackingMap map, EvictionEventCallback callback);

    /**
     * <p>Called when the <code>BackingMap</code> associated with this evictor is
     * destroyed.</p>
     *
     * <p>This method is the opposite of the <code>initialize</code> method. When
     * it is called, the Evictor can free up any resources it uses.</p>
     *
     * @see ObjectGrid#destroy()
     */
    void destroy();

    /**
     * Called after a transaction has committed to allow the  evictor to track
     * object usage in the <code>BackingMap</code>.
     * <p>
     * This method also reports any entries that have been successfully evicted.
     * Note, this method is not called for transactions that are rolled back. If
     * there is a need to track object usage for rolled back transactions, the
     * evictor must implement the <code>RollbackEvictor</code> interface as
     * well.
     * <p>
     * This method is called after a transaction has completed. Consequently,
     * all transaction locks that were acquired by the completed transaction
     * are no longer held. Potentially, multiple threads could call this method
     * concurrently and each thread would be completing its own transaction.
     * Since transaction locks are already released by the completed transaction,
     * this method must provide its own synchronization to ensure it is thread safe.
     *
     * For an Evictor in an {@link ObjectMap} that is configured to use {@link OutputFormat#RAW}
     * for the keys or values, the keys and values objects in the LogSequence
     * will be {@link SerializedKey} or {@link SerializedValue} objects respectively. If required, you
     * can use the {@link SerializedEntry#getObject()} method to retrieve (possibly inflating
     * the serialized object) the original key or value object.
     *
     * To override the map's output format configuration, use the {@link PluginOutputFormat} 
     * annotation in the implementation class.
     *
     * @param sequence the <code>LogSequence</code> of changes committed to the map
     *
     * @see RollbackEvictor
     */
    void apply(LogSequence sequence);

    /**
     * This method is called to activate the Evictor. Until this method is called,
     * the Evictor must not use the EvictionEventCallback interface to
     * evict any map entries.  If it does use the EvictionEventcallback interface to
     * evict map entries prior to activate being called, an IllegalStateException
     * is thrown.
     */
    void activate();

    /**
     * This method is called to deactivate the Evictor. Once this method is called,
     * the Evictor must quit using the EvictionEventCallback interface to
     * evict any map entries.  If it does use the EvictionEventcallback interface after
     * this method is called, an IllegalStateException is thrown.
     */
    void deactivate();
}
