// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2009
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $HeadURL$
// $Id$
//
// Source File Description: 
//
//  Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// b15582       7.0       04/17/2009   jian.tang Add ExceptionMapper for mapping exceptions
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.plugins;


/**
 * The <code>ExceptionMapper</code> interface allows users to plug in specific exception mapping functionality. 
 * When ObjectGrid gets a generic <code>Throwable</code> from a third-party product, this interface can be 
 * used to map it to a different <code>Throwable</code> which can be more friendly or consumable to ObjectGrid.  
 * <p>
 * For example, when a <code>JPALoader</code> or a <code>JPAEntityLoader</code> is used, the application could implement 
 * this interface to map a generic <code>javax.persistence.PersistenceException</code> to more consumable exceptions. 
 * The implementation class could introspect the SQL state and error code of the <code>java.sql.SQLException</code> chained in the 
 * JPA exception and throw a <code>LoaderNotAvailableException</code> if the SQL state or error code indicates 
 * the database server or network is not functional or the database runs out of resources.
 * <p>
 * Here is an example:
 * <pre>
 *    public Throwable map(Throwable original) {
 * 
 *         Throwable cause = original;
 * 
 *         while (cause != null) {
 *             // keep looping to check the next chained exception
 * 
 *             if (cause instanceof SQLException) {
 *                 // Only check if the exception is an SQLException
 * 
 *                 SQLException sqle = (SQLException) cause;
 *                 
 *                 // If the loader not available SQL state set contains this SQL state, then
 *                 // we return a LoaderNotAvailableException with the original exception chained in it.
 *                 if (loaderNotAvailableSQLStateSet.contains(sqle.getSQLState())) {
 *                     return new LoaderNotAvailableException(original);
 *                 }
 *                 
 *                 // If the loader not available SQL error code set contains this error code, then
 *                 // we return a LoaderNotAvailableException with the original exception chained in it
 *                 if (loaderNotAvailableSQLErrorSet.contains(new Integer(sqle.getErrorCode()))) {
 *                     return new LoaderNotAvailableException(original);
 *                 }
 *             }
 *             
 *             // Get the next chained exception
 *             Throwable newCause = cause.getCause();
 *             
 *             // Safe-guard check to avoid indefinite loop if the exception chains itself
 *             if (newCause == cause) {
 *                 // Always return the original exception if cannot map it.
 *                 return original;
 *             } else {
 *                 cause = newCause;
 *             }
 *         }
 * 
 *         // Always retrun the original exception if cannot map it.
 *         return original;
 *     }
 * </pre>
 * Currently, the {@link ExceptionMapper} can be configured in the following ObjectGrid beans:
 * <ul>
 * <li>JPATxCallback</li>
 * </ul>
 * 
 * @since 7.0
 */
public interface ExceptionMapper {

    /**
     * This method maps an <code>Throwable</code> to a more friendly or consumable <code>Throwable</code> if possible.
     * If the passed-in <code>Throwable</code> cannot be mapped, the implementation is expected to return it.   
     *    
     * @param original the <code>Throwable</code>  to be mapped
     * @return the mapped <code>Throwable</code> , or the original <code>Throwable</code>  if it cannot be mapped.  
     * 
     */
    Throwable map(Throwable original);
    
}
