// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2009
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  LogElement.java
//
// Source File Description:
//
//    Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M1_B.1  WASX.XD   01/14/2005   sutter    new code drop
// XD60_M2_B.2  WASX.XD   03/21/2005   sutter    M2 Deliverable
// XD60_M3_B.10 WASX.XD   04/07/2005   sutter    Re-factor datagrid/datagrid.impl
// XD60_M3_B.11 WASX.XD   04/12/2005   kjlaw     add setVersionedObject method for Loader to use.
// 267383       WASX.XD   04/20/2005   jhanders  touch support
// 258492       WASX.XD   05/05/2005   kjlaw     add changes for rollback of BackingMap changes.
// ------------ --------- ------------ --------- -----------------------------------------

package com.ibm.websphere.objectgrid.plugins;

import java.io.Serializable;

import com.ibm.websphere.objectgrid.ObjectMap;
import com.ibm.websphere.objectgrid.OutputFormat;
import com.ibm.websphere.objectgrid.Session;
import com.ibm.websphere.objectgrid.em.EntityManager;
import com.ibm.websphere.objectgrid.io.XsDataInputStream;
import com.ibm.websphere.objectgrid.plugins.io.KeySerializerPlugin;
import com.ibm.websphere.objectgrid.plugins.io.ValueSerializerPlugin;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.SerializedEntry;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.SerializedKey;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.SerializedValue;

/**
 * LogElements are the individual entries within a LogSequence. A LogElement has
 * attributes such as operation type (delete, insert, update, etc.), current value,
 * last access time, versioned value, etc. A LogElement is created during a transaction
 * to record in-flight operations.
 *
 * For a LogElement on an {@link ObjectMap} that is configured to use {@linkplain OutputFormat#RAW} for the
 * keys or values, the keys or values objects in the LogElement will be {@link SerializedKey} or 
 * {@link SerializedValue} objects respectively. If required, you can use the 
 * {@link SerializedEntry#getObject()} method to retrieve (possibly inflating the serialized object) 
 * the original key or value object.
 *
 * To override the map's output format configuration, use the {@link PluginOutputFormat} 
 * annotation in the caller of the LogElement.
 *
 * @since WAS XD 6.0, XC10
 *
 * @see LogSequence
 */
public interface LogElement extends Serializable {

    /**
     * The <code>Type</code> class is used to represent a <code>LogElement</code>
     * type.
     *
     * @since WAS XD 6.0
     */
    // TODO This is the second interface that has defined the Type class.  Should it
    // be pulled out to its own file scope for more general use?
    public class Type implements Comparable {
        private final String name;
        private final int code;
        Type(String name, int code) {
            this.name = name;
            this.code = code;
        }

        /**
         * Gets the type code for this object.
         *
         * @return the type code
         */
        public int getCode() {
            return code;
        }

        /**
         * @see Comparable#compareTo(Object)
         */
        public int compareTo(Object object) {
            Type type = (Type) object;
            return code - type.code;
        }

        /**
         * @see Object#toString()
         */
        public String toString() {
            return ("LogElement Type: " + name);
        }
    }

    /**
     * The type code constant for INSERT.
     *
     * @see #INSERT
     * @see Type#getCode()
     */
    public int CODE_INSERT = 0;

    /**
     * The type code constant for <code>UPDATE</code>.
     *
     * @see #UPDATE
     * @see Type#getCode()
     */
    public int CODE_UPDATE = 1;

    /**
     * The type code constant for <code>DELETE</code>.
     *
     * @see #DELETE
     * @see Type#getCode()
     */
    public int CODE_DELETE = 2;

    /**
     * The type code constant for <code>EVICT</code>.
     *
     * @see #EVICT
     * @see Type#getCode()
     */
    public int CODE_EVICT = 3;

    /**
     * The type code constant for <code>FETCH</code>.
     *
     * @see #FETCH
     * @see Type#getCode()
     */
    public int CODE_FETCH = 4;

    /**
     * The type code constant for <code>TOUCH</code>.
     *
     * @see #TOUCH
     * @see Type#getCode()
     */
    public int CODE_TOUCH = 5;

    /**
     * The type code constant for <code>CLEAR</code>.
     *
     * @see #CLEAR
     * @see Type#getCode()
	 * @since WAS XD 6.1.0.3
     */
    public int CODE_CLEAR = 7;

    /**
     * The type code constant for <code>LOCK</code>.
     *
     * @see #LOCK
     * @see Type#getCode()
     * @since 8.6
     */
    public int CODE_LOCK = 8;

    /**
     * The type code constant for <code>UPSERT</code>.
     *
     * @see #UPSERT
     * @see Type#getCode()
     * @since 8.6
     */
    public int CODE_UPSERT = 9;

    /**
     * The code constant for <code>UNDO_NOT_NEEDED</code>.  Used to indicate no
     * operation is needed to undo the changes for this <code>LogElement</code>
     * since this <code>LogElement</code> was never processed or it was an operation
     * that does not require an undo operation.
     */
    public int CODE_UNDO_NOT_NEEDED = -1;

    /**
     * The <code>Type</code> that represents the INSERT operation.
     */
    public final static Type INSERT = new Type("INSERT", CODE_INSERT);

    /**
     * The <code>Type</code> that represents the UPDATE operation.
     */
    public final static Type UPDATE = new Type("UPDATE", CODE_UPDATE);

    /**
     * The <code>Type</code> that represents the DELETE operation.
     */
    public final static Type DELETE = new Type("DELETE", CODE_DELETE);

    /**
     * The <code>Type</code> that represents the EVICT operation.
     */
    public final static Type EVICT = new Type("EVICT", CODE_EVICT);

    /**
     * The <code>Type</code> that represents the FETCH operation.
     */
    public final static Type FETCH = new Type("FETCH", CODE_FETCH);

    /**
     * The <code>Type</code> that represents the TOUCH operation.
     */
    public final static Type TOUCH = new Type("TOUCH", CODE_TOUCH);

    /**
     * The <code>Type</code> that represents the CLEAR operation.
	 * @since WAS XD 6.1.0.3
     */
    public final static Type CLEAR = new Type("CLEAR", CODE_CLEAR);

    /**
     * The <code>Type</code> that represents the LOCK operation.
     * @since 8.6
     */
    public final static Type LOCK = new Type("LOCK", CODE_LOCK);

    /**
     * The <code>Type</code> that represents the UPSERT operation.
     * @since 8.6
     */
    public final static Type UPSERT = new Type("UPSERT", CODE_UPSERT);

    /**
     * The <code>Type</code> that represents an UNDO action is NOT required for
     * this <code>LogElement</code>.
     */
    public final static Type UNDO_NOT_NEEDED = new Type("UNDO_NOT_NEEDED", CODE_UNDO_NOT_NEEDED);

    /**
     * Gets the type of this <code>LogElement</code>.  The type indicates what operation
     * needs to be applied to the map entry.
     *
     * @return the type of this <code>LogElement</code>.
     *         It can be one of: INSERT, UPDATE, DELETE, EVICT, FETCH, or TOUCH.
     *
     * @see Type
     */
    Type getType();

    /**
     * Gets the value for this <code>LogElement</code>.
     * <p>
     * For a LogElement on an {@link ObjectMap} that is configured to use a {@link ValueSerializerPlugin}, the
     * values in the LogSequence will be {@link SerializedValue} objects. If required, you
     * can use the {@link SerializedEntry#getObject()} method to retrieve (possibly inflating
     * the serialized object) the original value object.
     * <p>
     * The original value represents the new value that should be applied to the
     * <code>BackingMap</code> and <code>Loader</code>.  This value can be cast
     * to <code>ValueProxyInfo</code> when a value interface is in use in order
     * to determine the set of dirty attributes.
     *
     * @return the value in case of INSERT, UPDATE, or FETCH, <code>null</code>
     *         in the case of DELETE or EVICT.
     *
     * @see ValueProxyInfo
     */
    Object getCurrentValue();

    /**
     * Returns the <code>CacheEntry</code> for this key.
     *
     * The key, current committed value, etc. can be accessed from the
     * <code>CacheEntry</code>.
     *
     * @return the entry in the cache that is requested to be updated.
     *
     * @see CacheEntry#getCommittedValue()
     * @see #getKey()
     */
    CacheEntry getCacheEntry();

    /**
     * Answers <tt>true</tt> if this change has NOT been applied to the loader.
     * <p>
     * Changes can previously be applied to a loader using the
     * <code>ObjectMap.flush()</code> or <code>Session.flush()</code> methods.
     * This method reveals whether the change in this <code>LogElement</code>
     * has already been applied to the Loader using one of those methods.
     *
     * @return <tt>true</tt> if this change has NOT been applied to the loader.
     *
     * @see ObjectMap#flush()
     * @see Session#flush()
     */
    boolean isPending();

    /**
     * Gets the versioned object at the time the object was first associated
     * with the transaction.
     * <p>
     *
     * For a LogElement on an {@link ObjectMap} that is configured to use a
     * or {@link ValueSerializerPlugin}, the versioned object will be returned
     * as an {@link XsDataInputStream}, read
     * will be {@link SerializedKey} or {@link SerializedValue} objects respectively. If required, you
     * can use the {@link SerializedEntry#getObject()} method to retrieve (possibly inflating
     * the serialized object) the original key or value object.
     * For a LogElement on an {@link ObjectMap} that is configured to use a {@link ValueSerializerPlugin}
     * that generates version objects, the version object will be the data stream
     * representing the data.
     *
     * @return The versioned object.
     *
     * @see OptimisticCallback
     */
    Object getVersionedValue();

    /**
     * Used to update the versioned object after an update of map entry occurs.
     * <p>
     * The Loader can use this method when it is using an optimistic strategy and uses the
     * <code>OptimisticCallback.updateVersionedObjectForValue(Object)</code> method to
     * get an updated version object.
     *
     * @param v The versioned object.
     *
     * @see OptimisticCallback#updateVersionedObjectForValue(Object)
     */
    void setVersionedValue(Object v);

    /**
     * Returns the last access time associated with this <code>LogElement</code>.
     *
     * @return last access time
     */
    long getLastAccessTime();

    /**
     * Returns what operation must be performed to "undo" a prior change the
     * transaction made to the map entry.
     * <p>
     * Note, an undo type of <code>UNDO_NOT_NEEDED</code> is returned if nothing
     * needs to be undone for this <code>LogElement</code>.
     *
     * @return the "undo" type of this <code>LogElement</code>.
     *         It can be one of: INSERT, UPDATE, DELETE, or UNDO_NOT_NEEDED
     */
    public Type getUndoType();

    /**
     * Gets the "before image" of the value object.
     * <p>
     * The "before image" is the value object that existed in map entry prior
     * to applying a change to map entry. Note, it is possible for a
     * <code>null</code> reference to be returned (e.g. in the case where a new
     * map entry is created).
     * <p>
     * For a LogElement on an {@link ObjectMap} that is configured to use {@link OutputFormat#RAW} for the values,
     * the value will be a {@link SerializedValue} object. If required, you
     * can use the {@link SerializedEntry#getObject()} method to retrieve (possibly inflating
     * the serialized object) the original value object.
     * <p>
     * To override the map's output format configuration, use the {@link PluginOutputFormat} 
     * annotation in the caller of the LogElement.
     * 
     * @return the value prior to applying the change
     */
    public Object getBeforeImage();

    /**
     * Gets the "after image" value object.
     * <p>
     * The "after image" is the value object that existed in map entry after
     * applying a change to the map entry.  Note, it is possible for a
     * <code>null</code> reference to be returned (e.g. in the case where an
     * existing map entry is removed/evicted).
     * <p>
     * For a LogElement on an {@link ObjectMap} that is configured to use {@link OutputFormat#RAW} for the value,
     * the value will be a {@link SerializedValue} object. If required, you
     * can use the {@link SerializedEntry#getObject()} method to retrieve (possibly inflating
     * the serialized object) the original value object.
     * <p>
     * To override the map's output format configuration, use the {@link PluginOutputFormat} 
     * annotation in the caller of the LogElement.
     *
     * @return the value after applying the change
     */
    public Object getAfterImage();

    /**
     * Answers <tt>true</tt> if this LogElement is a result of a cascade operation. This only applies to
     * ObjectGrid EntityManager programming model.
     * <p>
     * ObjectGrid EntityManager supports cascade operations. For example, when persisting an
     * entity P, if P has a relation to entity C with <code>CascadeType.PERSIST</code> enabled, C will also be persisted
     * as a result of the cascade operation. The method isCascaded() returns true for
     * the LogElement object which represents C, and the method returns false for the LogElement object which
     * represents P.
     *
     * @see EntityManager
     * @since 6.1.0.5 FIX1
     * @return <tt>true</tt> if the LogElement object is a result of cascade operation.
     */
    public boolean isCascaded();

    /**
     * Returns the key for this LogElement.<p>
     * <p>
     * For a LogElement on an {@link ObjectMap} that is configured to use {@link OutputFormat#RAW} for the keys,
     * the value will be a {@link SerializedKey} object. If required, you
     * can use the {@link SerializedEntry#getObject()} method to retrieve (possibly inflating
     * the serialized object) the original key object.
     * <p>
     * To override the map's output format configuration, use the {@link PluginOutputFormat} 
     * annotation in the caller of the LogElement.
     * <p>
     * This method can be used instead of LogElement.getCacheEntry().getKey().
     *
     * @return the key for this LogElement.
     *
     * @see CacheEntry#getKey()
     * @since 7.0
     */
    public Object getKey();
}
