// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2006
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// 1.4 XD/ws/code/objectgrid.core/src/com/ibm/websphere/objectgrid/plugins/LogSequenceTransformer.java, WAS.objectgrid, WASX.XD, b0614.107 3/10/06 13:59:02
//
// Module  :  LogSequenceTransformer.java
//
// Source File Description: Reference JavaDoc
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M4_B.17 WASX.XD   Jun 3, 2005  sutter    New code drop
// 281469       WASX.XD   06/07/2005   sutter    Complete Distributed Commit processing
// d337874      WASX.XD   01/31/2006   rajeshk   LogSequence can not be serialized for some ObjectGrid instances
// d353959      WXD601.XD 03/10/2006   jhanders  Make LogSequenceTransformer working with near cache
// ------------ --------- ------------ --------- -----------------------------------------

package com.ibm.websphere.objectgrid.plugins;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.ObjectGrid;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.ObjectGridManagerImpl;
import com.ibm.ws.objectgrid.map.LogSequenceImpl;
import com.ibm.ws.xs.NLSConstants;

/**
 * This class is used to serialize and de-serialize <code>LogSequence</code>s. It provides an alternative to the
 * standard readObject/writeObject mechanism that is part of the <code>LogSequence</code> implementation. By using
 * these alternative methods, the user has more control over what parts and how the LogSequence is serialized and
 * inflated.
 * 
 * @since WAS XD 6.0
 * 
 * @see LogSequence
 */
public class LogSequenceTransformer {

    private final static String CLASS_NAME = LogSequenceTransformer.class.getName();

    private final static TraceComponent tc = Tr.register(CLASS_NAME, Constants.TR_GROUP_NAME,
            Constants.TR_RESOURCE_BUNDLE_NAME);

    private static final int COLLECTION_START = 900;

    private static final int COLLECTION_END = 901;

    /**
     * Provides an alternative to the standard serialization method of <code>writeObject</code>.
     * <p>
     * This method allows the caller to provide a filter for determining which <code>LogElement</code>s to include in
     * the serialization process. The <code>DistributionMode</code> parameter allows the caller to control the
     * serialization process.
     * 
     * @param logSequences
     *            a collection of LogSequences that need to be serialized into the stream. Caller must guarantee the
     *            collection contains only LogSequence objects.
     * @param stream
     *            the stream to serialize the LogSequence into. Caller must guarantee this argument is not
     *            <code>null</code>
     * @param filter
     *            a filter callback for determining whether to include a given LogElement in the serialized output. This
     *            argument is optional. A <code>null</code> value indicates not to filter the LogSequences
     * @param mode
     *            Conditional or unconditional distribution of changes. The call must guarantee this argument is not
     *            <code>null</code>
     * 
     * @throws IOException
     *             if an error occurs during serialization of the LogSequences
     * 
     * @see LogSequenceFilter
     * @see DistributionMode
     */
    public static void serialize(Collection /* LogSequence */logSequences, ObjectOutputStream stream,
            LogSequenceFilter filter, DistributionMode mode) throws IOException {

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.entry(tc, "serialize");
        }

        stream.writeInt(COLLECTION_START);
        stream.writeInt(logSequences.size());
        Iterator iter = logSequences.iterator();
        while (iter.hasNext()) {
            LogSequenceImpl logSequence = (LogSequenceImpl) iter.next();
            logSequence.serialize(stream, filter, mode);
        }
        stream.writeInt(COLLECTION_END);

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.exit(tc, "serialize");
        }
    }

    /**
     * Provides an alternative to the standard serialization method of <code>readObject</code>.
     * <p>
     * The caller needs to pass in the stream that contains the serialized version of the LogSequences that were
     * serialized using the <code>serialize</code> method.
     * 
     * @param stream
     *            the stream to de-serialize the LogSequences from.
     * @param objectGrid
     *            the ObjectGrid instance that owns this LogSequence.
     * 
     * @return a Collection of LogSequences
     * 
     * @throws IOException
     *             if an error occurs during deserialization of the LogSequences
     * @throws ClassNotFoundException
     *             if a class isn't found during deserialization of the LogSequences
     * 
     * @see #serialize(Collection, ObjectOutputStream, LogSequenceFilter, DistributionMode)
     */
    public static Collection /* LogSequence */inflate(ObjectInputStream stream, ObjectGrid objectGrid)
            throws IOException, ClassNotFoundException {

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.entry(tc, "inflate");
        }

        int tag;
        if ((tag = stream.readInt()) != COLLECTION_START) {
            Tr.error(tc, NLSConstants.REQUIRED_FIELD_NOT_FOUND_CWOBJ0024, new Object[] { new Integer(tag), CLASS_NAME });
            throw new IOException("Invalid stream content (start) for Collection of LogSequences.");
        }
        int size = stream.readInt();
        Collection logSequences = new ArrayList(size);
        for (int i = size - 1; i >= 0; --i) {
            LogSequenceImpl logSequence = new LogSequenceImpl(objectGrid);
            logSequence.inflate(stream);
            logSequences.add(logSequence);
        }
        if ((tag = stream.readInt()) != COLLECTION_END) {
            Tr.error(tc, NLSConstants.REQUIRED_FIELD_NOT_FOUND_CWOBJ0024, new Object[] { new Integer(tag), CLASS_NAME });
            throw new IOException("Invalid stream content (end) for Collection of LogSequences.");
        }

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.exit(tc, "inflate");
        }

        return logSequences;
    }
}