// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2006
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// 1.5 XD/ws/code/objectgrid/src/com/ibm/websphere/objectgrid/plugins/OptimisticCollisionException.java, WAS.objectgrid, WASX.XD, b0614.107 2/14/06 17:23:18
//
// Module  :  OptimisticCollisionException.java
//
// Source File Description:
//
//    See javadoc for this class.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M3_B.25  WASX.XD   Apr 14, 2005 kjlaw    new code drop
// XD601_M3_B.1  WASX.XD   Oct 24, 2005 kjlaw    update javadoc.
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.plugins;

import com.ibm.websphere.objectgrid.BackingMap;
import com.ibm.websphere.objectgrid.ObjectGridException;
import com.ibm.websphere.objectgrid.TxID;
import com.ibm.websphere.objectgrid.plugins.builtins.NoVersioningOptimisticCallback;

/**
 * This exception is thrown when optimistic locking strategy is used
 * and more than one update transaction collides on the same map entry
 * of a ObjectGrid instance.  The first transaction to commit updates the
 * version object for the map entry.  Other transactions that read this
 * same map entry prior to the commit have the prior version object.
 * When the other transactions try to commit, the version object it read
 * will not match the version that was last committed.  Thus the other
 * transactions are prevented from updating a map entry using "stale" data.
 * <p>
 * The default OptimisticCallback used by the runtime if one is not provided
 * by the application will cause this exception to be thrown if there is not a well
 * constructed equals(Object) method on your value object since the entire value
 * object is used as the version object.
 * <p>
 * Since this exception indicates the map entry contains stale data, ObjectGrid
 * will invalidate the stale map entry or entries as identified by the key parameter that
 * is passed to the <code>OptimisticCollisionException(String, String, String, Object)</code>
 * method. If this exception is thrown by a <code>Loader</code> plugin and a
 * <code>null</code> reference is used as the key parameter by the
 * loader, ObjectGrid assumes the loader does not know which entry caused the
 * exception. In this case, ObjectGrid will use the <code>LogSequence</code>
 * that was passed to the <code>Loader.batchUpdate(TxID, LogSequence)</code>
 * method to determine which map entries to invalidate.  Each <code>LogElement</code>
 * in the <code>LogSequence</code> of type update or delete will be invalidated.
 *
 * @since WAS XD 6.0
 *
 * @see NoVersioningOptimisticCallback
 * @see OptimisticCallback
 * @see BackingMap#setOptimisticCallback(OptimisticCallback)
 * @see Loader#batchUpdate(TxID, LogSequence)
 * @see LogElement#DELETE
 * @see LogElement#UPDATE
 */
public class OptimisticCollisionException extends ObjectGridException
{
    /**
     * The <code>serialVersionUID</code> for this class.
     */
    private static final long serialVersionUID = -3455749769693986460L;

    final private String ivMapName;
    final private Object ivKey;
    final private String ivGridName;

    /**
     * Constructs a new OptimisticCollisionException with the specified detail
     * message and key of a map of an ObjectGrid. The cause is not initialized,
     * and may subsequently be initialized by a call to the <code>initCause</code>
     * method.
     *
     * @param message   the detail message. The detail message is saved for later
     *                  retrieval by the <code>getMessage</code> method.
     * @param gridName  the name of the grid the map belongs to.
     * @param mapName   the name of the map.
     * @param key       the key or array of keys that caused the optimistic collision
     *                  exception to occur.  If more than a single key caused the
     *                  exception, then an array of object should be used for this
     *                  parameter.  Each array element in this case identifies a
     *                  single map entry that caused the exception to occur.  This is
     *                  useful when a Loader uses the batch update support of a JDBC driver.
     *                  A null reference should be passed if unable to determine which key
     *                  or set of keys caused this exception to occur.
     *
     * @see #getGridName()
     * @see #getKey()
     * @see #getMapName()
     * @see #getMessage()
     * @see #initCause(Throwable)
     */
    public OptimisticCollisionException(String message, String gridName, String mapName, Object key)
    {
        super(message);
        ivGridName = gridName;
        ivMapName = mapName;
        ivKey = key;
    }

    /**
     * Gets the name of the ObjectGrid instance the optimistic collision occurred in.
     *
     * @return the name of the ObjectGrid instance.
     */
    final public String getGridName()
    {
        return ivGridName;
    }

    /**
     * Gets the map name the optimistic collision occurred in.
     *
     * @return name of map.
     */
    final public String getMapName()
    {
        return ivMapName;
    }

    /**
     * Gets the key(s) that caused this exception to occur.
     *
     * @return key object or array of key objects that caused this exception to occur.
     */
    final public Object getKey()
    {
        return ivKey;
    }
}
