// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2012
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  InverseRangeIndex.java
//
// Source File Description:
//
//  Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// t9392        8.6       08/09/2012   boggar    New code drop
// ------------ --------- ------------ --------- -----------------------------------------

package com.ibm.websphere.objectgrid.plugins.index;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Properties;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.BackingMap;
import com.ibm.websphere.objectgrid.ObjectGridException;
import com.ibm.websphere.objectgrid.ObjectGridRuntimeException;
import com.ibm.websphere.objectgrid.OutputFormat;
import com.ibm.websphere.objectgrid.TxID;
import com.ibm.websphere.objectgrid.plugins.BackingMapPlugin;
import com.ibm.websphere.objectgrid.plugins.LogSequence;
import com.ibm.websphere.objectgrid.plugins.io.MapSerializerPlugin;
import com.ibm.websphere.objectgrid.plugins.io.annotations.PluginOutputFormat;
import com.ibm.websphere.objectgrid.plugins.io.datadescriptor.MapDataDescriptor;
import com.ibm.websphere.projector.md.EntityMetadata;
import com.ibm.ws.ffdc.FFDCFilter;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.ObjectGridManagerImpl;
import com.ibm.ws.objectgrid.index.inverserange.AttributeNamesParser;
import com.ibm.ws.objectgrid.index.inverserange.CompositeRangeAttribute;
import com.ibm.ws.objectgrid.index.inverserange.CompositeRangeSearchAttribute;
import com.ibm.ws.objectgrid.index.inverserange.InverseRangeConstants;
import com.ibm.ws.objectgrid.index.inverserange.InverseRangeSearchKey;
import com.ibm.ws.objectgrid.index.inverserange.RangeAttribute;

/**
 * InverseRangeIndex is a built-in index implementation of <code>MapIndexPlugin</code> for storing range data and could be used for looking up 
 * single item with in a range. 
 * <P>
 *
 * This is a final class and can not be extended by other classes.
 *  
 * <p>
 * 
 * When used with Java Objects (POJO), InverseRangeIndex uses Java reflection to dynamically introspect objects stored in a map when they are 
 * updated. It can index any individual fields or properties in the key or value portion of a map.  The index implementation uses
 * an algorithm to structurally store the data and retrieve using a specific look up key.
 * 
 * <P>
 * 
 * The following properties can be used to configure the index using the ObjectGrid
 * deployment descriptor XML file or by using API:
 * <table>
 * <tr>
 *   <td valign="top"><b><u>Property Name</u></b></td></td><td valign="top"><b><u>Applicability</u></b></td><td valign="top"><b><u>Description</u></b></td>
 * <tr>
 *   <td valign="top"><b>Name</b></td><td valign="top">POJO, Serializer</td><td valign="top">The name of the index.  The name must be unique for each map.</td>
 * </tr>
 * <tr>
 *   <td valign="top"><b>AttributeName</b></td><td valign="top">POJO, Serializer</td><td valign="top">The comma-delimited names of the attributes to index.<br><br>
 *   For field-access indexes, the attribute names are equivalent to the field names.<br>
 *   For property-access indexes, the attribute names is the JavaBean-compatible property names.<br>
 *   For indexes associated with a {@linkplain MapSerializerPlugin} configured BackingMap, the attribute name is a path to the attribute, 
 *   which can be in the KeySerializerPlugin or ValueSerializerPlugin.  See the {@link #setAttributeName(String)} method for details.  
 *   </td>
 * </tr>
 * <tr>
 *   <td valign="top"><b>FieldAccessAttribute</b></td><td valign="top">POJO</td><td valign="top">
 *   If true, the object is accessed using the fields directly.  If not specified or false, the
 *   attribute's getter method is used to access the data.</td>
 * </tr>
 * <tr>
 *   <td valign="top"><b>AddressableKeyName</b></td><td valign="top">POJO</td><td valign="top">
 *   If this is set, the index will check the attribute names suffixed with value of AddressableKeyName and dot as path separator for key attribute.
 *   Default value for this is "key". All attribute names which do not have this prefix will be treated as value attributes. <BR>
 *   This property will be ignored when using a serializer. 
 * </tr>
 * </table>
 *
 * <P>
 * The following is an ObjectGrid descriptor XML file example that defines the built-in InverseRangeIndex:
 * <Pre>
 *  &lt;?xml version="1.0" encoding="UTF-8"?&gt;
 *  &lt;objectGridConfig xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
 *   xsi:schemaLocation="http://ibm.com/ws/objectgrid/config ../objectGrid.xsd"
 *   xmlns="http://ibm.com/ws/objectgrid/config"&gt;
 *      &lt;objectGrids&gt;
 *          &lt;objectGrid name="inverseRangeIndexSampleGrid"&gt;
 *              &lt;backingMap name="person" pluginCollectionRef="product" readOnly="false"
 *                             preloadMode="false" lockStrategy="OPTIMISTIC" copyMode="COPY_ON_READ_AND_COMMIT" /&gt;
 *          &lt;/objectGrid&gt;
 *      &lt;/objectGrids&gt;
 *      &lt;backingMapPluginCollections&gt;
 *          &lt;backingMapPluginCollection id="product"&gt;
 *              &lt;bean id="MapIndexPlugin" className="com.ibm.websphere.objectgrid.plugins.index.InverseRangeIndex"&gt;
 *                 &lt;property name="Name" type="java.lang.String" value="inverse_range_index" description="index name" /&gt;
 *                 &lt;property name="AttributeName" type="java.lang.String" value="productName[KeyAttribute.productName],promotionDate[KeyAttribute.startPromotionDate ,KeyAttribute.endPromotionDate], RAM[KeyAttribute.minRAM,KeyAttribute.maxRAM], condition[KeyAttribute.condition], KeyAttribute.country" description="attribute name" /&gt;
 *                 &lt;property name="AddressableKeyName" type="java.lang.String" value="KeyAttribute" description="default is key" /&gt;
 *                 &lt;property name="FieldAccessAttribute" type="boolean" value="false" description="default is false" /&gt;
 *              &lt;/bean&gt;
 *          &lt;/backingMapPluginCollection&gt;
 *      &lt;/backingMapPluginCollections&gt;
 *   &lt;/objectGridConfig&gt;
 * </Pre>
 * <p>
 *
 * These properties can also be set programmatic way, using the JavaBean pattern. For example:
 * <pre>
 *      BackingMap productBackingMap = ivObjectGrid.getMap("product");
 *
 *      InverseRangeIndex inverseRangeIndex = new InverseRangeIndex();
 *      inverseRangeIndex.setName("inverse_range_index");
 *      inverseRangeIndex.setAddressableKeyName("KeyAttribute");
 *      inverseRangeIndex.setAttributeName("productName[KeyAttribute.productName],promotionDate[KeyAttribute.startPromotionDate ,KeyAttribute.endPromotionDate], RAM[KeyAttribute.minRAM,KeyAttribute.maxRAM], condition[KeyAttribute.condition], KeyAttribute.country" description="attribute name");
 *      inverseRangeIndex.setFieldAccessAttribute(false);
 *      productBackingMap.addMapIndexPlugin(inverseRangeIndex);
 * </pre>
 * <P>
 *
 * @since WAS XS 8.6
 *
 * @see MapIndexPlugin
 * @see MapIndex
 * @see MapRangeIndex
 */

@PluginOutputFormat(keyFormat=OutputFormat.RAW, valueFormat=OutputFormat.RAW)
public final class InverseRangeIndex implements MapIndexPlugin, BackingMapPlugin {

    final static String CLASS_NAME = InverseRangeIndex.class.getName();

    final static TraceComponent tc = Tr
            .register(CLASS_NAME, Constants.TR_GROUP_NAME, Constants.TR_RESOURCE_BUNDLE_NAME);

    private HashIndex hashIndex = new HashIndex();

    private String attibuteNameString;

    private String[] attributeNamesArray;

    private Integer[] attributeNamesTypeArray;

    private String[] searchAttributeNamesArray;

    private Integer[] searchAttributeNamesTypeArray;

    private String addressableKeyName = MapDataDescriptor.DEFAULT_ADDRESSABLEKEYNAME;
    

    /* (non-Javadoc)
     * @see com.ibm.websphere.objectgrid.lifecycle.BackingMapPlugin#initialize()
     */
    public void initialize() {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.entry(tc, "initialize idx=" + hashIndex.indexName + ", map=" + hashIndex.backingMap.getName());
        }        
        hashIndex.initialize();
        hashIndex.setInverseRangeIndex(this);
        
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.exit(tc, "initialize");
        }
    }

    /* (non-Javadoc)
     * @see com.ibm.websphere.objectgrid.plugins.Initializable#isInitialized()
     */
    public boolean isInitialized() {
        return hashIndex.isInitialized();
    }


    /* (non-Javadoc)
     * @see com.ibm.websphere.objectgrid.lifecycle.BackingMapPlugin#destroy()
     */
    public void destroy() {
        hashIndex.destroy();
    }

    /* (non-Javadoc)
     * @see com.ibm.websphere.objectgrid.plugins.Destroyable#isDestroyed()
     */
    public boolean isDestroyed() {
        return hashIndex.isDestroyed();
    }

    /* (non-Javadoc)
     * @see com.ibm.websphere.objectgrid.plugins.BackingMapPlugin#setBackingMap(com.ibm.websphere.objectgrid.BackingMap)
     */
    public void setBackingMap(BackingMap map) {
        hashIndex.setBackingMap(map);
    }

    /* (non-Javadoc)
     * @see com.ibm.websphere.objectgrid.plugins.BackingMapPlugin#getBackingMap()
     */
    public BackingMap getBackingMap() {
        return hashIndex.getBackingMap();
    }

    /*
     * (non-Javadoc)
     *
     * @see com.ibm.websphere.objectgrid.plugins.index.MapIndexPlugin#setProperties(java.util.Properties)
     */
    public void setProperties(Properties properties) {
       
        setName(properties.getProperty("indexName"));
        setAttributeName(properties.getProperty("attributeName"));
        setAddressableKeyName(properties.getProperty("addressableKeyName"));

        String ivFieldAccessAttributeString = properties.getProperty("fieldAccessAttribute");
        if (ivFieldAccessAttributeString != null && ivFieldAccessAttributeString.equals("true")) {
            setFieldAccessAttribute(true);
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see com.ibm.websphere.objectgrid.plugins.index.MapIndexPlugin#getProperties()
     */
    public Properties getProperties() {

        Properties p = new Properties();

        if (hashIndex.indexName != null) {
            p.put("indexName", hashIndex.indexName );
        }
        
        if (hashIndex.attributeName != null) {
            p.put("attributeName", hashIndex.attributeName);
        }
        
        if (addressableKeyName != null) {
            p.put("addressableKeyName", addressableKeyName );
        }
        
        p.put("fieldAccessAttribute", hashIndex.isFieldAccessAttribute());

        return p;
    }

    /*
     * This is not applicable for InverseRangeIndex. Invoking this method will result in an UnsupportedOperationException.
     * 
     * (non-Javadoc)
     *
     * @see com.ibm.websphere.objectgrid.plugins.index.MapIndexPlugin#setEntityMetadata(com.ibm.websphere.projector.md.EntityMetadata)
     */
    public void setEntityMetadata(EntityMetadata entityMetadata) {
        
        throw new UnsupportedOperationException("setEntityMetadata() is not applicable for InverseRAngeIndex.");

    }

    /*
     * (non-Javadoc)
     *
     * @see com.ibm.websphere.objectgrid.plugins.index.MapIndexPlugin#isRangeIndex()
     */
    public boolean isRangeIndex() {
        return hashIndex.isRangeIndex();
    }

    /**
     * Extracts the attribute value from a specified object. 
     *
     * @param valueOrKey
     *            The value or key object.
     *
     * @throws ObjectGridRuntimeException
     *             is thrown if any exception occurs attempting to extract the attribute value from the value or key Object.
     */
    public Object getAttribute(Object valueOrKey) {
        return hashIndex.getAttribute(valueOrKey);
    }

    /**
     * Retrieves the attribute name or comma-delimited list
     * of attribute names that was set with the
     * {@link #setAttributeName(String)} method.
     *
     * @return the name specified as an argument to the
     *         <code>setAttributeName(String)</code> method or <code>null</code>
     *         if the <code>setAttributeName</code> method was not previously
     *         called or if more than one attribute is defined.
     *
     * @see #setAttributeName(String)
     *
     */
    public String getAttributeName() {
        return hashIndex.getAttributeName();
    }

    /**
     * Sets the comma-delimited values of attribute names to include in the inverse range index.<p>
     * 
     * There are three different type of attributes:
     *  
     * <ul>
     * <li><code>Non-Range attribute</code> - Non-Range attribute that is used as identifier. 
     * <li><code>Simple Range Attribute</code> - A Range Attribute contains boundary values for a given Range. <BR>
     *  For example: min/max or start/finish or begin/end or high/low
     * <li><code>Multi-Range Attribute</code> - An array/List of Range Attributes where each element 
     *  in again an array/List with two boundary values. 
     * </ul>
     * 
     *  The value for <code>attributeName</code> should contain one or more non-range attributes and one or more simple range attributes(or Multi-Range Attribute).
     *  An attribute name should not contain a mixture of simple range attributes and multi-Range attributes.<br><br>
     *  
     * <p>
     * <b>Syntax for AttributeName:</b>
     * <pre>
     * attribute_name_string ::= ({simple_attribute}, {range_attribute}) | ({simple_attribute}, multi_range_attribute) 
     * 
     *      simple_attribute ::= (search_attribute_name, "[", index_attribute_name, "]") | (index_attribute_name);
     *       range_attribute ::= search_attribute_name "[" low_index_attribute_name "," high_index_attribute_name "]";
     * multi_range_attribute ::= [search_attribute_list_name] "[[" index_attribute_list_name "]]";
     * </pre>
     * <p>
     * Attribute names for indexing key:
     * </p>
     * <code>index_attribute_name</code> is non-range attribute name to be part of inverse range index key.<BR><BR>
     * 
     * <code>low_index_attribute_name</code> is low boundary for range attribute name to be part of inverse range index key.<BR><BR>
     * 
     * <code>high_index_attribute_name</code> is high boundary for range attribute name to be part of inverse range index key.<BR><BR>
     * 
     * <code>index_attribute_list_name</code> is attribute name for List/array of range values. Each element in List/array should
     * again be an array/List with two values, which are low and high boundary values for a range.<BR><BR>
     * 
     * <p>
     * Attribute names for search key:
     * </p>
     * <code>search_attribute_name</code> is attribute name to be part of inverse range search key. This is required for
     * range_attribute and optional for simple_attribute. When not specified for simple_attribute, index_attribute_name
     * will be used. <BR><BR>
     * 
     * <code>search_attribute_list_name</code> is attribute name of List/array type, to be part of inverse range search key. This is
     * optional for multi_range_attribute. When not specified, index_attribute_list_name will be used.<BR><BR>
     * 
     * <p><b> Example with Simple Range Attribute:</b></p>

     *  <code>productName[key.productName],promotionDate[key.startPromotionDate ,key.endPromotionDate], RAM[key.minRAM,key.maxRAM], condition[key.condition], key.country</code> <BR><BR>
     *  
     *  In the above example:<BR>
     *   <ul><li><code>productName, condition</code> and <code>country</code> are Non-Range Attributes looked up in the key and same names will be used for search key.<br>
     *   <li><code>startPromotionDate</code> and <code>endPromotionDate</code> are read from the key and treated as one Simple Range Attribute. <code>promotionDate</code> will be read from the search key for findAll(Object searchKey) operation.
     *   <li><code>minRAM</code> and <code>maxRAM</code> are read from the key and treated as one Simple Range Attribute. <code>RAM</code> will be read from the search key for findAll(Object searchKey) operation.<br><BR>
     *   </ul>
     *   
     * <p><b> Example with Multi-Range Attribute:</b></p>
     * 
     *  <code>identifier[key.identifier],rangeValues[[key.rangeValues]]</code><BR><BR>
     *  
     *  In the above example:
     *   <ul><li><code>identifier</code> is a Non-Range Attribute and same name will be used  for the search key
     *   <li><code>rangeValues</code> is treated as a Multi-Range Attribute. Which can be an array/List of elements and each 
     *   element should be an array/List with two boundary values. Same name will be used for search key. In search key each element 
     *   in array/List will contain only one value for search instead of two boundary values.<BR><BR>
     *   </ul>
     * Note that the <code>AddressableKeyName</code> property is only applicable for indexing key attribute 
     * names and not for search key attributes.<BR><BR>  
     * 
     * The {@linkplain #setFieldAccessAttribute(boolean)} identifies how to access the cache object:  
     * by field or by getter method. <p><BR>
     * 
     * 
     *    
     * @see MapIndexPlugin#setAttributeName(String)
     */

    public void setAttributeName(String attributeName) {

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.entry(tc, "setAttributeName - input value is " + attributeName);
        }        
        
        hashIndex.setAttributeName(attributeName);

        this.attibuteNameString = attributeName;

        hashIndex.isComposite = true;

        AttributeNamesParser parser = null;

        try {

            parser = new AttributeNamesParser(attributeName, true);

        } catch (ObjectGridException e) {

            FFDCFilter.processException(e, CLASS_NAME + ".setAttributeName", "700");
            if (tc.isErrorEnabled()) {
                Tr.error(tc,
                        "Exception in processing attributeName string for InverseRangeIndex. InverseRangeIndedex will not work. "
                                + ": " + e.toString());
            }
            throw new ObjectGridRuntimeException("Exception in parsing attributeName string for InverseRangeIndex. "
                    + e.getMessage());

        }

        this.attributeNamesArray = parser.getAttributeNames();
        this.searchAttributeNamesArray = parser.getSearchAttributeNames();
        this.attributeNamesTypeArray = parser.getAttributeNamesType();
        this.searchAttributeNamesTypeArray = parser.getSearchAttributeNamesType();

        hashIndex.attributeName = attributeName;
        hashIndex.attributeNamesArray = this.attributeNamesArray;
        hashIndex.originalAttributeNamesArray = this.attributeNamesArray;

        hashIndex.searchAttributeNamesArray = this.searchAttributeNamesArray;
        hashIndex.originalSearchAttributeNamesArray = this.searchAttributeNamesArray;

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.exit(tc, "setAttributeName - attributeNamesArray : " + Arrays.toString(attributeNamesArray) + 
                    " searchAttributeNamesArray : " + Arrays.toString(searchAttributeNamesArray));
        }        
        
    }

    /**
     * Setter for the addressableKeyName property.<p>
     * 
     * default value is "key".
     *
     * @param addressableKeyName
     */
    public void setAddressableKeyName(String addressableKeyName) {
        this.addressableKeyName = addressableKeyName;
    }

    /**
     * return the value for the addressableKeyName property.<p>
     * 
     * @return addressableKeyName
     */
    public String getAddressableKeyName() {
        return this.addressableKeyName;
    }
    

    /**
     * Setter for the fieldAccessAttribute property.<p>
     * 
     * This property does not apply BackingMaps that have an EntityManager API entity associated or if the 
     * BackingMap has a {@linkplain MapSerializerPlugin} configured.
     *
     * @param isFieldAccessAttribute
     */
    public void setFieldAccessAttribute(boolean isFieldAccessAttribute) {
        hashIndex.setFieldAccessAttribute(isFieldAccessAttribute);
    }

    /**
     * return the value of FieldAccessAttribute property
     *
     * @return fieldAccessAttribute property
     */
    public boolean isFieldAccessAttribute() {
        return hashIndex.isFieldAccessAttribute();
    }

    
    /*
     * (non-Javadoc)
     *
     * @see com.ibm.websphere.objectgrid.plugins.index.MapIndexPlugin#getName()
     */
    public String getName() {
        return hashIndex.getName();
    }

    /**
     * Sets the name of this index.
     *
     * @param indexName
     *            the name of the index
     */
    public void setName(String indexName) {
        hashIndex.setName(indexName);
    }

    /**
     * @see MapIndexPlugin#getIndexProxy(MapIndexInfo)
     */
    public Object getIndexProxy(MapIndexInfo map) {
        return hashIndex.getIndexProxy(map);
    }

    /**
     * @see MapIndexPlugin#doBatchUpdate(TxID,LogSequence)
     */
    public void doBatchUpdate(TxID txid, LogSequence sequence) throws ObjectGridRuntimeException {
        hashIndex.doBatchUpdate(txid, sequence);
    }

    /**
     * @see MapIndexPlugin#undoBatchUpdate(TxID,LogSequence)
     */
    public void undoBatchUpdate(TxID txid, LogSequence sequence) throws ObjectGridException {
        hashIndex.undoBatchUpdate(txid, sequence);
    }

    /*
     * This gets inverse range key as an Object Array with identifier portion of key as first element and
     * CompositeRangeAttribute as second element.
     */

    Object[] getInverseRangeKey(Object compositeValue) {

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.entry(tc, "getInverseRangeKey");
        }        
        
        Object[] compositeAttributeValues = (Object[]) compositeValue;
        ArrayList<Object> identifier = new ArrayList<Object>();
        ArrayList<Object> rangeValues = new ArrayList<Object>();

        for (int i = 0; i < compositeAttributeValues.length; i++) {
            if (attributeNamesTypeArray[i] == InverseRangeConstants.SIMPLE_ATTRIBUTE) {
                identifier.add(compositeAttributeValues[i]);
            } else if (attributeNamesTypeArray[i] == InverseRangeConstants.RANGE_ATTRIBUTE) {
                if (compositeAttributeValues[i] == null & compositeAttributeValues[i + 1] == null) {
                    rangeValues.add(null);
                } else {
                    RangeAttribute rangeAttribute = null;

                    try {
                        rangeAttribute = new RangeAttribute(compositeAttributeValues[i],
                                compositeAttributeValues[i + 1]);
                    } catch (Throwable e) {
                        FFDCFilter.processException(e, CLASS_NAME + ".getInverseRangeKey", "701");
                        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                            Tr.debug(
                                    tc,
                                    "Exception in evaluating attribute value for InverseRangeIndex. " + ": "
                                            + e.getMessage());
                        }
                        throw new ObjectGridRuntimeException(
                                "Exception in evaluating attribute value for InverseRangeIndex. " + e.getMessage());
                    }

                    rangeValues.add(rangeAttribute);
                }
                i++;
            } else if (attributeNamesTypeArray[i] == InverseRangeConstants.MULTI_RANGE_ATTRIBUTES) {
                Object[] attributeList = null;

                if (compositeAttributeValues[i].getClass().isArray()) {
                    attributeList = (Object[]) compositeAttributeValues[i];
                } else if (compositeAttributeValues[i] instanceof List) {
                    attributeList = ((List) compositeAttributeValues[i]).toArray();
                } else {
                    throw new ObjectGridRuntimeException(
                            "Multi range attributes must be passed as a java.util.List or an array. Attribute \""
                                    + attributeNamesArray[i]
                                    + "\" defined as a multi range style attribute in configuration. But, passed in as "
                                    + compositeAttributeValues[i].getClass().getName() + ".");
                }

                for (int j = 0; j < attributeList.length; j++) {
                    Object[] values = null;
                    
                    if(attributeList[j].getClass().isArray()) values = (Object[]) attributeList[j];
                    else if (attributeList[j] instanceof List) values = ((List)attributeList[j]).toArray();
                    else throw new ObjectGridRuntimeException(
                            "A range attribute must be passed as a java.util.List or an array. Range attribute \""
                                    + attributeList[j] + "passed in as "
                                    + attributeList[j].getClass().getName() + ".");
 

                    if (values.length != 2) {
                        throw new ObjectGridRuntimeException(
                                "Elements in muti range attribute must be a java.util.List or array with two entries. First entry represents lower limit where as second entry represents upper limit for a range style attribute.");
                    }

                    if (values[0] == null & values[1] == null) {
                        rangeValues.add(null);
                    } else {
                        RangeAttribute rangeAttribute = null;

                        try {
                            rangeAttribute = new RangeAttribute(values[0], values[1]);
                        } catch (Throwable e) {
                            FFDCFilter.processException(e, CLASS_NAME + ".getInverseRangeKey", "702");
                            if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                                Tr.debug(tc, "Exception in evaluating attribute value for InverseRangeIndex. " + ": "
                                        + e.getMessage());
                            }
                            throw new ObjectGridRuntimeException(
                                    "Exception in evaluating attribute value for InverseRangeIndex. " + e.getMessage());
                        }
                        rangeValues.add(rangeAttribute);
                    }
                }
            }
        }

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.exit(tc, "getInverseRangeKey - " + " identifier : " + identifier + " compositeRangeAttributes " + rangeValues.toString());
        }        
        
        return new Object[] { identifier, new CompositeRangeAttribute(rangeValues.toArray()) };
    }

    /*
     * 
     * returns the InverseRangeSearchKey which contains identifier and CompositeSeatchAttribute
     */
    InverseRangeSearchKey getInverseRangeSearchKey(Object compositeValue) {

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.entry(tc, "InverseRangeSearchKey");
        }        
        
        Object[] compositeAttributeValues = (Object[]) compositeValue;
        ArrayList<Object> identifier = new ArrayList<Object>();
        ArrayList<Object> rangeValues = new ArrayList<Object>();

        for (int i = 0; i < compositeAttributeValues.length; i++) {
            if (searchAttributeNamesTypeArray[i] == InverseRangeConstants.SIMPLE_ATTRIBUTE) {
                identifier.add(compositeAttributeValues[i]);
            } else if (searchAttributeNamesTypeArray[i] == InverseRangeConstants.RANGE_ATTRIBUTE) {
                rangeValues.add(compositeAttributeValues[i]);
            } else if (attributeNamesTypeArray[i] == InverseRangeConstants.MULTI_RANGE_ATTRIBUTES) {

                Object[] attributeList = null;

                if (compositeAttributeValues[i].getClass().isArray()) {
                    attributeList = (Object[]) compositeAttributeValues[i];
                } else if (compositeAttributeValues[i] instanceof List) {
                    attributeList = ((List) compositeAttributeValues[i]).toArray();
                } else {
                    throw new ObjectGridRuntimeException(
                            "Multi range attributes must be passed as a java.util.List or an array. Attribute \""
                                    + attributeNamesArray[i]
                                    + "\" defined as a multi range style attribute in configuration. But, passed in as "
                                    + compositeAttributeValues[i].getClass().getName() + ".");
                }

                for (int j = 0; j < attributeList.length; j++) {
                    rangeValues.add(attributeList[j]);
                }
            }
        }

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.exit(tc, "getInverseSearchRangeKey - " + " identifier : " + identifier + " compositeRangeSearchAttributes " + rangeValues.toString());
        } 
        
        return new InverseRangeSearchKey(identifier, new CompositeRangeSearchAttribute(rangeValues.toArray()));
    }

}
