// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005 - 2007
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  MapIndex.java
//
// Source File Description:
//
//    Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M2_B.2  WASX.XD   02/23/2005   newport   new code drop
// 271122       WASX.XD   04/28/2005   kjlaw     add FinderException to findAll method.
// 285430       WASX.XD   06/21/2005   pdykes    Pre-GM changes to remove non-supported public interfaces
// b1153        XD6.1     05/30/2006   cheng1    B25: Indexing enhancement
// b1153        XD6.1     06/12/2006   cheng1    B25: add findAll() to return all keys
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.plugins.index;

import java.io.InvalidObjectException;
import java.io.ObjectStreamException;
import java.io.Serializable;
import java.util.Iterator;

import com.ibm.websphere.objectgrid.ObjectMap;
import com.ibm.websphere.objectgrid.OutputFormat;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.SerializedEntry;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.SerializedKey;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.SerializedValue;

/**
 * This interface is an abstract index that can be created on a Map. The index can be used to perform efficient look ups
 * and possibly other operations such as relational operations on an attribute in a Map. The MapIndex is provided with
 * all update events and maintains an index that can be used to issue simple queries against the index later. The index
 * could use an index defined callback to make an index on composite attributes.
 *
 * @since WAS XD 6.0.1
 *
 * @see ObjectMap#getIndex(String)
 */
public interface MapIndex {


    /**
     * This class represents directives for finding types of values in an index.<p>
     *
     * @since 7.0
     */
    public static final class SpecialValue implements Serializable {
        private static final long serialVersionUID = -5996516491414025549L;
        static SpecialValue EMPTY_VALUE_INSTANCE = new SpecialValue((byte)0);
        private byte ivValueType;

        private SpecialValue(byte valueType) {
            ivValueType = valueType;
        }

        /**
         * Returns a string representation of the <code>SpecialValue</code>.
         *
         * @return a string representation of the <code>SpecialValue</code>.
         */
        public String toString() {
            if(ivValueType == 0) {
                return "EMPTY_VALUE";
            } else {
                return "UNKNOWN: " + ivValueType; //$NON-NLS-1$
            }
        }

        private Object readResolve() throws ObjectStreamException {
            if (ivValueType == 0) {
                return EMPTY_VALUE;
            }
            throw new InvalidObjectException("Unknown MapIndex.SpecialValue");
        }

        public int hashCode() {
            final int prime = 31;
            int result = 1;
            result = prime * result + ivValueType;
            return result;
        }

        public boolean equals(Object obj) {
            if (this == obj)
                return true;
            if (obj == null)
                return false;
            if (getClass() != obj.getClass())
                return false;
            final SpecialValue other = (SpecialValue) obj;
            if (ivValueType != other.ivValueType)
                return false;
            return true;
        }

    }

    /**
     * A SpecialValue constant that indicates a particular attribute value should
     * be excluded from an index query.
     *
     * @since 7.0
     */
    public static final SpecialValue EMPTY_VALUE = SpecialValue.EMPTY_VALUE_INSTANCE;

    /**
     * Returns the keys for the entries that have the specified attribute value.<p>
     * 
     * For maps that are configured to have {@linkplain ObjectMap#getKeyOutputFormat()} set to {@linkplain OutputFormat#RAW}, 
     * the keys returned will be {@link SerializedKey} objects. If required, you can use the {@link SerializedEntry#getObject()} 
     * method to retrieve (possibly inflating the serialized object) the original key or value object.
     *
     * If the index is a composite index
     * (having more than one attribute), the attributeValue parameter must be the type of Object[] of the same size as
     * the number of attributes defined in the index. Each attribute value element in the Object array is corresponding
     * to the list of attribute names defined for the composite index. Use constant value {@link #EMPTY_VALUE}
     * to indicate a particular attribute value to be excluded from value evaluation during index searching.<p>
     *
     * For example, if there is a composite index that has attributes defined:  state,city,zip.<p>
     *
     * To find all keys that match a specific state, city and zip:<br>
     * <code>
     * Iterator result = idx.findAll(new Object[]{"MN", "Rochester", "55901"});
     * </code><p>
     *
     * To find all keys that match only the state:<br>
     * <code>
     * Iterator result = idx.findAll(new Object[]{"MN", MapIndex.EMPTY_VALUE, MapIndex.EMPTY_VALUE});
     * </code>
     *
     * @param attributeValue
     *            a reference to the attribute value to search for. In composite index case, this must be the type of
     *            Object[].
     *
     * @return the list of the keys for the entries with an attribute equal to the specified attribute value.
     *
     * @throws FinderException
     *             if an exception or retry limit is reached when concurrent transactions updating the index prevent
     *             findAll from completing.
     * @throws IllegalArgumentException
     *             For a composite index, the IllegalArgumentException will be thrown in each of the following
     *             conditions: If the attributeValue is null or is not of type Object[], and also if the attributeValue
     *             is of type Object[] but has length unequal to the length of the configured attribute names.
     */
    // TODO should argument be of type Comparable?
    Iterator /* <Object key> */findAll(Object attributeValue) throws FinderException;

    /**
     * Returns the keys of the entire map. Be cautious when executing this method from an ObjectGrid client process. It
     * will have significant impact on network traffic and memory footprint in client process because all keys will be
     * fetched from ObjectGrid servers into client process.
     *
     * For maps that are configured to have {@linkplain ObjectMap#getKeyOutputFormat()} set to {@linkplain OutputFormat#RAW}, 
     * the keys returned will be {@link SerializedKey} objects. If required, you can use the {@link SerializedEntry#getObject()} 
     * method to retrieve (possibly inflating the serialized object) the original key or value object.
     *
     * @return the list of the keys for all entries in the map.
     *
     * @throws FinderException
     *             if an exception or retry limit is reached when concurrent transactions updating the index prevent
     *             findAll from completing.
     * @since WAS XD 6.1
     */
    Iterator /* <Object key> */findAll() throws FinderException;
}
