//IBM Confidential OCO Source Material
//5724-J34 (C) COPYRIGHT International Business Machines Corp. 2012
//The source code for this program is not published or otherwise divested
//of its trade secrets, irrespective of what has been deposited with the
//U.S. Copyright Office.
//
//$Id$
//
//Module  :  PartitionKey.java
//
//Source File Description:
//
// PartitionKey annotation definition
//
package com.ibm.websphere.objectgrid.plugins.io.annotations;

import static java.lang.annotation.ElementType.FIELD;
import static java.lang.annotation.ElementType.TYPE;
import static java.lang.annotation.RetentionPolicy.RUNTIME;

import java.lang.annotation.Retention;
import java.lang.annotation.Target;

/**
 * Specifies a single attribute to use to calculate the partition hash code.<p>
 * 
 * (Note: this annotation is currently only used when eXtreme scale data format (XDF) is enabled)<br>
 *
 * The &#64;PartitionKey annotation can be specified for the class using a path
 * syntax to identify a single attribute.  Multiple attributes can be specified using
 * additional &#64;PartitionKey annotations on each field, using the {@link #order()} attribute
 * to specify the order in which the hash codes will be calculated.<p>
 * <p>
 * <b>The PartitionKey is not inheritable.</b>
 * </p>
 * Use the {@link PartitionKeys} annotation
 * to specify more complex attribute combinations at the class level.<p>
 *
 * The following examples illustrate how to identify a top-level, primitive field:<br>
 *
 * <strong>Simple class annotation example:</strong>
 * <pre>
 * &#64;PartitionKey("deptId")
 * class Employee {
 *     int empId;
 *     int deptId;
 * }
 * </pre>
 *
 * <strong>Simple field annotation example:</strong>
 * <pre>
 * class Employee {
 *     int empId;
 *
 *     &#64;PartitionKey
 *     int deptId;
 * }
 * </pre>
 *
 * <strong>Simple, multiple field annotation example:</strong>
 * <pre>
 * class Employee {
 *     int empId;
 *
 *     &#64;PartitionKey(order=0)
 *     int deptId;
 *
 *     &#64;PartitionKey(order=1)
 *     String country;
 * }
 * </pre>
 *
 *
 * The following examples illustrate how to address an attribute inside an embedded class.  In this case, the
 * path separator is defined in the data grid as a "." character:<br>
 * <strong>Embedded class annotation example:</strong>
 * <pre>
 * &#64;PartitionKey("deptKey.id")
 * class Employee {
 *     DepartmentKey deptKey;
 * }
 *
 * class DepartmentKey {
 *     int id;
 * }
 * </pre>
 *
 * <strong>Embedded field annotation example:</strong>
 * <pre>
 * class Employee {
 *     &#64;PartitionKey("id")
 *     DepartmentKey deptKey;
 * }
 *
 * class DepartmentKey {
 *     int id;
 * }
 * </pre>
 *
 *
 * @since 8.6
 */
@Target({ TYPE, FIELD })
@Retention(RUNTIME)
public @interface PartitionKey {

    /**
     * Identifies the path to the attribute that should be included as part of the partition key.
     * @return the attribute path.
     */
    String value() default "";

    /**
     * Order of the fields that contribute to the partition calculation.
     *
     * @return the field order
     */
    short order() default 0;

}
