// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2009
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// 1.8 XD/ws/code/objectgrid/src/com/ibm/websphere/objectgrid/security/MapPermission.java, WAS.objectgrid, WASX.XD, b0614.107 2/14/06 17:24:01
//
// Module  :  MapPermission.java
//
// Source File Description:
//
// Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M3_B.13 WASX.XD   03/01/2005   jitang    ObjectGrid - JAAS / Security - Drop 1 - Basic Support
// 344949       WXD601.XD 02/07/2006   jhanders  Changes for review from javadoc updates
// b14971       6.1.0.5   02/09/2009   jian.tang Make MapPermission accept OG name or map name with "." in it
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.security;

import java.security.Permission;
import java.util.StringTokenizer;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.ObjectMap;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.security.PermissionHelper;

/**
 * This class represents permissions to the ObjectMap. It has five different
 * actions:
 * </p>
 * <ul>
 * <li>READ</li>
 * <li>WRITE</li>
 * <li>INSERT</li>
 * <li>REMOVE</li>
 * <li>INVALIDATE</li>
 * </ul>
 *
 * Please refer to the <code>ObjectMap</code> class for the permissions
 * needed for each method.
 *
 * @since WAS XD 6.0
 *
 * @see Permission
 * @see ObjectMap
 */
public final class MapPermission extends Permission {
    private static final long serialVersionUID = -7144398636300360859L;

    private final static String CLASS_NAME = MapPermission.class.getName();

    private final static boolean DEBUG = false;

    private final static TraceComponent TC = Tr.register(
            CLASS_NAME,
            Constants.TR_SECURITY_GROUP_NAME,
            Constants.TR_RESOURCE_BUNDLE_NAME);

    /**
     * mask indicating the permissions
     */
    private int mask = 0;

    /**
     * The total number of possible Permissions
     */
    public static final int NUMBER_PERMISSION = 5;

    /**
     * Read permission constant
     */
    public static final int READ = 0x01;

    /**
     * Write permission constant
     */
    public static final int WRITE = 0x02;

    /**
     * Insert permission constant
     */
    public static final int INSERT = 0x04;

    /**
     * Remove permission constant
     */
    public static final int REMOVE = 0x08;

    /**
     * Invalidate permission constant
     */
    public static final int INVALIDATE = 0x10;

    /**
     * An array with the string representation of each permission.
     * <p>
     * The indexes to the String representation are the <code>int</code>
     * constants for each permission (i.e. READ, WRITE, INSERT, REMOVE,
     * INVALIDATE)
     */
    public static final String[] PERMISSIONS = new String[INVALIDATE + 1];

    /**
     * All permissions String
     */
    public static final String ALL_PERMISSION = "all";

    /**
     * All permission constant
     */
    public static final int ALL = READ | WRITE | INSERT | REMOVE | INVALIDATE;

    /**
     * parsed names
     * */
    private String[] names = new String[2];

    /**
     * Whether the name has been parsed into ObjectGrid and map name yet.
     */
    private boolean isNameParsed = false;

    static {
        PERMISSIONS[READ] = "read";
        PERMISSIONS[WRITE] = "write";
        PERMISSIONS[INSERT] = "insert";
        PERMISSIONS[REMOVE] = "remove";
        PERMISSIONS[INVALIDATE] = "invalidate";
    }

    /**
     * Constructs a new MapPermission for the named map with the specified
     * actions.
     * <p>
     * The map name for this permission should be in the format of
     * <ObjectGrid name>.<map name>, for example "og1.map1".  A special value
     * of "*" is used to represent all objectgrids or all maps, for example
     * "*.map1", "og1.*", or "*.*". A "*" is used to substitue a whole
     * ObjectGrid name or map name. A partial substitution like "og*.map" is not
     * supported.
     *
     * @param name the full name of the map. It is also the permission name.
     * @param actions
     *            the comma separated list of actions, such as "read", "write",
     *            "remove", "insert", and "invalidate".
     *
     * @throws IllegalArgumentException if name is <code>null</code> or malformed
     *         or if actions is <code>null</code> or contains an invalidate
     *         action name
     */
    public MapPermission(String name, String actions) {
        super(name == null ? "" : name.trim());

        if (DEBUG && TC.isEntryEnabled())
            Tr.entry(TC, "<init>", new Object[] { name, actions });

        if (name == null) {
            throw new IllegalArgumentException("Name cannot be null.");
        }
        if (actions == null) {
            throw new IllegalArgumentException("Actions cannot be null.");
        }
        parseAction(actions);
        isNameParsed = PermissionHelper.parseName(getName(), names);

        if (DEBUG && TC.isEntryEnabled())
            Tr.exit(TC, "<init>", this);
    }

    /**
     * Constructs a new MapPermission for the named map with the specified
     * actions.
     * <p>
     * This method is used for the ObjectGrid map of which the ObjectGrid name
     * or map name contains a ".". In this case, we pass the ObjectGrid name and
     * map name separately as the parameters.
     * <p>
     * ObjectGrid uses &lt;ObjectGrid name&gt;.&lt;map name&gt; as the permission name. Therefore,
     * your application still needs to make sure there is no name conflicts for the
     * &lt;ObjectGrid name&gt;.&lt;map name&gt;. For example, you should not have an ObjectGrid
     * with name "og" and one of its maps with name "session.data", and another
     * ObjectGrid with name "og.session" and one of its maps with name "data".
     *
     * @param ogName the ObjectGrid name.
     * @param mapName the map name.
     * @param actions
     *            the comma separated list of actions, such as "read", "write",
     *            "remove", "insert", and "invalidate".
     *
     * @throws IllegalArgumentException if ObjectGrid name or map name is <code>null</code>
     *         or if actions is <code>null</code> or contains an invalidate
     *         action name
     * @since 6.1.0.5 FIX2
     */
    public MapPermission(String ogName, String mapName, String actions) {
        super((ogName == null ? "" : ogName.trim()) + "." + (mapName == null ? "" : mapName.trim()));

        if (DEBUG && TC.isEntryEnabled())
            Tr.entry(TC, "<init>", new Object[] { ogName, mapName, actions });

        if (ogName == null) {
            throw new IllegalArgumentException("ObjectGrid name cannot be null.");
        }
        if (mapName == null) {
            throw new IllegalArgumentException("ObjectGrid Map name cannot be null.");
        }
        if (actions == null) {
            throw new IllegalArgumentException("Actions cannot be null.");
        }
        parseAction(actions);
        names[0] = ogName;
        names[1] = mapName;
        isNameParsed = true;

        if (DEBUG && TC.isEntryEnabled())
            Tr.exit(TC, "<init>", this);
    }

    /**
     * Constructs a new MapPermission for the named map with the specified
     * actions.
     * <p>
     * The map name for this permission should be in the format of
     * <ObjectGrid name>.<map name>, for example "og1.map1".  A special value
     * of "*" is used to represent all objectgrids or all maps, for example
     * "*.map1", "og1.*", or "*.*". A "*" is used to substitue a whole
     * ObjectGrid name or map name. A pPartial substitution like "og*.map" is not
     * supported.
     *
     * @param name the full name of the map. It is also the permission name.
     * @param actions
     *            the actions mask. For example,
     *            MapPermission.READ|MapPermission.WRITE
     *
     * @throws IllegalArgumentException if name is <code>null</code> or malformed
     *         or if actions is less than 0 or greater than <code>ALL</code>
     */
    public MapPermission(String name, int actions) {
        super(name == null ? "" : name.trim());
        if (DEBUG && TC.isEntryEnabled())
            Tr.entry(TC, "<init>", new Object[] { name, new Integer(actions) });

        if (name == null) {
            throw new IllegalArgumentException("Name cannot be null.");
        }
        parseIntAction(actions);
        isNameParsed = PermissionHelper.parseName(getName(), names);

        if (DEBUG && TC.isEntryEnabled())
            Tr.exit(TC, "<init>", this);
    }

    /**
     * Constructs a new MapPermission for the named map with the specified
     * actions.
     * <p>
     * This method is used for the ObjectGrid map of which the ObjectGrid name
     * or map name contains a ".". In this case, we pass the ObjectGrid name and
     * map name as the parameters.
     * <p>
     * ObjectGrid uses &lt;ObjectGrid name&gt;.&lt;map name&gt; as the permission name. Therefore,
     * your application still needs to make sure there is no name conflicts for the
     * &lt;ObjectGrid name&gt;.&lt;map name&gt;. For example, you should not have an ObjectGrid
     * with name "og" and one of its maps with name "session.data", and another
     * ObjectGrid with name "og.session" and one of its maps with name "data".
     *
     * @param ogName the ObjectGrid name.
     * @param mapName the map name.
     * @param actions
     *            the actions mask. For example,
     *            MapPermission.READ|MapPermission.WRITE
     *
     * @throws IllegalArgumentException if ObjectGrid name or map name is <code>null</code>
     *         or if actions is less than 0 or greater than <code>ALL</code>
     * @since 6.1.0.5 FIX2
     */
    public MapPermission(String ogName, String mapName, int actions) {
        super((ogName == null ? "" : ogName.trim()) + "." + (mapName == null ? "" : mapName.trim()));

        if (DEBUG && TC.isEntryEnabled())
            Tr.entry(TC, "<init>", new Object[] { ogName, mapName, new Integer(actions) });

        if (ogName == null) {
            throw new IllegalArgumentException("ObjectGrid name cannot be null.");
        }
        if (mapName == null) {
            throw new IllegalArgumentException("ObjectGrid Map name cannot be null.");
        }
        parseIntAction(actions);
        names[0] = ogName;
        names[1] = mapName;
        isNameParsed = true;

        if (DEBUG && TC.isEntryEnabled())
            Tr.exit(TC, "<init>", this);
    }

    /**
     * Parses the action string
     *
     * @param action
     *            the action string
     */
    private void parseAction(String action) {
        StringTokenizer st = new StringTokenizer(action, ",");

        mask = 0;
        while (st.hasMoreTokens()) {
            String tok = st.nextToken().trim();
            if (tok.equals(PERMISSIONS[READ])) {
                mask |= READ;
            }
            else if (tok.equals(PERMISSIONS[WRITE])) {
                mask |= WRITE;
            }
            else if (tok.equals(PERMISSIONS[INSERT])) {
                mask |= INSERT;
            }
            else if (tok.equals(PERMISSIONS[REMOVE])) {
                mask |= REMOVE;
            }
            else if (tok.equals(PERMISSIONS[INVALIDATE])) {
                mask |= INVALIDATE;
            }
            else if (tok.equals(ALL_PERMISSION)) {
                mask |= ALL;
            }
            else if (!tok.equals("")) {
                throw new IllegalArgumentException("Unknown action " + tok);
            }
        }
    }

    /**
     * Parses the action integer
     *
     * @param action
     *            the action integer
     */
    private void parseIntAction(int action) {
        if (action > ALL || action < 0) {
            throw new IllegalArgumentException(
            "Action mask value should be between 0 and MapPermission.ALL.");
        }

        if (action == 0) {
            return;
        }
        else if (action == ALL) {
            mask |= ALL;
            return;
        }
        else {
            if ((action & READ) == READ) {
                mask |= READ;
            }
            if ((action & WRITE) == WRITE) {
                mask |= WRITE;
            }
            if ((action & INSERT) == INSERT) {
                mask |= INSERT;
            }
            if ((action & REMOVE) == REMOVE) {
                mask |= REMOVE;
            }
            if ((action & INVALIDATE) == INVALIDATE) {
                mask |= INVALIDATE;
            }
        }
    }

    /**
     * Checks if the specified permission is "implied" by
     * this object.
     * <P>
     * More specifically, this method returns true if:<p>
     * <ul>
     * <li> <i>permission</i>'s class is the same as this object's class, and<p>
     * <li> <i>permission</i>'s name is equal to or (in the case of wildcards)
     *      is implied by this object's
     *      name. For example, "a.*" implies "a.b"., and<p>
     * <li> <i>permission</i>'s actions are equal to or are a subset of
     *         this object's actions
     * </ul>
     *
     * @param permission the permission to check against.
     *
     * @return true if the passed permission is equal to or
     * implied by this permission, false otherwise.
     */
    public boolean implies(Permission permission) {
        if (DEBUG && TC.isEntryEnabled())
            Tr.entry(TC, "implies", new Object[] { this, permission });
        if (!(permission.getClass().getName().equals(CLASS_NAME))) {
            if (DEBUG && TC.isEntryEnabled())
                Tr.exit(TC, "implies", "Not a MapPermission object, return false");
            return false;
        }

        MapPermission p = (MapPermission) permission;

        // The name part must be either the wildcard *, or the name must match
        // our name

        if (!getName().equals(p.getName())) {

            if (isNameParsed) {
                if (p.isNameParsed) {
                    // this  has og name and map name
                    // other has og name and map name
                    String[] otherNames = p.getParsedNames();
                    for (int i = 1; i >= 0; i--) {
                        if (!names[i].equals("*") && !names[i].equals(otherNames[i])) {
                            if (DEBUG && TC.isEntryEnabled())
                                Tr.exit(TC, "implies", "Name not match, return false");
                            return false;
                        }
                    }
                } else {
                    // this  has og name and map name
                    // other doesn't have og name and map name, only has full name.

                    if (names[0].equals("*")) {
                        if (!names[1].equals("*") && !p.getName().endsWith("."+names[1])){
                            return false;
                        }
                    } else {
                        if (names[1].equals("*")) {
                            if (!p.getName().startsWith(names[0] + ".")) {
                                return false;
                            }
                        } else {
                            // since their full name is not equal, must be false.
                            return false;
                        }
                    }
                }
            } else {
                // this doesn't have og name and map name, only has full name, so it doesn't contain *
                // so implies only return true if full name are equal.
                // However, they are not equal, so it is always false.
                return false;
            }
        }

        if ((mask & p.mask) != p.mask) {
            if (DEBUG && TC.isEntryEnabled())
                Tr.exit(TC, "implies", "permission mask not match, return false");
            return false;
        }

        // Only if both the action and name match do we return true.
        if (DEBUG && TC.isEntryEnabled())
            Tr.exit(TC, "implies", "true");
        return true;
    }

    /**
     * Checks two MapPermission objects for equality. Two MapPermission
     * objects are equal if and only if their names and actions are equal.
     * <P>
     * Do not use the <code>equals</code> method for making access control
     * decisions; use the <code>implies</code> method.
     *
     * @param o the object we are testing for equality with this object.
     *
     * @return <code>true</code> if both MapPermission objects are equivalent.
     */
    public boolean equals(Object o) {
        if (!(o.getClass().getName().equals(CLASS_NAME)))
            return false;

        // check the name and action mask for deep equality
        MapPermission p = (MapPermission) o;
        return ((p.getName().trim().equals(getName().trim())) && (p.mask == mask));
    }

    /**
     * Returns the hash code value for this MapPermission object.
     *
     * @return a hash code value for this object.
     */
    public int hashCode() {
        // a hash code for permissions,
        return getName().hashCode() ^ mask;
    }

    /**
     * Returns the actions as a String.
     *
     * @return the actions of this Permission.
     */
    public String getActions() {
        return getActions(mask);
    }

    /**
     * Returns the actions as a String for the specified mask.
     *
     * @param m the mask
     *
     * @return the action string representation for the mask
     */
    public static String getActions(int m) {
        if (m > ALL || m < 0) {
            throw new IllegalArgumentException(
            "Action mask should be between 0 and MapPermission.ALL.");
        }
        if (m==0) {
            return "";
        }

        int index = 1;
        boolean isCommaNeeded = false;
        StringBuffer sb = new StringBuffer(32);

        while (index <= INVALIDATE) {
            if ((index & m) == index) {
                if (isCommaNeeded) {
                    sb.append(", ");
                }
                else {
                    isCommaNeeded = true;
                }
                sb.append(PERMISSIONS[index]);
            }
            index *= 2;
        }

        return new String(sb);
    }

    /**
     * Returns an array which contains the object grid name and map name.
     *
     * @return the parsed name array
     */
    public String[] getParsedNames() {
        return names;
    }

    /**
     * Returns an int mask value which represents the permission's actions.
     * <p>
     * For example, for a MapPermission which has read and insert permissions,
     * the returned value will be 5 (MapPermission.READ|MapPermission.INSERT).
     *
     * @return an int value which represents the permission.
     */
    public int getActionsInInt() {
        return mask;
    }
}