// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005-2007
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $HeadURL$
// $Id$
//
// Module  :  KeyStoreLoginAuthenticator.java
//
// Source File Description: 
//
//  Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M3_B.13 WASX.XD   04/05/2005   jitang    ObjectGrid - JAAS / Security - Drop 1 - Basic Support
// b10193       WASX.XD   10/16/2007   jitang    Add doPriv block
//
// ------------ ---------- ------------ -------------------------------------------------
package com.ibm.websphere.objectgrid.security.plugins.builtins;

import java.security.AccessController;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;

import javax.security.auth.Subject;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.security.plugins.Authenticator;
import com.ibm.websphere.objectgrid.security.plugins.Credential;
import com.ibm.websphere.objectgrid.security.plugins.ExpiredCredentialException;
import com.ibm.websphere.objectgrid.security.plugins.InvalidCredentialException;
import com.ibm.ws.ffdc.FFDCFilter;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.ObjectGridManagerImpl;
import com.ibm.ws.objectgrid.security.auth.callback.UserPasswordCallbackHandlerImpl;

/**
 * This class is an implementation of the <code>Authenticator</code> interface 
 * when a user name and password are used as a credential.
 * <p>
 * When user ID and password authentication is used, the credential passed to the 
 * <code>authenticate(Credential)</code> method is a UserPasswordCredential object. 
 * <p>
 * This implementation will use a <code>KeyStoreLoginModule</code> to authenticate 
 * the user into the key store using the JAAS login module "KeyStoreLogin". The key 
 * store can be configured as an option to the <code>KeyStoreLoginModule</code> 
 * class. Please see the <code>KeyStoreLoginModule</code> class for more details 
 * about how to set up the JAAS login configuration file.
 * <p>
 * This class is fully supported for production use.
 *
 * @since WAS XD 6.0.1
 * 
 * @see Authenticator
 * @see KeyStoreLoginModule
 * @see UserPasswordCredential
 */
public class KeyStoreLoginAuthenticator implements Authenticator {

    static final String CLASS_NAME = KeyStoreLoginAuthenticator.class.getName();

    static final TraceComponent TC = Tr.register(
            CLASS_NAME,
            Constants.TR_SECURITY_GROUP_NAME,
            Constants.TR_RESOURCE_BUNDLE_NAME);

    /**
     * Creates a new KeyStoreLoginAuthenticator.
     */
    public KeyStoreLoginAuthenticator() {
        super();
    }

    /**
     * Authenticates a <code>UserPasswordCredential</code>.
     * <p>
     * Uses the user name and password from the specified UserPasswordCredential
     * to login to the KeyStoreLoginModule named "KeyStoreLogin".
     * 
     * @throws InvalidCredentialException if credential isn't a 
     *         UserPasswordCredential or some error occurs during processing
     *         of the supplied UserPasswordCredential
     * 
     * @throws ExpiredCredentialException if credential is expired.  This exception
     *         is not used by this implementation 
     * 
     * @see Authenticator#authenticate(Credential)
     * @see KeyStoreLoginModule
     */
    public Subject authenticate(final Credential credential) throws InvalidCredentialException, ExpiredCredentialException {
        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
            Tr.entry(TC, "authenticate", credential);

        if (credential == null) {
            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.exit(
                    TC,
                    "authenticate",
                    "InvalidCredentialException(null credential)");
            throw new InvalidCredentialException("Supplied credential is null");
        }

        if (! (credential instanceof UserPasswordCredential) ) {
            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.exit(TC, "authenticate", "InvalidCredentialException(wrong credential type)");
            throw new InvalidCredentialException("Supplied credential is not a UserPasswordCredential");
        }

        Subject subject;
        try {
            subject = (Subject) AccessController.doPrivileged(new PrivilegedExceptionAction() {
                public Object run() throws InvalidCredentialException, ExpiredCredentialException {
                    UserPasswordCredential cred = (UserPasswordCredential) credential;
                    LoginContext lc = null;
                    try {
                        lc = new LoginContext("KeyStoreLogin",
                                new UserPasswordCallbackHandlerImpl(cred.getUserName(), cred.getPassword().toCharArray()));

                        if (ObjectGridManagerImpl.isTraceEnabled && TC.isDebugEnabled())
                            Tr.debug(TC, "authenticate", "lc="+lc);

                        lc.login();

                        Subject s = lc.getSubject();

                        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                            Tr.exit(TC, "authenticate", s);
                        return s;
                    }
                    catch (LoginException le) {
                        FFDCFilter.processException(
                                le,
                                CLASS_NAME + ".authenticate",
                                "120",
                                this);
                        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                            Tr.exit(TC, "authenticate", "LoginException");
                        throw new InvalidCredentialException(le);
                    }
                    catch (IllegalArgumentException ile) {
                        FFDCFilter.processException(
                                ile,
                                CLASS_NAME + ".authenticate",
                                "121",
                                this);
                        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                            Tr.exit(TC, "authenticate", "IllegalArgumentException");
                        throw new InvalidCredentialException(ile);
                    }
                            }
            });
        } catch (PrivilegedActionException e) {
            Throwable t = e.getCause();
            
            if (t instanceof InvalidCredentialException) {
                throw (InvalidCredentialException) t;
            }
            else if (t instanceof ExpiredCredentialException) {
                throw (ExpiredCredentialException) t;
            }
            else {
                throw new RuntimeException(e);
            }
        }
        
        return subject;
    }

}