// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2006
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// 1.9 XD/ws/code/objectgrid.impl/src/com/ibm/websphere/objectgrid/security/plugins/builtins/KeyStoreLoginModule.java, WAS.objectgrid, WASX.XD, b0614.107 2/14/06 19:06:01
//
// Module  :  KeyStoreLoginModule.java
//
// Source File Description: 
//
//    Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M3_B.13 WASX.XD   03/21/2005   jitang    ObjectGrid - JAAS / Security - Drop 1 - Basic Support
//   
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.security.plugins.builtins;

import java.io.File;
import java.io.FileInputStream;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.UnrecoverableKeyException;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.NameCallback;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;
import javax.security.auth.x500.X500Principal;
import javax.security.auth.x500.X500PrivateCredential;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.ObjectGridRuntimeException;
import com.ibm.ws.ffdc.FFDCFilter;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.ObjectGridManagerImpl;
import com.ibm.ws.objectgrid.util.ObjectGridUtil;

/**
 * The KeyStoreLoginModule is a keystore authentication login module based on 
 * JAAS authentication.
 * <p>
 * A login configuration should provide an option "<code>keyStoreFile</code>" to 
 * indicate where the keystore file is located. If the <code>keyStoreFile</code> 
 * value contains a system property in the form, <code>${system.property}</code>,
 * it will be expanded to the value of the system property.
 * <p>
 * If an option "<code>keyStoreFile</code>" is not provided, the default keystore
 * file name is <code>"${java.home}${/}.keystore"</code>.
 * <p>
 * Here is a Login module configuration example: 
 * <pre><code>
 *    KeyStoreLogin {
 *        com.ibm.websphere.objectgrid.security.plugins.builtins.KeystoreLoginModule required
 *            keyStoreFile="${user.dir}${/}security${/}.keystore";
 *    };
 * </code></pre>
 * 
 * This login module returns a principal of class javax.security.auth.x500.X500Principal.
 * 
 * @since WAS XD 6.0.1  
 * 
 * @see LoginModule
 */
public class KeyStoreLoginModule implements LoginModule {

    private static final String CLASS_NAME = KeyStoreLoginModule.class.getName();

    private static final TraceComponent TC = Tr.register(
            CLASS_NAME,
            Constants.TR_SECURITY_GROUP_NAME,
            Constants.TR_RESOURCE_BUNDLE_NAME);

    /** 
     * Key store file property name 
     */
    public static final String KEY_STORE_FILE_PROPERTY_NAME = "keyStoreFile";

    /**
     * Key store type. Only JKS is supported 
     */
    public static final String KEYSTORE_TYPE = "JKS";

    /**
     * The default key store file name 
     */
    public static final String DEFAULT_KEY_STORE_FILE = "${java.home}${/}.keystore";

    private CallbackHandler handler;

    private Subject subject;

    private boolean debug = false;

    private Set principals = new HashSet();

    private Set publicCreds = new HashSet();

    private Set privateCreds = new HashSet();

    protected KeyStore keyStore;

    /**
     * Creates a new KeyStoreLoginModule.
     */
    public KeyStoreLoginModule() {
        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled()) {
            Tr.entry(TC, "<init>");
            Tr.exit(TC, "<init>", this);
        }
    }

    /**
     * Initializes the login module.
     * 
     * @see LoginModule#initialize(Subject, CallbackHandler, Map, Map)
     */
    public void initialize(Subject sub, CallbackHandler callbackHandler,
            Map mapSharedState, Map mapOptions) {

        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
            Tr.entry(TC, "initialize", new Object[] { this, sub, callbackHandler,
                    mapSharedState, mapOptions });

        // initialize any configured options
        debug = "true".equalsIgnoreCase((String) mapOptions.get("debug"));

        if (sub == null)
            throw new IllegalArgumentException("Subject is not specified"); 

        if (callbackHandler == null)
            throw new IllegalArgumentException(
            "CallbackHander is not specified"); 

        // Get the key store path
        String sKeyStorePath = (String) mapOptions
            .get(KEY_STORE_FILE_PROPERTY_NAME);

        // If there is no key store path, the default one is the .keystore
        // file in the java home directory
        if (sKeyStorePath == null) {
            sKeyStorePath = DEFAULT_KEY_STORE_FILE;
        }

        // Replace the system enviroment variable
        sKeyStorePath = ObjectGridUtil.replaceVar(sKeyStorePath, TC);

        File fileKeyStore = new File(sKeyStorePath);
        FileInputStream fisFileKeyStore=null;
        try {
            KeyStore store = KeyStore.getInstance("JKS");
            store.load(fisFileKeyStore=new FileInputStream(fileKeyStore), null);

            // Save the key store
            keyStore = store;

            if (debug) {
                System.out.println("[KeyStoreLoginModule] initialize: Successfully loaded key store");
            }
            if (ObjectGridManagerImpl.isTraceEnabled && TC.isDebugEnabled())
                Tr.debug(TC, "initialize", "Successfully loaded key store " + sKeyStorePath);
        }
        catch (Exception e) {
            FFDCFilter.processException(
                    e,
                    CLASS_NAME + ".initialize",
                    "173",
                    this);

            ObjectGridRuntimeException re = new ObjectGridRuntimeException(
                    "Failed to load keystore: " + fileKeyStore.getAbsolutePath()); 
            re.initCause(e);
            if (debug) {
                System.out.println("[KeyStoreLoginModule] initialize: Key store loading failed with exception "
                        + e.getMessage());
            }
            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.exit(TC, "initialize", "Exception");
        }finally {
        	try {
                if(fisFileKeyStore!=null) fisFileKeyStore.close();
        	} catch(Exception e){};        	
        }

        this.subject = sub;
        this.handler = callbackHandler;

        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
            Tr.exit(TC, "initialize");
    }

    /**
     * Authenticates a user based on the keystore file.
     * 
     * @see LoginModule#login()
     */
    public boolean login() throws LoginException {

        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
            Tr.entry(TC, "login", this);
        if (debug) {
            System.out.println("[KeyStoreLoginModule] login: entry");
        }

        String name = null;
        char pwd[] = null;

        if (keyStore == null || subject == null || handler == null) {
            throw new LoginException("Module initialization failed"); 
        }

        NameCallback nameCallback = new NameCallback("Username:");
        PasswordCallback pwdCallback = new PasswordCallback("Password:", false);

        try {
            handler.handle(new Callback[] { nameCallback, pwdCallback });
        }
        catch (Exception e) {
            FFDCFilter.processException(e, CLASS_NAME + ".login", "223", this);

            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.exit(TC, "login", "SecurityException");
            throw new LoginException("Callback failed: " + e); 
        }

        name = nameCallback.getName();
        char[] tempPwd = pwdCallback.getPassword();

        if (tempPwd == null) {
            // treat a NULL password as an empty password
            tempPwd = new char[0];
        }
        pwd = new char[tempPwd.length];
        System.arraycopy(tempPwd, 0, pwd, 0, tempPwd.length);

        pwdCallback.clearPassword();

        if (debug) {
            System.out.println("[KeyStoreLoginModule] login: "
                    + "user entered user name: " + name);
        }
        if (ObjectGridManagerImpl.isTraceEnabled && TC.isDebugEnabled())
            Tr.debug(TC, "login", "userName="+name);

        // Validate the user name and password
        try {
            validate(name, pwd);
        }
        catch (SecurityException se) {
            FFDCFilter.processException(se, CLASS_NAME + ".login", "252", this);

            principals.clear();
            publicCreds.clear();
            privateCreds.clear();
            LoginException le = new LoginException(
            "Exception encountered during login");
            le.initCause(se);

            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.exit(TC, "login", "SecurityException");
            throw le;
        }

        if (debug) {
            System.out.println("[KeyStoreLoginModule] login: exit");
        }
        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
            Tr.exit(TC, "login", Boolean.TRUE);
        return true;
    }

    /**
     * Indicates the user is accepted. 
     * <p>
     * This method is called only if the user is authenticated by all modules in
     * the login configuration file. The principal objects will be added to the 
     * stored subject.
     * 
     * @return false if for some reason the principals cannot be added; true
     *         otherwise
     * 
     * @exception LoginException
     *                LoginException is thrown if the subject is readonly or if
     *                any unrecoverable exceptions is encountered.
     * 
     * @see LoginModule#commit()
     */
    public boolean commit() throws LoginException {
        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
            Tr.entry(TC, "commit", this);

        if (debug) {
            System.out.println("[KeyStoreLoginModule] commit: entry");
        }

        if (principals.isEmpty()) {
            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.entry(TC, "commit", "Exception-Commit is called out of sequence");
            throw new IllegalStateException("Commit is called out of sequence");
        }

        if (subject.isReadOnly()) {
            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.entry(TC, "commit", "Exception-Subject is Readonly");
            throw new LoginException("Subject is Readonly");
        }

        subject.getPrincipals().addAll(principals);
        subject.getPublicCredentials().addAll(publicCreds);
//        subject.getPrivateCredentials().addAll(privateCreds);

        principals.clear();
        publicCreds.clear();
        privateCreds.clear();

        if (debug) {
            System.out.println("[KeyStoreLoginModule] commit: exit");
        }
        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
            Tr.exit(TC, "commit", Boolean.TRUE);
        return true;
    }

    /**
     * Indicates the user is not accepted
     * 
     * @see LoginModule#abort()
     */
    public boolean abort() throws LoginException {
        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
            Tr.entry(TC, "abort", this);
        boolean b = logout();
        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
            Tr.entry(TC, "abort", new Boolean(b));
        return b;
    }

    /**
     * Logs the user out. Clear all the maps.
     * 
     * @see LoginModule#logout()
     */
    public boolean logout() throws LoginException {

        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
            Tr.entry(TC, "logout", this);

        // Clear the instance variables
        principals.clear();
        publicCreds.clear();
        privateCreds.clear();

        // clear maps in the subject
        if (!subject.isReadOnly()) {
            if (subject.getPrincipals() != null) {
                subject.getPrincipals().clear();
            }

            if (subject.getPublicCredentials() != null) {
                subject.getPublicCredentials().clear();
            }

            if (subject.getPrivateCredentials() != null) {
                subject.getPrivateCredentials().clear();
            }
        }
        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
            Tr.exit(TC, "logout", Boolean.TRUE);
        return true;
    }

    /**
     * Validates the user name and password based on the keystore.
     * 
     * @param userName user name
     * @param password password
     * @throws SecurityException if any exceptions encountered
     */
    private void validate(String userName, char password[])
        throws SecurityException {

        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
            Tr.entry(TC, "validate", new Object[]{this, userName, "xxxxxx"});

        PrivateKey privateKey = null;

        // Get the private key from the keystore
        try {
            privateKey = (PrivateKey) keyStore.getKey(userName, password);
        }
        catch (NoSuchAlgorithmException nsae) {
            FFDCFilter.processException(nsae, CLASS_NAME + ".validate", "399", this);

            SecurityException se = new SecurityException();
            se.initCause(nsae);
            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.exit(TC, "validate", "NoSuchAlgorithmException");
            throw se;
        }
        catch (KeyStoreException kse) {
            FFDCFilter.processException(kse, CLASS_NAME + ".validate", "406", this);
            SecurityException se = new SecurityException();
            se.initCause(kse);
            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.exit(TC, "validate", "KeyStoreException");
            throw se;
        }
        catch (UnrecoverableKeyException uke) {
            FFDCFilter.processException(uke, CLASS_NAME + ".validate", "412", this);
            SecurityException se = new SecurityException();
            se.initCause(uke);
            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.exit(TC, "validate", "UnrecoverableKeyException");
            throw se;
        }

        if (privateKey == null) {
            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.exit(TC, "validate", "Exception-Invalidate name");
            throw new SecurityException("Invalid name: " + userName); 
        }

        // Check the certificats
        Certificate certs[] = null;
        try {
            certs = keyStore.getCertificateChain(userName);
        }
        catch (KeyStoreException kse) {
            FFDCFilter.processException(kse, CLASS_NAME + ".validate", "430", this);
            SecurityException se = new SecurityException();
            se.initCause(kse);
            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.exit(TC, "validate", "KeyStoreException");
            throw se;
        }

        if (debug) {
            System.out.println("  Print out the certificates:");
            for (int i = 0; i < certs.length; i++) {
                System.out.println("  certificate " + i);
                System.out.println("    " + certs[i]);
            }
            Tr.debug(TC, "  End of printing out the certificates:");
        }

        if (certs != null && certs.length > 0) {

            // If the first certificate is an X509Certificate
            if (certs[0] instanceof X509Certificate) {
                try {
                    // Get the first certificate which represents the user
                    X509Certificate certX509 = (X509Certificate) certs[0];

                    // Create a principal
                    X500Principal principal = new X500Principal(certX509
                            .getIssuerDN()
                            .getName());
                    principals.add(principal);

                    if (debug) {
                        System.out.println("  Principal added: " + principal);
                    }
                    if (ObjectGridManagerImpl.isTraceEnabled && TC.isDebugEnabled())
                        Tr.debug(TC, "validate", new Object[] {
                                "principal added:" + principal, "principal name=" + principal.getName()});
                    // Create the certification path object and add it to the 
                    // public credential set
                    CertificateFactory factory = CertificateFactory
                        .getInstance("X.509");
                    java.security.cert.CertPath certPath = factory
                        .generateCertPath(Arrays.asList(certs));
                    publicCreds.add(certPath);

                    // Add the private credential to the private credential set
                    privateCreds.add(new X500PrivateCredential(certX509,
                            privateKey, userName));

                }
                catch (CertificateException ce) {
                    FFDCFilter.processException(ce, CLASS_NAME + ".validate", "476", this);
                    SecurityException se = new SecurityException();
                    se.initCause(ce);
                    if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                        Tr.exit(TC, "validate", "CertificateException");
                    throw se;
                }
            }
            else {
                // The first certificate is not an X509Certificate
                // We just add the certificate to the public credential set
                // and the private key to the private credential set.
                publicCreds.add(certs[0]);
                privateCreds.add(privateKey);
            }
        }
        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
            Tr.exit(TC, "validate");
    }
}