// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2006
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// 1.7 XD/ws/code/objectgrid.impl/src/com/ibm/websphere/objectgrid/security/plugins/builtins/LDAPAuthenticator.java, WAS.objectgrid, WASX.XD, b0614.107 3/16/06 17:40:16
//
// Module  :  LDAPAuthenticator.java
//
// Source File Description: 
//
//  Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD601_M2_Bx  WASX.XD   10/03/2005   jitang    ObjectGrid security support
//
// ------------ ---------- ------------ -------------------------------------------------
package com.ibm.websphere.objectgrid.security.plugins.builtins;

import javax.security.auth.Subject;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.security.plugins.Authenticator;
import com.ibm.websphere.objectgrid.security.plugins.Credential;
import com.ibm.websphere.objectgrid.security.plugins.ExpiredCredentialException;
import com.ibm.websphere.objectgrid.security.plugins.InvalidCredentialException;
import com.ibm.ws.ffdc.FFDCFilter;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.ObjectGridManagerImpl;
import com.ibm.ws.objectgrid.security.auth.callback.UserPasswordCallbackHandlerImpl;

/**
 * This class is an implementation of the <code>Authenticator</code> interface 
 * when a user name and password are used as a credential, and the authentication 
 * is against an LDAP server.
 * <p>
 * When basic authentication is used, the credential passed to the 
 * <code>authenticate(Credential)</code> method is a 
 * <code>UserPasswordCredential</code> object. 
 * <p>
 * This implementation uses the <code>SimpleLDAPLoginModule</code> class to authenticate 
 * the user into an LDAP server using the JAAS login module "LDAPLogin". The 
 * parameters for the LDAP server connection can be configured as options to the 
 * LDAPLoginModule. Please refer to the <code>SimpleLDAPLoginModule</code> class for 
 * more details about what properties should be configured.
 * <p>
 * This class is fully supported for production use.
 *
 * @since WAS XD 6.0.1
 * 
 * @see Authenticator
 * @see SimpleLDAPLoginModule
 * @see UserPasswordCredential
 */
public class LDAPAuthenticator implements Authenticator {

    private static final String CLASS_NAME = LDAPAuthenticator.class.getName();

    private static final TraceComponent TC = Tr.register(
            CLASS_NAME,
            Constants.TR_SECURITY_GROUP_NAME,
            Constants.TR_RESOURCE_BUNDLE_NAME);

    /**
     * Creates a new LDAPAuthenticator
     */
    public LDAPAuthenticator() {
        super();
    }

    /**
     * Authenticates a <code>UserPasswordCredential</code>.
     * <p>
     * Uses the user name and password from the specified UserPasswordCredential
     * to login to the LDAPLoginModule named "LDAPLogin".
     * 
     * @throws InvalidCredentialException if credential isn't a 
     *         UserPasswordCredential or some error occurs during processing
     *         of the supplied UserPasswordCredential
     * 
     * @throws ExpiredCredentialException if credential is expired.  This exception
     *         is not used by this implementation 
     * 
     * @see Authenticator#authenticate(Credential)
     * @see LDAPLoginModule
     */
    public Subject authenticate(Credential credential) throws InvalidCredentialException, ExpiredCredentialException {

        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
            Tr.entry(TC, "authenticate", credential);

        if (credential == null) {
            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.exit(
                    TC,
                    "authenticate",
                    "InvalidCredentialException(null credential)");
            throw new InvalidCredentialException("Supplied credential is null");
        }

        if (! (credential instanceof UserPasswordCredential) ) {
            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.exit(TC, "authenticate", "InvalidCredentialException(wrong credential type)");
            throw new InvalidCredentialException("Supplied credential is not a UserPasswordCredential");
        }

        UserPasswordCredential cred = (UserPasswordCredential) credential;
        LoginContext lc = null;
        try {
            lc = new LoginContext("LDAPLogin",
                    new UserPasswordCallbackHandlerImpl(cred.getUserName(), cred.getPassword().toCharArray()));

            if (ObjectGridManagerImpl.isTraceEnabled && TC.isDebugEnabled())
                Tr.debug(TC, "authenticate", "lc=" + lc);

            lc.login();

            Subject subject = lc.getSubject();

            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.exit(TC, "authenticate", subject);
            return subject;
        }
        catch (LoginException le) {
            FFDCFilter.processException(le, CLASS_NAME + ".authenticate", "120", this);
            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.exit(TC, "authenticate", le);
            throw new InvalidCredentialException(le);
        }
        catch (IllegalArgumentException ile) {
            FFDCFilter.processException(ile, CLASS_NAME + ".authenticate", "121", this);
            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.exit(TC, "authenticate", "IllegalArgumentException:"+ile);
            throw new InvalidCredentialException(ile);
        }
    }
}