// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2006
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// 1.5 XD/ws/code/objectgrid.impl/src/com/ibm/websphere/objectgrid/security/plugins/builtins/LDAPLoginModule.java, WAS.objectgrid, WASX.XD, b0614.107 2/21/06 20:31:05
//
// Module  :  LDAPLoginModule.java
//
// Source File Description:
//
//  Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD601_M2_Bx  WASX.XD   10/03/2005   jitang    ObjectGrid security support
// d343450      WXD601.XD 02/02/2006   jitang    Deal with hrdepartment field not existing in the LDAP server
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.security.plugins.builtins;

import java.util.Map;

import javax.naming.NamingException;
import javax.security.auth.Subject;
import javax.security.auth.callback.Callback;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.callback.NameCallback;
import javax.security.auth.callback.PasswordCallback;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginException;
import javax.security.auth.spi.LoginModule;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.ws.ffdc.FFDCFilter;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.ObjectGridManagerImpl;

/**
 * This LDAPLoginModule authenticates a user name and password to the
 * configured LDAP server.
 * <p>
 * A login configuration should provide at least the following two options:
 * <ul>
 * <li>providerURL: The LDAP server provider URL </li>
 * <li>factoryClass: The ldap context factory implementation class</li>
 * </ul>
 * <p>
 * Here is a Login module configuration example used by the sample
 * LDAPAuthenticator:
 * <pre><code>
 * LDAPLogin {
 *     com.ibm.websphere.objectgrid.security.plugins.builtins.LDAPLoginModule required
 *     providerURL="ldap://bluepages.ibm.com:389/"
 *     factoryClass="com.sun.jndi.ldap.LdapCtxFactory"
 *     debug=true;
 * };
 * </code></pre>
 *
 * The providerURL points to the IBM bluepages LDAP server with the port number
 * 389. The initial context factory is com.sun.jndi.ldap.LdapCtxFactory.
 *
 * @since WAS XD 6.0.1
 * @see LoginModule
 * @deprecated This plug-in is deprecated in version 8.6.  Use the {@link SimpleLDAPLoginModule} LoginModule.
 */
public class LDAPLoginModule implements LoginModule {

    private static final String CLASS_NAME = LDAPLoginModule.class.getName();

    private static final TraceComponent TC = Tr.register(
            CLASS_NAME,
            Constants.TR_SECURITY_GROUP_NAME,
            Constants.TR_RESOURCE_BUNDLE_NAME);

    /**
     * LDAP server provider URL property name
     */
    public static final String PROVIDER_URL = "providerURL";

    /**
     * LDAP context factory implementation class property name
     */
    public static final String FACTORY_CLASS = "factoryClass";

    private Subject subject;

    private CallbackHandler callbackHandler;

    private Map options;

    // configurable option
    private boolean debug = false;

    // the authentication status
    private boolean succeeded = false;

    private boolean commitSucceeded = false;

    // username and password
    private String username;

    private char[] password;

    // testUser's SamplePrincipal
    private SimpleUserPrincipal userPrincipal;
    private SimpleDeptPrincipal deptPrincipal;

    /**
     * Initializes this <code>LoginModule</code>.
     *
     * @param _subject the <code>Subject</code> to be authenticated.
     * @param _callbackHandler a <code>CallbackHandler</code> for communicating
     *          with the end user (prompting for user names and
     *          passwords, for example).
     * @param _sharedState shared <code>LoginModule</code> state.
     * @param _options options specified in the login
     *          <code>Configuration</code> for this particular
     *          <code>LoginModule</code>.
     */
    public void initialize(Subject _subject, CallbackHandler _callbackHandler, Map _sharedState,
            Map _options) {

        this.subject = _subject;
        this.callbackHandler = _callbackHandler;
        this.options = _options;

        if (subject == null)
            throw new IllegalArgumentException("Subject is not specified");

        if (callbackHandler == null)
            throw new IllegalArgumentException("CallbackHander is not specified");

        // initialize any configured options
        debug = "true".equalsIgnoreCase((String) options.get("debug"));
    }

    /**
     * Authenticates the user by a user name and password.
     *
     * @return true in all cases since this <code>LoginModule</code>
     *      should not be ignored.
     *
     * @exception FailedLoginException if the authentication fails.
     * @exception LoginException if this <code>LoginModule</code>
     *      is unable to perform the authentication.
     */
    public boolean login() throws LoginException {

        if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled() || debug)
            Tr.entry(TC, "login", this);

        if (subject == null || callbackHandler == null) {
            throw new LoginException("Module initialization failed");
        }

        NameCallback nameCallback = new NameCallback("Username:");
        PasswordCallback pwdCallback = new PasswordCallback("Password:", false);

        try {
            callbackHandler.handle(new Callback[] { nameCallback, pwdCallback });
        }
        catch (Exception e) {
            FFDCFilter.processException(e, CLASS_NAME + ".login", "223", this);

            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.exit(TC, "login", "SecurityException");
            throw new LoginException("Callback failed: " + e);
        }

        username = nameCallback.getName();
        char[] tempPwd = pwdCallback.getPassword();

        if (tempPwd == null) {
            // treat a NULL password as an empty password
            tempPwd = new char[0];
        }
        password = new char[tempPwd.length];
        System.arraycopy(tempPwd, 0, password, 0, tempPwd.length);

        pwdCallback.clearPassword();

        if (ObjectGridManagerImpl.isTraceEnabled && TC.isDebugEnabled() || debug)
            Tr.debug(TC, "[LDAPLoginModule] login: " + "user entered user name: " + username);

        // LDAP authentication
        LDAPAuthenticatorHelper helper = new LDAPAuthenticatorHelper((String) options
                .get(PROVIDER_URL), (String) options.get(FACTORY_CLASS));

        try {
            String[] principalNames = null;
            if ((principalNames = helper.authenticate(username, new String(password))) != null) {

                // authentication succeeded!!!
                if (ObjectGridManagerImpl.isTraceEnabled && TC.isDebugEnabled() || debug)
                    Tr.debug(TC, "[LDAPLoginModule] " + "authentication succeeded");

                userPrincipal = new SimpleUserPrincipal(principalNames[0]);

                if (principalNames[1] != null) {
                    deptPrincipal = new SimpleDeptPrincipal(principalNames[1]);
                }

                succeeded = true;
            }
            else {

                // authentication failed -- clean out state
                if (ObjectGridManagerImpl.isTraceEnabled && TC.isDebugEnabled() || debug)
                    Tr.debug(TC, "[LDAPLoginModule] " + "authentication failed");
                succeeded = false;
                username = null;
                for (int i = 0; i < password.length; i++)
                    password[i] = ' ';
                password = null;
                if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled() || debug)
                    Tr.exit(TC, "login", "FailedLoginException");
                throw new FailedLoginException("Login failed");
            }

            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled() || debug)
                Tr.exit(TC, "login", Boolean.TRUE);
            return true;
        }
        catch (NamingException ne) {
            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled() || debug)
                Tr.exit(TC, "login", ne);

            // Do not chain naming exception here because sometimes the LDAP context
            // object is not serializable, such as
            // com.sun.jndi.ldap.LdapCtx.
            LoginException le = new LoginException(ne.getMessage());
            throw le;
        }
    }

    /**
     * Called if the LoginContext's overall authentication succeeded
     * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
     * succeeded).
     * <p>
     * If this LoginModule's own authentication attempt succeeded (checked by
     * retrieving the private state saved by the <code>login</code> method),
     * this method associates a <code>SimpleUserPrincipal</code> and
     * <code>SimpleDeptPrincipal</code> with the <code>Subject</code> located in
     * the <code>LoginModule</code>.  If this LoginModule's own
     * authentication attempted failed, this method removes
     * any state that was originally saved.
     *
     * @exception LoginException if the commit fails.
     *
     * @return true if this LoginModule's own login and commit
     *      attempts succeeded, or false otherwise.
     */
    public boolean commit() throws LoginException {
        if (succeeded == false) {
            return false;
        }
        // add a Principal (authenticated identity)
        // to the Subject

        // assume the user we authenticated is the SamplePrincipal
        if (!subject.getPrincipals().contains(userPrincipal)) {
            subject.getPrincipals().add(userPrincipal);
        }

        if (deptPrincipal != null && !subject.getPrincipals().contains(deptPrincipal)) {
            subject.getPrincipals().add(deptPrincipal);
        }

        if (ObjectGridManagerImpl.isTraceEnabled && TC.isDebugEnabled() || debug) {
            Tr.debug(TC, "[LDAPLoginModule] " + "added SamplePrincipal to Subject");
        }

        // in any case, clean out state
        username = null;
        for (int i = 0; i < password.length; i++)
            password[i] = ' ';
        password = null;

        commitSucceeded = true;
        return true;
    }

    /**
     * Called if the LoginContext's overall authentication failed.
     * (the relevant REQUIRED, REQUISITE, SUFFICIENT and OPTIONAL LoginModules
     * did not succeed).
     * <p>
     * If this LoginModule's own authentication attempt
     * succeeded (checked by retrieving the private state saved by the
     * <code>login</code> and <code>commit</code> methods),
     * then this method cleans up any state that was originally saved.
     *
     * @exception LoginException if the abort fails.
     *
     * @return false if this LoginModule's own login and/or commit attempts
     *      failed, and true otherwise.
     */
    public boolean abort() throws LoginException {
        if (succeeded == false) {
            return false;
        }
        else if (succeeded == true && commitSucceeded == false) {
            // login succeeded but overall authentication failed
            succeeded = false;
            username = null;
            if (password != null) {
                for (int i = 0; i < password.length; i++)
                    password[i] = ' ';
                password = null;
            }
            userPrincipal = null;
        }
        else {
            // overall authentication succeeded and commit succeeded,
            // but someone else's commit failed
            logout();
        }
        return true;
    }

    /**
     * Logs out the user.
     * <p>
     * This method removes the <code>SamplePrincipal</code>
     * that was added by the <code>commit</code> method.
     *
     * @exception LoginException if the logout fails.
     *
     * @return true in all cases since this <code>LoginModule</code>
     *          should not be ignored.
     */
    public boolean logout() throws LoginException {

        subject.getPrincipals().remove(userPrincipal);
        succeeded = false;
        succeeded = commitSucceeded;
        username = null;
        if (password != null) {
            for (int i = 0; i < password.length; i++)
                password[i] = ' ';
            password = null;
        }
        userPrincipal = null;
        return true;
    }
}