// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2006
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// 1.3 XD/ws/code/objectgrid.core/src/com/ibm/websphere/objectgrid/security/plugins/builtins/WSSubjectSourceImpl.java, WAS.objectgrid, WASX.XD, b0614.107 2/14/06 18:43:11
//
// Module  :  WSSubjectSourceImpl.java
//
// Source File Description: 
//
//  Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M3_B.13 WASX.XD   04/01/2005   jitang    DataGrid - JAAS / Security - Drop 1 - Basic Support
// d288501      WXD60.XD  07/06/2005   vandersn  Add default constructor for plugin use
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.security.plugins.builtins;

import java.security.AccessController;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;

import javax.security.auth.Subject;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.security.ObjectGridSecurityException;
import com.ibm.websphere.objectgrid.security.plugins.SubjectSource;
import com.ibm.websphere.security.WSSecurityException;
import com.ibm.websphere.security.auth.WSSubject;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.ObjectGridManagerImpl;

/**
 * This class is a implementation of the <code>SubjectSource</code> plugin 
 * interface for use when an ObjectGrid is accessed in a WebSphere
 * Application Server.
 * <p> 
 * This implementation gets either a runAs Subject or a caller Subject 
 * from WebSphere Application Server thread. Users can use the 
 * <code>setType(int)</code> method to indicate what type of Subject will be 
 * retrieved. If <code>RUN_AS_SUBJECT</code> is used, the runAs Subject will be 
 * returned; if <code>CALLER_SUBJECT</code> is used, the caller Subject will be 
 * returned.
 *
 * @since WAS XD 6.0
 *
 * @see SubjectSource
 */
public class WSSubjectSourceImpl implements SubjectSource {

    /** 
     * A constant representing the runAs Subject type 
     */
    public static final int RUN_AS_SUBJECT = 1;

    /** 
     * A constant representing the caller Subject type 
     */
    public static final int CALLER_SUBJECT = 2;

    private static final String CLASS_NAME = WSSubjectSourceImpl.class.getName();

    static final TraceComponent TC = Tr.register(
            CLASS_NAME,
            Constants.TR_SECURITY_GROUP_NAME,
            Constants.TR_RESOURCE_BUNDLE_NAME);

    private int type = RUN_AS_SUBJECT;

    /**
     * Creates a new WSSubjectSourceImpl with the RunAs Subject type 
     */
    public WSSubjectSourceImpl(){
        super();
    }

    /**
     * Creates a new WSSubjectSourceImpl with the specified Subject type
     *
     * @param type the Subject type
     *
     * @throws IllegalArgumentException if the specified type is invalid
     */
    public WSSubjectSourceImpl(int type) {
        super();
        setType(type);
    }

    /**
     * Gets the subject type
     *
     * @return   the argument that was passed to the constructor or the 
     *           <code>setType(int)</code> method, or the default value of 
     *           <code>RUN_AS_SUBJECT</code> if the default constructor is used
     * 
     * @see #RUN_AS_SUBJECT
     * @see #setType(int) 
     */
    public int getType() {
        return type;
    }

    /**
     * Sets the subject type, either <code>RUN_AS_SUBJECT</code> or 
     * <code>CALLER_SUBJECT</code>
     *
     * @param type the subject type
     * 
     * @throws IllegalArgumentException if the specified type is invalid
     */
    public void setType(int type) {
        if (type != RUN_AS_SUBJECT && type != CALLER_SUBJECT) {
            throw new IllegalArgumentException("The passed in subject type is not a valid type.");
        }
        this.type = type;
    }

    /**
     * Converts a passed-in Subject to a new Subject object.
     *
     * @return the converted subject
     *
     * @throws ObjectGridSecurityException if an error occurs during processing
     *
     * @see SubjectSource#getSubject()
     */
    public Subject getSubject() throws ObjectGridSecurityException {

        try {
            final int finalType = type;
            return (Subject) AccessController.doPrivileged(new PrivilegedExceptionAction() {
                public Object run() throws ObjectGridSecurityException{
                    Subject s = null;
                    if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                        Tr.entry(TC, "getSubject", this);
                    try {
                        if (finalType == RUN_AS_SUBJECT) {
                            s = WSSubject.getRunAsSubject();
                        }
                        else if (finalType == CALLER_SUBJECT) {
                            s = WSSubject.getCallerSubject();
                        }
                    }
                    catch (WSSecurityException wse) {
                        throw new ObjectGridSecurityException(wse);
                    }
                    if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                        Tr.exit(TC, "getSubject", s);
                    return s;
                }
            });
        }
        catch (PrivilegedActionException pae) {
            if (pae.getException() instanceof ObjectGridSecurityException) {
                throw (ObjectGridSecurityException) pae.getException();
            }

            throw new ObjectGridSecurityException(pae);
        }
    }
}
