// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2006
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// 1.4 XD/ws/code/objectgrid.core/src/com/ibm/websphere/objectgrid/security/plugins/builtins/WSSubjectValidationImpl.java, WAS.objectgrid, WASX.XD, b0614.107 2/14/06 18:43:14
//
// Module  :  WSSubjectValidationImpl.java
//
// Source File Description: 
//
// Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M3_B.13 WASX.XD   04/08/2005   jitang    ObjectGrid - JAAS / Security - Drop 1 - Basic Support
//
// ------------ ---------- ------------ -------------------------------------------------
package com.ibm.websphere.objectgrid.security.plugins.builtins;

import java.security.AccessController;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;

import javax.security.auth.Subject;
import javax.security.auth.login.LoginContext;
import javax.security.auth.login.LoginException;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.security.plugins.InvalidSubjectException;
import com.ibm.websphere.objectgrid.security.plugins.SubjectValidation;
import com.ibm.ws.ffdc.FFDCFilter;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.ObjectGridManagerImpl;
import com.ibm.ws.objectgrid.security.auth.callback.WSCredTokenCallbackHandlerImpl;

/**
 * This class is an implementation of the <code>SubjectValidation</code>
 * interface in a WebSphere Application Server environment.
 * <p>
 * A typical scenario of using this class is described as below:
 * <UL>
 * <LI>The client has already been authenticated to WAS before it 
 * accesses ObjectGrid.</LI>
 * <LI>The client is running locally on the ObjectGrid server side.</LI>
 * <LI>The client retrieves the runAs Subject from the thread and 
 * passes the subject to ObjectGrid server.</LI>
 * <LI>The ObjectGrid runtime will call the <code>validateSubject(Subject)</code>
 * method to validate the subject has not been tampered with.</LI>
 * </UL>
 * 
 * Here is the code sample:
 * <pre>
 * //get the subject from the thread 
 * Subject subject = com.ibm.websphere.security.auth.WSSubject.getRunAsSubject();
 * 
 * // use the subject to get a Session object 
 * Session session = objectgrid.getSession(subject);
 * </pre>
 * 
 * @since WAS XD 6.0
 * 
 * @see InvalidSubjectException
 * @see SubjectValidation
 */
public class WSSubjectValidationImpl implements SubjectValidation {

    private static final String CLASS_NAME = WSSubjectValidationImpl.class.getName();

    static final TraceComponent TC = Tr.register(
            CLASS_NAME,
            Constants.TR_SECURITY_GROUP_NAME,
            Constants.TR_RESOURCE_BUNDLE_NAME);

    /**
     * Creates a new WSSubjectValidationImpl
     */
    public WSSubjectValidationImpl() {
        super();
    }

    /**
     * Validates that the subject has not been tampered with.
     *  
     * @see SubjectValidation#validateSubject(Subject)
     */
    public Subject validateSubject(final Subject subject) throws InvalidSubjectException {
        try {
            Subject s = (Subject) AccessController.doPrivileged(new PrivilegedExceptionAction() {
                public Object run() throws LoginException, SecurityException {
                    if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                        Tr.entry(TC, "validateSubject", new Object[] { this, subject });
                    LoginContext lc = new LoginContext("WSLogin",
                        new WSCredTokenCallbackHandlerImpl(subject));

                    lc.login();

                    Subject s1 = lc.getSubject();
                    if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                        Tr.exit(TC, "validateSubject", s1);
                    return s1;
                }
            });

            return s;
        }
        catch (PrivilegedActionException e) {
            FFDCFilter.processException(e, CLASS_NAME + ".validateSubject", "116", this);

            // No error will logged in plug-in

            if (ObjectGridManagerImpl.isTraceEnabled && TC.isEntryEnabled())
                Tr.exit(TC, "validateSubject", "LoginException");
            throw new InvalidSubjectException(e.getException());
        }
    }
}