// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2006-2007
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  Tuple.java
//
// Source File Description:
//
//    See javadoc.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// b995         6.1       05/22/2006   cdjohnson New part.
// b1591        6.1       07/19/2006   cdjohnson Feature enhancements for M4
// b5192        6.1       01/25/2007   cdjohnson Added equals and hashCode javadoc.
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.projector;

import com.ibm.websphere.projector.md.TupleMetadata;

/**
 * A Tuple is a data structure that holds primitive attributes and associations (relationships) to other tuples.  The metadata
 * for the Tuple is contained in the associated TupleMetadata.<p>
 * 
 * Tuple instances are created using the {@link TupleMetadata#createTuple()} method.<p>
 * 
 * The Tuple interface should not be directly implemented.
 *
 * @since WAS XD 6.1
 */
public interface Tuple
{
    
	/**
	 * This returns the metadata for this tuple.
	 * @return the metadata for this tuple.
	 */
	TupleMetadata getMetadata();

    /**
     * Retreive an attribute value from the specified position.
     * @param position the position of the attribute within this tuple.
     * @return the value of the attribute.
     * @throws IndexOutOfBoundsException if the position is greater than {@link TupleMetadata#getNumAttributes()}-1 or less than 0.
     */
    Object getAttribute(int position);

    /**
     * Set the attribute value at the specified position.
     * @param position the position of the attribute within this tuple.
     * @param value the value to set the attribute to.
     * @return the value previously at the specified position.
     * @throws IndexOutOfBoundsException if the position is greater than {@link TupleMetadata#getNumAttributes()}-1 or less than 0.
     */
   Object setAttribute(int position, Object value);

    /**
     * Set the attribute values at the specified positions.
     * @param values the values to set each attribute to.
     * @throws IndexOutOfBoundsException if the input array size is greater than {@link TupleMetadata#getNumAttributes()}.
     */
    void setAttributes(Object[] values);

    /**
     * Retrieve the association Tuple key from the specified position and index.
     * @param position the position of the association key(s) within this tuple.
     * @param index the index of an association key for a given association.
     * @return the Tuple key for the specified association index.
     * @throws IndexOutOfBoundsException if the position is greater than {@link TupleMetadata#getNumAssociations()()}-1 or less than 0 or if the
     * index is greater than {@link #getAssociationSize(int)} or less than 0;
     */
    Tuple getAssociation(int position, int index);

    /**
     * Retrieve all association Tuple keys from the specified position.
     * @param position the position of the association key(s) within this tuple.
     * @return the Tuple keys for the specified association.
     * @throws IndexOutOfBoundsException if the position is greater than {@link TupleMetadata#getNumAssociations()()}-1 or less than 0.
     */
    Tuple[] getAssociations(int position);

    /**
     * Add a single association key for an association.
     * @param position the position of the association key(s) within this tuple.
     * @param key the Tuple key to add to the current association.
     * @throws IndexOutOfBoundsException if the position is greater than {@link TupleMetadata#getNumAssociations()()}-1 or less than 0.
     */
    void addAssociationKey(int position, Tuple key);

    /**
     * Add multiple association keys for an association.
     * @param position the position of the association key(s) within this tuple.
     * @param keys the Tuple keys to add to the current association.
     * @throws IndexOutOfBoundsException if the position is greater than {@link TupleMetadata#getNumAssociations()()}-1 or less than 0.
     */
    void addAssociationKeys(int position, Tuple[] keys);

    /**
     * Retrieve the number of association keys for an association.
     * @param position the position of the association key(s) within this tuple.
     * @return the number of keys that the association has.
     */
    int getAssociationSize(int position);
    
    /**
     * Indicates whether a Tuple is "equal to" this one.<p>
     * 
     * Tuples are equal if the attributes and associations are equal.<p>
     *   
     * Attribute equality is determined by using the <tt>equals</tt> method on each
     * attribute.  Array attributes return true if they are "deeply equal".  Each
     * element in the array is checked for equality as a normal attribute.  Nested
     * arrays are therefore checked for equality.<p>
     * 
     * Association equality is determined by using the <tt>equals</tt> method on 
     * each association key Tuple.  If the association is a <tt>OneToMany</tt> or <tt>ManyToMany</tt>
     * relationship, the keys are compared using the order in which the keys were added to the 
     * association.<p>
     * 
     * @param   object the reference Tuple with which to compare.
     * @return  <code>true</code> if this object is the same as the object
     *          argument; <code>false</code> otherwise.
     * @see     #hashCode()
     * @see     java.util.Hashtable
     */
    boolean equals(Object object);
    
    /**
     * Returns the hash code value for this Tuple.  The hash code
     * complies with the fundamental contract of {@link java.lang.Object#hashCode()}
     * and is calculated based on the attributes and association keys within
     * this tuple as described in the <tt>equals</tt> method.
     * 
     * @return  a hash code value for this Tuple.
     * @see     java.lang.Object#equals(java.lang.Object)
     * @see     java.util.Hashtable
     */
    int hashCode();
    
}
