// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2006
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  AssociationType.java
//
// Source File Description:
//
//    See javadoc.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// b995         6.1       05/22/2006   cdjohnson New part.
// b2203        6.1       08/16/2006   cdjohnson Fixed toString()
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.projector.md;

import java.io.Serializable;


/**
 * The type of relationship or association between entities.
 * @see TupleAssociation
 * @since WAS XD 6.1
 */
public final class AssociationType implements Serializable
{
    
    private static final long serialVersionUID = 6316227490365985331L;

    /**
     * Specifies a single-valued association to another entity that has one-to-one
     * multiplicity. It is not normally necessary to specify the associated target entity explicitly since it can
     * usually be inferred from the type of the object being referenced.
     */
    public final static AssociationType OneToOne = new AssociationType(0);
    
    /**
     * Specifies a muliti-valued association to another entity that has one-to-many
     * multiplicity.
     */
    public final static AssociationType OneToMany = new AssociationType(1);
    
    /**
     * Specifies a single-valued association to another entity that has many-to-one
     * multiplicity. It is not normally necessary to specify the associated target entity explicitly since it can
     * usually be inferred from the type of the object being referenced.
     */
    public final static AssociationType ManyToOne = new AssociationType(2);

    /**
     * Specifies a many-valued association with many-to-many multiplicity. If the
     * Collection is defined using generics to specify the element type, the associated target entity class does
     * not need to be specified; otherwise it must be specified.<p>
     * 
     * Every many-to-many association has two sides, the owning side and the non-owning, or inverse, side.
     * If the association is bidirectional, either side may be designated as the owning side
     */
    public final static AssociationType ManyToMany = new AssociationType(3);
    
    private final static AssociationType[] PRIVATE_VALUES = new AssociationType[]{OneToOne, OneToMany, ManyToOne, ManyToMany};

    /**
     * Id that uniquely identifies the AssociationType instance.
     */
    final private int assocType;

    /**
     * Hidden CTOR for enumerated type idiom to use.
     *
     * @param assocType is the unique id for the contructed AssociationType instance.
     */
    private AssociationType( int assocType )
    {
        this.assocType = assocType;
    }

    /**
     * Hidden default CTOR for enumerated type idiom to use.
     */
    private AssociationType()
    {
        this(-1);
    }

    /**
     * Returns a string representation of the <code>AssociationType</code>.
     *
     * @return  a string representation of the <code>AssociationType</code>.
     */
    public String toString() {
        switch (assocType) {
        case 0: return "OneToOne"; //$NON-NLS-1$
        case 1: return "OneToMany"; //$NON-NLS-1$
        case 2: return "ManyToOne"; //$NON-NLS-1$
        case 3: return "ManyToMany"; //$NON-NLS-1$
        default: return "UNKNOWN: " + assocType; //$NON-NLS-1$
        }
    }
    
    /**
     * Retreive the value of this AssociationType.  Can be useful when 
     * using with a switch statement.
     * @return the int value of this AssociationType. 
     */
    public int getAssociationType() {
        return assocType;
    }
    
    /*
     * Required to avoid creating a different object instance
     * each tiem the object is deserialized.
     */
    private Object readResolve() { //throws ObjectStreamException {
        return PRIVATE_VALUES[assocType];
    }

    /**
     * Retrieves the hash code of this AssociationType object.
     * @return the hash code value of this object.
     */
    public int hashCode() {
        final int PRIME = 31;
        int result = 1;
        result = PRIME * result + assocType;
        return result;
    }

    /**
     * Returns true if the input object is equivalent to
     * this AssociationType.
     * 
     * @param obj the object with which to compare.
     * @return true if the input Object is equivalent to this AssociationType
     */
    public boolean equals(Object obj) {
        if (this == obj)
            return true;
        if (obj == null)
            return false;
        if (getClass() != obj.getClass())
            return false;
        final AssociationType other = (AssociationType) obj;
        if (assocType != other.assocType)
            return false;
        return true;
    }
    
    
}
